
(** * Algebra
  Copyright INRIA (2009) Apics Team (Jose Grimm).
*)

(* $Id: group0.v,v 1.17 2009/11/10 17:39:54 grimm Exp $ *)
Require Export group.
Require Export group1.
Require Export set2.

Set Implicit Arguments.


Module Transformation.

Definition Tcovers (a b : Bset) (f : EE) :=
  forall x, inc x b -> exists y, (inc y a) & (f y = x).
Definition Tinjects (a : Bset) (f : EE) :=
  forall x y , inc x a -> inc y a -> f x = f y -> x = y.

Definition Tinverse (a : Bset) (f : EE) (y : Bset) :=
  choose (fun x => (inc x a & f x = y)).

Definition Taxioms (a b: Bset) (f : EE) := forall x, inc x a -> inc (f x) b.

Lemma compose_Taxioms : forall a b c f g,
  Taxioms b c f -> Taxioms a b g -> Taxioms a c (fun x  => f (g x)).
Proof. uf Taxioms. ir.  auto.
Qed.

Lemma identity_Taxioms : forall a, Taxioms a a (fun x : Bset => x).
Proof. uf Taxioms.  tv. Qed.

Definition Tsurjective (a b : Bset) (f : EE) := 
  (Taxioms a b f & Tcovers a b f).

Lemma surjective_inverse_Taxioms : forall a b f,
  Tsurjective a b f -> Taxioms b a (Tinverse a f).
Proof. uf Tsurjective. uf Taxioms. uf Tinverse.  uf Tcovers. ir. ee.
  set (q:= (choose (fun x0 : Bset => inc x0 a & f x0 = x))).
  assert (inc q a & f q = x). uf q. app choose_pr.  app H1. nin H2; am.
Qed.

Lemma Tsurjective_inverse_right :  forall a b f x ,
  Tsurjective a b f -> inc x b -> f (Tinverse a f x) = x.
Proof. uf Tinverse. uf Tsurjective. uf Tcovers.  ir. ee.
  set (q:= (choose (fun x0  => inc x0 a & f x0 = x))).
  assert (inc q a & f q = x). uf q. app choose_pr.  app H1. nin H2; am.
Qed.

Definition Tinjective (a b : Bset) (f : EE) := 
  ( Taxioms a b f & Tinjects a f).

Definition Tbijective (a b : Bset) (f : EE) :=
  ( Taxioms a b f & Tsurjective a b f & Tinjective a b f).

Definition are_Tinverse (a b : Bset) (f g : EE) :=
  Taxioms a b f & Taxioms b a g
  & (forall x, inc x a -> g (f x) = x)
  & (forall y, inc y b -> f (g y) = y).

Lemma Tbijective_inverse :  forall a b f,
  Tbijective a b f -> are_Tinverse a b f (Tinverse a f).
Proof. uf  Tbijective; uf are_Tinverse. ir. destruct H as [H0 [H1 H2]].
  cp (surjective_inverse_Taxioms H1). ee. am. am.
  ir. cp (H0 _ H3). cp (Tsurjective_inverse_right H1 H4). 
  nin H2. app H6. app H. ir. app (Tsurjective_inverse_right H1 H3).
Qed.

Lemma Tinverse_bijective : forall a b f g, 
  are_Tinverse a b f g -> Tbijective a b f.
Proof. uf  Tbijective. uf are_Tinverse. ir. ee. am. red. ee. am.  red.
  ir. exists (g x). split. app H0. app H2. red. split. am. red.  ir.
  wrr H1. wr H5. sy; app H1. 
Qed.

Lemma Ttrans_sub_bijective : forall x y f u, Tinjective x y f -> sub u x ->
  Tbijective u (fun_image u f) f.
Proof. ir. assert (Taxioms u (fun_image u f) f). red. app inc_fun_image. 
  red. ee. am. red. ee. am. red. ir. rwi fun_image_rw H2. am.
  red. split. am. red in H; ee. red.  ir. app H2. app H0. app H0.
Qed. 

Lemma Tidentity_bijective : forall x, Tbijective x x (fun y => y).
Proof. ir. cp (identity_Taxioms (a:=x)). red. split. am. split. split.
  am. red. ir. exists x0. split. am. tv. split. am. red. tv.
Qed. 

Lemma Tsubidentity_injective : forall x y, sub x y ->
  Tinjective x y (fun z => z). 
Proof. ir. split. red. ir. app H. red. tv.
Qed.

Lemma Tcompose_injective : forall x y z f g,
  Tinjective x y f -> Tinjective y z g ->
  Tinjective x z (fun y=> g (f y)).
Proof. ir. nin H. nin H0. split. apply compose_Taxioms with y. am. am.
  red. ir. app H1. app H2. app H. app H.
Qed. 

Lemma Tcompose_surjective : forall x y z f g,
  Tsurjective x y f ->  Tsurjective y z g ->
  Tsurjective x z (fun y=> g (f y)).
Proof. ir. nin H; nin H0. split. apply compose_Taxioms with y;  am.
  red. ir. nin (H2 x0). ee. nin (H1 x1); ee. exists x2; ee; try am.
  rww H7. am. am. 
Qed. 

Lemma Tcompose_bijective : forall x y z f g,
  Tbijective x y f ->  Tbijective y z g ->
  Tbijective x z (fun y => g (f y)).
Proof.  uf Tbijective. ir. ee. apply compose_Taxioms with y; am.
  apply Tcompose_surjective with y; am.
  apply Tcompose_injective with y; am. 
Qed. 

(* M stands for Map *)

Definition Maxioms (a b f : Bset) :=
  fgraph f & domain f = a & sub (range f) b.

Lemma Mcompose_axioms : forall a b c f g,
   Maxioms b c f -> Maxioms a b g -> Maxioms a c (fcompose f g).
Proof. uf Maxioms. ir. assert (fcomposable f g). unfold fcomposable. ee.
  am. am.  rww H3. assert (fgraph (fcompose f g)).  
  app fcompose_axioms. ee. am. rww fcomposable_domain.  
  red. ir. awi H7. nin H7. cp (pr2_V H2 H7). awi H8. 
  cp (inc_pr1_domain H2 H7). rwi fcomposable_domain H9. rwi pr1_pair H9.  
  bwi H8. rw H8. app H6. app inc_V_range. rw H5. app H4. app inc_V_range. 
  am.  am.  am.  nin H2; am. 
Qed. 

Lemma identity_Maxioms : forall a, Maxioms a a (identity a).
Proof. ir. unfold Maxioms. ee. ap identity_axioms. rww identity_domain.
  uf identity. red. ir. rwi inc_create_range H. nin H. ee. wrr H0. 
Qed. 

Definition Msurjective (a b f : Bset) :=
  Maxioms a b f & Tcovers a b (fun x => V x f).

Definition Minjective (a b f : Bset) :=
  Maxioms a b f & Tinjects a (fun x  => V x f).

Definition Mbijective (a b f : Bset) :=
  Maxioms a b f & Msurjective a b f & Minjective a b f.


Definition isomap x y := choose (Mbijective x y).

Definition isotrans x y u:= V u (isomap x y).

Definition are_isomorphic x y :=
  exists f, Mbijective x y f.

Lemma trans_map_axioms : forall x y f,
  Taxioms x y f = Maxioms x y (L x f).
Proof. ir. ap iff_eq; ir. red. ee. gprops. bw. rw create_range. 
  red. ir. awi H0. nin H0. ee. wr H1.  app H.
  red. ir. red in H; ee. app H2. rw create_range. aw.  exists x0. ee. am. tv.
Qed.

Lemma trans_map_injective : forall x y f,
  Tinjective x y f = Minjective x y (L x f).
Proof. ir. uf Minjective.  uf Tinjective. rw trans_map_axioms. ap iff_eq; ir. 
  ee. am. red. ir. bwi H3. app H0. am. am. ee. am. red. ir.
  app H0. bw.
Qed.

Lemma trans_map_surjective : forall x y f,
  Tsurjective x y f = Msurjective x y (L x f).
Proof. uf Tsurjective; uf Msurjective. ir. rw trans_map_axioms. 
  ap iff_eq; ir. ee. am. red.  ir. nin (H0 _ H1). exists x1. nin H2. bw. 
  intuition. ee. am. red. ir.  nin (H0 _ H1).  nin H2. bwi H3. exists x1. 
  intuition. am.
Qed.

Lemma trans_map_bijective : forall x y f,
  Tbijective x y f = Mbijective x y (L x f).
Proof. uf Tbijective. uf Mbijective. ir. rw trans_map_axioms.
  rw  trans_map_surjective. rw  trans_map_injective. tv.
Qed.

Lemma trans_bij_isomorphic : forall x y f,
  Tbijective x y f -> are_isomorphic x y.
Proof. ir. red. exists  (L x f).  wrr trans_map_bijective.
Qed. 

Lemma isomap_bij : forall x y,
  are_isomorphic x y -> Mbijective x y (isomap x y). 
Proof. ir. uh H. cp (choose_pr H). am. 
Qed. 

Lemma create_rw : forall x f, x = domain f -> fgraph f -> 
  L x (fun x => V x f) = f. 
Proof. ir. rw H. app create_recovers.
Qed. 

Lemma isomap_isotrans_rw : forall x y,
  are_isomorphic x y -> isomap x y = L x (isotrans x y).
Proof. ir. cp (isomap_bij H). red in H0; ee. red in H0; ee. 
  uf isotrans.  rw create_rw. tv.  sy; am. am.
Qed.

Lemma isotrans_bij : forall x y,
  are_isomorphic x y -> Tbijective x y (isotrans x y).
Proof. ir. cp (isomap_bij H). rw trans_map_bijective. wrr isomap_isotrans_rw.
Qed. 

Lemma iso_isotrans_rw : forall x y,
  are_isomorphic x y = Tbijective x y (isotrans x y).
Proof. ir; ap iff_eq; ir.  app isotrans_bij. exists (L x (isotrans x y)). 
  wr trans_map_bijective. am. 
Qed. 

Lemma iso_trans_ex_rw : forall x y,
  are_isomorphic x y = (exists f, Tbijective x y f).
Proof. ir; ap iff_eq; ir. exists (isotrans x y). wr iso_isotrans_rw; am. 
  nin H. exists (L x x0). wrr trans_map_bijective. 
Qed. 

Lemma are_isomorphic_refl : forall x, are_isomorphic x x.
Proof. ir. red. exists (L x (fun y => y)). 
  wr trans_map_bijective. ap Tidentity_bijective. 
Qed.

Lemma are_inverse_symm : forall x y f g,
  are_Tinverse x y f g -> are_Tinverse y x g f.
Proof. ir. red. red in H. intuition.
Qed. 

Lemma are_isomorphic_symm : forall x y, 
  are_isomorphic x y -> are_isomorphic y x.
Proof. ir. rwi iso_isotrans_rw H. 
  exists (L y (Tinverse x (isotrans x y))).
  wr trans_map_bijective. apply Tinverse_bijective with (isotrans x y). 
  apply are_inverse_symm. app Tbijective_inverse. 
Qed.

Lemma are_isomorphic_trans : forall x z,
  (exists y, (are_isomorphic x y & are_isomorphic y z)) ->
  are_isomorphic x z.
Proof. ir. nin H. ee. rwi iso_isotrans_rw H. rwi iso_isotrans_rw H0. 
  rw iso_trans_ex_rw. 
  exists  (fun a => (isotrans x0 z (isotrans x x0 a))). 
  apply Tcompose_bijective with x0. am. am. 
Qed. 


Definition Acreate (x y a: Bset ) :=
  denote Source x (
    denote Target y (
      denote Arrow a (
        stop))). 


Lemma source_Acreate : forall x y a, 
  sourceC (Acreate x y a) = x. 
Proof. ir. uf sourceC; uf Acreate. drw.  
Qed. 

Lemma target_Acreate : forall x y a, 
  targetC (Acreate x y a) = y. 
Proof. ir. uf targetC; uf Acreate; drw. 
Qed. 

Lemma arrow_Acreate : forall x y a, 
  arrowC (Acreate x y a) = a. 
Proof. ir; uf arrowC; uf Acreate; drw.
Qed. 



Definition Alike a := a = Acreate (sourceC a) (targetC a) (arrowC a).

Lemma create_Alike : forall s t a,
  Alike (Acreate s t a) = True.
Proof. ir.  app iff_eq; ir.  uf Alike. rw source_Acreate. 
  rw target_Acreate.  rw arrow_Acreate. tv. 
Qed. 


Definition ev u x := V x (arrowC u). 

Section Construction.
  Variables (x y : Bset) (f : EE).

Definition Ucreate := Acreate x y (L (Ul x) f).


Lemma source_Ucreate : sourceC Ucreate = x. 
Proof. uf Ucreate. rw source_Acreate. tv. 
Qed. 

Lemma target_Ucreate : targetC Ucreate = y. 
Proof. uf Ucreate. rww target_Acreate.  
Qed. 

Lemma ev_Ucreate : forall a , inc a (Ul x) -> 
  ev Ucreate a = f a.
Proof. ir. uf ev. uf Ucreate. rw arrow_Acreate. bw.
Qed. 

Lemma arrow_Ucreate : arrowC Ucreate = L (Ul (sourceC Ucreate)) (ev Ucreate).
Proof. uf Ucreate. rw arrow_Acreate. rw source_Acreate. 
  uf ev. app function_extensionality1. 
  ir. rw arrow_Acreate.  bw.
Qed. 


End Construction.

Lemma Ucreate_extens :  forall x x' y y' f g,
  x = x' ->  y = y' ->
  (forall a, inc a (Ul x) -> f a = g a) -> Ucreate x y f = Ucreate x' y' g.
Proof. ir. wr H; wr H0. uf Ucreate. 
  ap uneq. app function_extensionality1. 
Qed.

Definition Urealizes (x y : Bset) (f : EE) (a : Bset) :=
  sourceC a = x
  & targetC a = y
  & (forall z, inc z (Ul x) -> ev a z = f z).

Definition Uproperty (x y : Bset) (f : EE) := Taxioms (Ul x) (Ul y) f.

Definition Ulike (a : Bset) := Ucreate (sourceC a) (targetC a) (ev a) = a.


Lemma Ucreate_like : forall x y f, Ulike (Ucreate x y f).
Proof. ir. uf Ulike. ap Ucreate_extens. ap source_Ucreate. 
  ap target_Ucreate. ir. ap ev_Ucreate. rwi source_Ucreate H. am. 
Qed. 

Lemma Ulike_Alike : forall u, Ulike u -> Alike u.
Proof. ir. red in H. red. ufi Ucreate H. 
  set (f:=L (Ul (sourceC u)) (ev u)) in H. 
  assert (arrowC u = f). wr H. rw arrow_Acreate. tv. rw H0. sy; am. 
Qed. 


Lemma Ucreate_realizes : forall x y f, 
  Urealizes x y f (Ucreate x y f).
Proof. ir. pose source_Ucreate. pose target_Ucreate. pose ev_Ucreate.
  unfold Urealizes. intuition.
Qed.

Lemma Urealizes_like_eq : forall x y f a,
  Ulike a -> Urealizes x y f a -> Ucreate x y f = a.
Proof. ir. ufi Ulike H. wr H. ap Ucreate_extens; ufi Urealizes H0; ee. 
  sy; am. ee; sy;am. ir; sy; ap H2; am. 
Qed. 


Definition Uaxioms (a : Bset) := Uproperty (sourceC a) (targetC a) (ev a).
Definition strong_Uaxioms (a : Bset) := (Uaxioms a & Ulike a).

Lemma Uaxioms_from_strong : forall a, strong_Uaxioms a -> Uaxioms a.
Proof.  ir. unfold strong_Uaxioms in H. ee;am.
Qed.

Lemma Ucreate_strong_axioms :  forall x y f, 
  Uproperty x y f -> strong_Uaxioms (Ucreate x y f).
Proof. ir. red. split. red. rw source_Ucreate. rw target_Ucreate.
  uf Uproperty. uf Taxioms. ir. rw ev_Ucreate. red in H.  red in H. app H.
  am. ap Ucreate_like.
Qed.

Lemma Ucreate_axioms : forall x y  f, 
  Uproperty x y f -> Uaxioms (Ucreate x y f).
Proof. ir. ap Uaxioms_from_strong. app Ucreate_strong_axioms.
Qed.

Definition Uidentity (x : Bset) := Ucreate x x (fun y : Bset => y).

Definition Ucompose (a b : Bset) :=
  Ucreate (sourceC b) (targetC a) (fun y  => ev a (ev b y)).

Lemma identity_strong_Uaxioms : forall x, strong_Uaxioms (Uidentity x).
Proof. ir.  uf Uidentity. ap Ucreate_strong_axioms.
  uf Uproperty. ap identity_Taxioms.
Qed.

Lemma identity_Uaxioms : forall x, Uaxioms (Uidentity x).
Proof.  ir. nin (identity_strong_Uaxioms x).  am.
Qed.

Lemma left_Uidentity :  forall a, strong_Uaxioms a -> 
  Ucompose (Uidentity (targetC a)) a = a.
Proof. ir. uf Ucompose. ap Urealizes_like_eq. nin H. am. uf Uidentity. 
  red. ee. tv. rww target_Ucreate. ir. rww ev_Ucreate. nin H. app H.
Qed.

Lemma right_Uidentity : forall a, strong_Uaxioms a -> 
  Ucompose a (Uidentity (sourceC a)) = a.
Proof. ir. uf Ucompose. ap Urealizes_like_eq. nin H; am.  red. uf Uidentity.
  rww source_Ucreate. ee. tv. tv. ir. rww ev_Ucreate.
Qed.

Lemma source_Uidentity : forall x, sourceC (Uidentity x) = x.
Proof.  ir. unfold Uidentity; rewrite source_Ucreate. tv.
Qed.

Lemma target_Uidentity : forall x, targetC (Uidentity x) = x.
Proof. ir. unfold Uidentity; rww target_Ucreate.
Qed.

Lemma ev_Uidentity : forall x y, inc y (Ul x) -> ev (Uidentity x) y = y.
Proof. ir. uf Uidentity. rw ev_Ucreate. tv. am.
Qed.

Lemma source_Ucompose : forall a b, sourceC (Ucompose a b) = sourceC b.
Proof. ir. uf Ucompose; rww source_Ucreate.
Qed.

Lemma target_Ucompose : forall a b, targetC (Ucompose a b) = targetC a.
Proof. ir. uf Ucompose; rww target_Ucreate.
Qed.

Lemma ev_Ucompose :  forall a b y,
  inc y (Ul (sourceC b)) -> ev (Ucompose a b) y = ev a (ev b y).
Proof. ir. uf Ucompose; rww ev_Ucreate.
Qed.

Definition Ucomposable (a b : Bset) :=
  (Uaxioms a & Uaxioms b & sourceC a = targetC b).

Lemma compose_strong_Uaxioms :
  forall a b, Ucomposable a b -> strong_Uaxioms (Ucompose a b).
Proof. ir. red. ee. red. rw source_Ucompose. rw target_Ucompose. red.
  red. ir. rw ev_Ucompose.  red in H. ee.  app H. rw H2. app H1. am.
  uf Ucompose. ap Ucreate_like.
Qed.

Lemma Ucompose_axioms : forall a b, Ucomposable a b -> Uaxioms (Ucompose a b).
Proof. ir. set (K := compose_strong_Uaxioms H). nin K; am. 
Qed. 

Lemma Uextens : forall a b,
  strong_Uaxioms a -> strong_Uaxioms b ->
  sourceC a = sourceC b -> targetC a = targetC b ->
  (forall x,
    inc x (Ul (sourceC a)) -> inc x (Ul (sourceC b)) -> ev a x = ev b x) ->
  a = b.
Proof. ir. red in H; red in H0; ee. red in H4. wr H4. sy.
  ap Urealizes_like_eq. am. red. ee. am. am. ir. app H3. rww H1.
Qed.

Lemma Uassociativity : forall a b c ,
  Ucomposable a b -> Ucomposable b c -> 
  Ucompose (Ucompose a b) c = Ucompose a (Ucompose b c).
Proof. ir. ap Uextens. ap compose_strong_Uaxioms. red. ee. ap Ucompose_axioms.
  am.  red in H0; ee; am. rw source_Ucompose.  red in H0; ee; am.
  ap compose_strong_Uaxioms. red. ee. red in H; ee; am. ap Ucompose_axioms.
  am. rw target_Ucompose. red in H; ee; am.  rw source_Ucompose.  
  rw source_Ucompose. rw source_Ucompose.  tv.
  rw target_Ucompose. rw target_Ucompose. rw target_Ucompose. tv.
  rw source_Ucompose. rw source_Ucompose.  rw source_Ucompose. ir.
  rw ev_Ucompose. rw ev_Ucompose. rw ev_Ucompose. rw ev_Ucompose. tv. am.
  rw source_Ucompose. am.  nin H0.  nin H3. rw H4. app H3. am.
Qed.

Definition Uinclusion (a b : Bset) := Ucreate a b (fun y : Bset => y).

Lemma inclusion_strong_Uaxioms : forall a b, sub (Ul a) (Ul b) -> 
  strong_Uaxioms (Uinclusion a b).
Proof. ir.  uf Uinclusion. ap Ucreate_strong_axioms. red. red.  ap H.
Qed.

Lemma inclusion_Uaxioms : forall a b, 
  sub (Ul a) (Ul b) -> Uaxioms (Uinclusion a b).
Proof. ir. ap Uaxioms_from_strong. app inclusion_strong_Uaxioms.
Qed.

Definition Uinjective (u : Bset) :=
  Uaxioms u & Tinjective (Ul (sourceC u)) (Ul (targetC u)) (ev u).

Lemma Uinjective_rewrite : forall u,
  Uaxioms u ->  Uinjective u =
  (forall x y,
    inc x (Ul (sourceC u)) ->  inc y (Ul (sourceC u)) -> ev u x = ev u y 
    -> x = y).
Proof. ir. ap iff_eq; ir. red in H0. ee. red in H4. ee. app H5.
  red. split. am.  red. red in H. red in H.   split. am. am.
Qed.

Lemma Uinjective_compose_with : forall u v w ,
  strong_Uaxioms u ->  strong_Uaxioms v ->
  Uinjective w ->
  Ucomposable w u -> Ucomposable w v -> Ucompose w u = Ucompose w v -> u = v.
Proof. ir. red in H2. red in H3. ee.
  app Uextens. transitivity (sourceC (Ucompose w u)). 
  rww source_Ucompose. rw H4. rww source_Ucompose.
  transitivity (sourceC w).  sy; am. am.
  ir. nin H1. nin H11. app H12.  rw H8. app H7. rw H6. app H5.
  transitivity (ev (Ucompose w u) x). rww ev_Ucompose. rw H4. rww ev_Ucompose. 
Qed. 

Definition Usurjective (u : Bset) :=
  Uaxioms u & Tsurjective (Ul (sourceC u)) (Ul (targetC u)) (ev u).

Lemma Usurjective_rewrite : forall u,
  Uaxioms u -> Usurjective u =
  (forall x,
    inc x (Ul (targetC u)) ->
    ex (fun y => (inc y (Ul (sourceC u)) & ev u y = x))).
Proof. ir. ap iff_eq; ir. red in H0. ee. nin H2.  app H3.
  red. split. am. red. split. red in H. red in H. am. red. ir. app H0.
Qed. 

Definition Uinverse (u : Bset) :=
  Ucreate (targetC u) (sourceC u) (Tinverse (Ul (sourceC u)) (ev u)).

Lemma surjective_inverse_Uaxioms :
 forall u, Usurjective u -> strong_Uaxioms (Uinverse u).
Proof. ir. red. uf Uinverse.  ee. ap Ucreate_axioms. red.
  ap surjective_inverse_Taxioms. red in H. ee. am.
  uf Tinverse. ap Ucreate_like.
Qed. 

Lemma Usurjective_composable_left :  forall u, Usurjective u -> 
  Ucomposable (Uinverse u) u.
Proof. ir. red. ee.  ap Uaxioms_from_strong. 
  ap surjective_inverse_Uaxioms. am.  red in H; ee; am.
  uf Uinverse. rww source_Ucreate.
Qed.

Lemma Usurjective_composable_right :
  forall u, Usurjective u -> Ucomposable u (Uinverse u).
Proof. ir. red. ee. red in H; ee; am. ap Uaxioms_from_strong. 
  ap surjective_inverse_Uaxioms. am.  uf Uinverse. rww target_Ucreate.
Qed. 

Lemma Usurjective_inverse_right :
  forall u, Usurjective u -> Ucompose u (Uinverse u) = Uidentity (targetC u).
Proof. ir. assert (Uaxioms u). red in H. ee; am.
  ap Uextens. ap compose_strong_Uaxioms. app Usurjective_composable_right.
  ap identity_strong_Uaxioms.
  rw source_Ucompose. uf Uinverse. rw source_Ucreate. rww source_Uidentity.
  rw target_Ucompose. rww target_Uidentity. ir. rwi source_Uidentity H2.
  rewrite ev_Ucompose. uf Uinverse. rww ev_Ucreate. rww ev_Uidentity.
  red in H. ee. ap (Tsurjective_inverse_right H3 H2).
  rwi source_Ucompose  H1. am.
Qed.

Definition Ubijective (u : Bset) := (Uinjective u & Usurjective u).

Lemma Ubijective_transformation_bijective :
 forall u, Ubijective u -> Tbijective (Ul (sourceC u)) (Ul (targetC u)) (ev u).
Proof. ir. red in H. ufi Uinjective H; ufi Usurjective H. red. ufi Uaxioms H.
  ufi Uproperty H. intuition.
Qed. 

Lemma Utransformation_bijective_bijective :  forall u,
 Uaxioms u ->
 Tbijective (Ul (sourceC u)) (Ul (targetC u)) (ev u) ->
 Ubijective u.
Proof. ir. red in H0.  red. uf Uinjective; uf Usurjective. intuition.
Qed. 

Definition are_Uinverse (u v : Bset) :=
  Ucomposable u v
  & Ucomposable v u
  & Ucompose u v = Uidentity (sourceC v)
  & Ucompose v u = Uidentity (sourceC u).

Lemma Uinverse_symm : forall u v, are_Uinverse u v -> are_Uinverse v u.
Proof.  uf are_Uinverse. ir. ee; am.
Qed. 

Lemma Uinverse_unique : forall u v w,
  strong_Uaxioms v -> strong_Uaxioms w -> 
  are_Uinverse u v -> are_Uinverse u w -> v = w.
Proof. ir. transitivity (Ucompose v (Uidentity (sourceC v))).
  sy. app right_Uidentity. red in H1; red in H2. ee.
  assert (sourceC v = targetC u). red in H6; ee; am.
  assert (sourceC v = sourceC w). rw H9. sy. red in H3; ee; am. rw H10.
  wr H4.  wr Uassociativity. rw H8.  red in H2. ee. rw H12.
  app left_Uidentity. am. am.
Qed. 

Definition has_Uinverse (u:Bset) := ex (fun v  => are_Uinverse u v).

Lemma are_Uinverse_Tinverse : forall u v,
  are_Uinverse u v ->
  are_Tinverse (Ul (sourceC u)) (Ul (targetC u)) (ev u) (ev v).
Proof. uf are_Uinverse. uf are_Tinverse. ir. ufi Ucomposable H. ee.
  red in H. am. wr H4. rw H6. red in H5. am.
  ir. transitivity (ev (Ucompose v u) x). sy; app ev_Ucompose. rw H2.
  app ev_Uidentity. ir. wri H4 H7. transitivity (ev (Ucompose u v) y).
  rww ev_Ucompose. rw H1. rww ev_Uidentity.
Qed. 

Lemma Uinverse_bijective : forall u v, are_Uinverse u v -> Ubijective u.
Proof. ir. ap Utransformation_bijective_bijective.  red in H; ee. red in H; ee.
  am. apply Tinverse_bijective with (ev v). app are_Uinverse_Tinverse.
Qed.

Lemma Ubijective_inverse : forall u, 
  Ubijective u -> are_Uinverse u (Uinverse u).
Proof. ir. red in H. ee.
  cp (Usurjective_composable_left H0).
  cp (Usurjective_composable_right H0).
  cp (Usurjective_inverse_right H0). 
  red. ee. am. am. rw H3. red in H1; ee. rw H5. app refl_equal.
  apply Uinjective_compose_with with u.
  app compose_strong_Uaxioms. app identity_strong_Uaxioms. am.
  red. rw target_Ucompose. ee.   red in H; ee. am. app Ucompose_axioms.
  red in H2. ee. am. red. split. red in H; ee; am. split.
  ap identity_Uaxioms. rww target_Uidentity. 
  wrr Uassociativity. rw H3. ap Uextens.
  ap compose_strong_Uaxioms. red. ee. ap identity_Uaxioms.  red in H; ee; am.
  rww source_Uidentity. ap compose_strong_Uaxioms. red. ee. red in H; ee; am.
  ap identity_Uaxioms. rww target_Uidentity. rw source_Ucompose.
  rww source_Ucompose. rw source_Uidentity. tv. rw target_Ucompose.
  rw target_Ucompose. rw target_Uidentity. tv. rw source_Ucompose.
  rw source_Ucompose. ir.
  repeat rewrite ev_Ucompose. repeat rewrite ev_Uidentity. tv. am.
  red in H. ee.  red in H. ee. app H. am. am.
Qed.


Lemma Ubijective_has_inverse : forall u, Ubijective u -> has_Uinverse u.
Proof. ir. red. exists (Uinverse u). app Ubijective_inverse.
Qed. 


Lemma Ubijective_eq_has_inverse : forall u, Ubijective u = has_Uinverse u.
Proof.  ir. ap iff_eq; ir. ap Ubijective_has_inverse. am.  nin H.  
  apply Uinverse_bijective with x. am.
Qed.

Lemma Ubijective_eq_inverse :
  forall u, Ubijective u = are_Uinverse u (Uinverse u).
Proof. ir. ap iff_eq; ir. app Ubijective_inverse.
  apply Uinverse_bijective with (Uinverse u); am.
Qed. 

Lemma Uhas_inverse_eq_inverse :
  forall u, has_Uinverse u = are_Uinverse u (Uinverse u).
Proof. ir. wr Ubijective_eq_inverse. rww Ubijective_eq_has_inverse.
Qed. 

Definition Uempty := stop.

Definition Uempty_initial (a : Bset) := Uinclusion Uempty a.

Lemma underlying_Uempty : Ul Uempty = emptyset.
Proof.  uf Uempty. uf Ul. drw. 
Qed. 

Lemma Uempty_initial_strong_axioms :
  forall a, strong_Uaxioms (Uempty_initial a).
Proof. ir. uf Uempty_initial. ap inclusion_strong_Uaxioms. 
  rw underlying_Uempty. app sub_emptyset_any.
Qed. 

Lemma Uuse_axioms : forall u x y a,
  Uaxioms u ->
  sourceC u = x -> targetC u = y -> inc a (Ul x) -> inc (ev u a) (Ul y).
Proof. ir. red in H. ee. red in H. red in H.  wr H1. app H. rww H0.
Qed. 

End Transformation.
Export Transformation.

Module Barrow.


Import Transformation.

Definition mor_saxioms := strong_Uaxioms.
Definition mor_axioms := Uaxioms.

Definition sourceU x := Ul (sourceC x). 
Definition targetU x := Ul (targetC x). 


Lemma map_axioms_from_mor:forall m, mor_saxioms m 
  -> Maxioms (sourceU m) (targetU m) (arrowC m).
Proof. uf sourceU; uf targetU. ir. red in H. red in H. ee. red in H0. 
  assert (arrowC (Ucreate (sourceC m) (targetC m) (ev m)) = arrowC m).
  rww H0. rwi arrow_Ucreate H1. rwi H0 H1. wr H1.
  wr trans_map_axioms. red in H. red in H. am.
Qed.

Lemma function_axioms_from_mor:forall m, 
  mor_saxioms m -> fgraph (arrowC m).
Proof. ir. nin (map_axioms_from_mor H). am.
Qed.

Lemma mor_domain:forall m,
  mor_saxioms m -> domain (arrowC m) = sourceU m.
Proof. ir. nin (map_axioms_from_mor H). nin H1; am.
Qed.

Lemma mor_target:forall m, 
  mor_saxioms m ->sub (range (arrowC m)) (targetU m).
Proof. ir. nin(map_axioms_from_mor H). nin H1. am.
Qed.

Lemma mor_inc_target:forall m x, 
  mor_saxioms m ->inc x (sourceU m) -> inc (ev m x) (targetU m).
Proof. ir. app mor_target.
  assert (inc x (domain (arrowC m))). rww mor_domain.
  rw frange_inc_rw. exists x. split. am. uf ev. tv.
  app function_axioms_from_mor.
Qed.

Lemma mor_inc_ev: forall f x, mor_axioms f ->
  inc x (sourceU f) -> inc (ev f x) (targetU f).
Proof. ir. red in H. red in H.  red in H. app H. Qed.


Definition mor_image (x m: Bset) := fun_image x (fun u=> ev m u).
Definition mor_inv_image (x m:Bset):= inverse_image x (arrowC m).
Definition mor_range (m:Bset) :=   range (L (sourceU m)(ev m)).

Lemma mor_image_inc_rw: forall x m y, 
  inc y (mor_image x m)= exists z, (inc z x & ev m z=y).
Proof. ir. uf mor_image. ap iff_eq. ir.  awi H. am.  ir.  aw.
Qed.

Lemma mor_range_inc_rw: forall f y,
  inc y (mor_range f) = (exists x, inc x (sourceU f) & ev f x = y).
Proof. ir. uf mor_range. rw frange_inc_rw. rw create_domain.
  app iff_eq. ir. nin H. ee. exists x.  bwi H0.
  split. am. sy;am. am. ir. nin H. exists x. bw.
  ee. am. sy. am. ee; am. gprops.
Qed.

Lemma range_is_image: forall f,
  mor_range f= mor_image (sourceU f) f.
Proof. ir. set_extens. rwi mor_range_inc_rw H. rw mor_image_inc_rw. am.
  rw mor_range_inc_rw. rwi mor_image_inc_rw H. am.
Qed.


Lemma mor_inv_image_pr : forall a m x , 
  mor_saxioms m->
  inc x (mor_inv_image a m) -> (inc x (sourceU m) & inc (ev m x) a).
Proof. ir. split. ufi mor_inv_image H0.  red in H. red in H. ee. red in H1. 
  wri H1 H0. rwi arrow_Ucreate H0. rwi H1 H0.
  pose (f := (L (sourceU m) (ev m))).
  assert (inc x (domain f)). app (inverse_image_sub H0).
  unfold f in H2. rwi create_domain  H2. am.
  uf ev. ufi mor_inv_image H. app inverse_image_pr. 
Qed. 


Lemma mor_inv_image_pr_s : forall a m x, 
  inc x (mor_inv_image a m) ->  inc (ev m x) a.
Proof. ir. uf ev. ufi mor_inv_image H. app inverse_image_pr. 
Qed. 

Lemma mor_inv_image_inc : forall a m x, mor_saxioms m ->
    inc x (sourceU m) -> inc (ev m x) a -> inc x (mor_inv_image a m). 
Proof. ir. uf mor_inv_image. app inverse_image_inc. rww mor_domain.
Qed. 

Lemma inverse_image_sub : 
  forall a m, mor_saxioms m-> sub (mor_inv_image a m) (sourceU m).
Proof. ir. uf mor_inv_image. wrr mor_domain. app inverse_image_sub. 
Qed.

Lemma increasing_image: forall m x z,
  sub x z -> sub (mor_image x m) (mor_image z m).
Proof. ir. red. ir. rwi mor_image_inc_rw H0. nin H0. ee. 
  rw mor_image_inc_rw. exists x1. ee. app H. am.
Qed.


Lemma increasing_invimage: forall m x z,
  mor_saxioms m->  sub x z -> sub (mor_inv_image x m) (mor_inv_image z m).
Proof. ir. red. ir. pose (mor_inv_image_pr H H1). ee.
  app mor_inv_image_inc. app H0.
Qed.


Lemma sub_mor_inv: forall m z, 
  sub (mor_image (mor_inv_image z m) m) z.
Proof. ir. red. ir. rwi mor_image_inc_rw H. nin H. ee. wrr H0.
  app mor_inv_image_pr_s.
Qed.


Lemma sub_inv_image: forall m z w,
  mor_saxioms m-> sub z (sourceU m) -> 
  sub (mor_image z m) w -> sub z (mor_inv_image w m).
Proof. ir. red. ir. app mor_inv_image_inc.  app H0. app H1. 
  rw mor_image_inc_rw. exists x. split.  am. tv.
Qed.

Lemma sub_inv_image2: forall m z w, 
  mor_saxioms m-> sub z (sourceU m) -> 
  sub z (mor_inv_image w m)-> sub (mor_image z m) w.  
Proof. ir. red. ir. rwi mor_image_inc_rw H2. nin H2. ee. red in H1.
  wr H3. app mor_inv_image_pr_s.  app H1. 
Qed.

Definition mor_restr (m z:Bset):=
  Ucreate z (targetC m) (ev m).

Lemma source_mor_restr: forall f z, sourceC (mor_restr f z) = z.
Proof. ir. uf mor_restr. app source_Ucreate. Qed.

Lemma target_mor_restr: forall f z, targetC (mor_restr f z) = targetC f.
Proof. ir. uf mor_restr. app target_Ucreate. Qed.

Lemma ev_mor_restr:forall a f z, 
  inc a (Ul z) -> ev (mor_restr f z) a = ev f a.
Proof. ir. uf mor_restr. app ev_Ucreate. Qed.

Lemma surjective_pr: forall f a,
  Usurjective f -> inc a (targetU f) ->
  (ev f (Tinverse (sourceU f) (ev f) a)  =a 
  &inc (Tinverse (sourceU f) (ev f) a ) (sourceU f)).
Proof. ir. induction H.  induction H1. red in H2. 
  pose (H2 _ H0). assert(Hy:= choose_pr e). induction Hy. split. am. am. 
Qed.


Definition mor_to_arrow (f:Bset) (hyp: mor_axioms f) (x:(sourceU f)) :=
 chooseT (fun a:(targetU f) => Ro a = ev f (Ro x))
 (inc_nonempty(hyp (Ro x)(R_inc x))).

Lemma mor_to_arrow_pr: forall f (x:(sourceU f)) (hyp:mor_axioms f),
  Ro(mor_to_arrow hyp x) = ev f (Ro x).
Proof. ir. uf mor_to_arrow. app chooseT_pr. apply (hyp (Ro x) (R_inc x)). Qed.


End Barrow.
Export Barrow.

Module Blist.
Import List.

(* nonempty lists *)


Inductive nelist(x:Bset):Bset :=
  singl: x-> nelist x
| necons : x-> nelist x -> nelist x.

Fixpoint nelist_to_list(x:Bset)(u:nelist x): list x :=
  match u with 
    | singl u => cons u nil
    | necons a b => cons a (nelist_to_list b)
end.
 
Fixpoint list_to_nelist(x:Bset)(u:list x)(e:x): nelist x :=
  match u with
    | nil => singl e
    | cons a nil => singl a
    | cons a b => necons a (list_to_nelist b e) end.

Fixpoint ne_append (x:Bset) (l1 l2: nelist x){struct l1}: nelist x :=
  match l1 with 
   | singl a  => necons a l2
   | necons a b => necons a (ne_append b l2)
end.

Fixpoint ne_flatten(x:Bset)(l:nelist (nelist x)):nelist x :=
  match l with
    | singl a => a
    | necons a b => ne_append a (ne_flatten b)
end.

Fixpoint flatten(x:Bset)(l:list (list x)):list x :=
  match l with
    | nil => nil
    | cons a b => app a (flatten b)
end.


Lemma nonempty_nelist: forall (x:Bset) (e:x) (a:nelist x),
  nelist_to_list a <> nil. 
Proof. ir. induction a. unfold nelist_to_list.  discriminate. 
  unfold nelist_to_list.  discriminate. 
Qed.

Lemma list_is_nelist: forall (x:Bset) (e:x) (a:list x),
  a <> nil ->
  nelist_to_list (list_to_nelist a e) = a.
Proof. ir. induction a. elim H. tv. 
  simpl. induction a0. simpl. tv.
  set (u:=list_to_nelist (a0 :: a1) e). simpl. uf u. rw IHa. tv. discriminate.
Qed.

Lemma nelist_is_nelist: forall (x:Bset) (e:x) (a:nelist x),
  list_to_nelist (nelist_to_list a) e = a.
Proof. ir. induction a. simpl. tv.  simpl. rw IHa.
  set (w:= nelist_to_list a). assert (w <> nil). uf w. app nonempty_nelist.
  induction w. elim H; tv. tv.
Qed.

Lemma check_append: forall (x:Bset)(a b:nelist x),
  app(nelist_to_list a)(nelist_to_list b) = nelist_to_list (ne_append a b).
Proof. ir. induction a. simpl. tv. simpl. rww IHa. 
Qed.

Lemma fun_simpl2 :
  forall (x y z:Bset)(u : x->y->z), forall (a a':x)(b b':y),
    a=a' -> b=b' -> u a b = u a' b'.
Proof. ir. rw H. rw H0. tv. Qed.

Lemma nelist_simpl2:forall (x:Bset)(y:x)(u v: nelist x),
  u =v -> necons y u =necons y v.
Proof. ir. tv.  rww H. Qed.


End Blist.
Export Blist.
