Require Export group.
Set Implicit Arguments.
Unset Strict Implicit.
Module Basic.

(* $Id: group1.v,v 1.8 2009/11/09 18:15:13 grimm Exp $ *)

Hint Resolve group_is_monoid cgroup_is_group ring_is_cgroup
  sfield_is_ring lmodule_is_cgroup rmodule_is_cgroup lrmodule_is_cgroup : mp.

Hint Resolve lmodule_aux_is_ring rmodule_aux_is_ring
  llrmodule_aux_is_ring rlrmodule_aux_is_ring
  lrmodule_is_lmodule lrmodule_is_rmodule: mp.

Ltac mp:= auto with mp.

Lemma mzero_U: forall g, monoid_axioms g  -> inc (zero g) (Ul g).
Proof. ir. destruct H as [H _]. am. Qed.

Lemma gzero_U: forall g, group_axioms g  -> inc (zero g) (Ul g).
Proof. ir. ap mzero_U. mp. 
Qed.

Lemma cgzero_U: forall g, cgroup_axioms g  -> inc (zero g) (Ul g).
Proof. ir. ap gzero_U. mp. Qed.

Lemma rzero_U: forall g, ring_axioms g  -> inc (zero g) (Ul g).
Proof. ir. app cgzero_U. mp. Qed.

Lemma fzero_U: forall g, sfield_axioms g  -> inc (zero g) (Ul g).
Proof. ir. app rzero_U. mp. Qed.

Lemma lmzero_U: forall g, lmodule_axioms g  -> inc (zero g) (Ul g).
Proof. ir. ap cgzero_U. mp. Qed.

Lemma rmzero_U: forall g, rmodule_axioms g  -> inc (zero g) (Ul g).
Proof. ir. ap cgzero_U. mp. Qed.

Lemma lrmzero_U: forall g, lrmodule_axioms g  -> inc (zero g) (Ul g).
Proof. ir. ap cgzero_U. mp. Qed.

Lemma rone_U: forall g, ring_axioms g  -> inc (one g) (Ul g).
Proof. ir. destruct H as [_ Rp]. destruct Rp as [_ [o _]]. app o. 
Qed.

Lemma fone_U: forall g, sfield_axioms g  -> inc (one g) (Ul g).
Proof. ir. ap rone_U. mp. Qed.


Lemma mplus_internal:  forall g x y, monoid_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. destruct H as [_ [H _]].  app H. Qed.

Lemma gplus_internal:  forall g x y, group_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app mplus_internal. mp. Qed.

Lemma cgplus_internal:  forall g x y, cgroup_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app gplus_internal. mp. Qed.

Lemma rplus_internal:  forall g x y, ring_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app cgplus_internal. mp. Qed.

Lemma fplus_internal:  forall g x y, sfield_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app rplus_internal. mp. Qed.

Lemma lmplus_internal:  forall g x y, lmodule_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app cgplus_internal. mp. Qed.

Lemma rmplus_internal:  forall g x y, rmodule_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app cgplus_internal. mp. Qed.

Lemma lrmplus_internal:  forall g x y, lrmodule_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dplus g x y) (Ul g).
Proof. ir. app cgplus_internal. mp. Qed.


Lemma gneg_internal:  forall g x, group_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. destruct H as [_ [H _]]. app H. Qed.

Lemma cgneg_internal: forall g x, cgroup_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. app gneg_internal. mp. Qed.

Lemma rneg_internal: forall g x, ring_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. app cgneg_internal. mp. Qed.

Lemma fneg_internal: forall g x , sfield_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. app rneg_internal. mp. Qed.

Lemma lmneg_internal: forall g x, lmodule_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. app cgneg_internal. mp. Qed.

Lemma rmneg_internal: forall g x, rmodule_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. app cgneg_internal. mp. Qed.

Lemma lrmneg_internal: forall g x, lrmodule_axioms g-> 
  inc x (Ul g) -> inc (neg g x) (Ul g).
Proof. ir. app cgneg_internal. mp. Qed.

Lemma rmult_internal:  forall g x y, ring_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dmult g x y) (Ul g).
Proof. ir. destruct H as [_ [H _]]. app H. Qed.

Lemma fmult_internal:  forall g x y, sfield_axioms g-> 
  inc x (Ul g) -> inc y (Ul g) -> inc (dmult g x y) (Ul g).
Proof. ir. app rmult_internal. mp. Qed.

Lemma lmmult_internal:  forall g x y, lmodule_axioms g-> 
  inc x (Ul (lauxiliary g)) -> inc y (Ul g) -> inc (lmult g x y) (Ul g).
Proof. ir. destruct H as [_ [_ H]]. destruct H as [_ [_ [_ [H _]]]]. app H. 
Qed.

Lemma rmmult_internal:  forall g x y, rmodule_axioms g-> 
  inc x (Ul (rauxiliary g)) -> inc y (Ul g) -> inc (rmult g x y) (Ul g).
Proof. ir. destruct H as [_ [_ H]]. destruct H as [_ [_ [_ [H _]]]]. app H. 
Qed.

Lemma llrmmult_internal:  forall g x y, lrmodule_axioms g-> 
  inc x (Ul (lauxiliary g)) -> inc y (Ul g) -> inc (lmult g x y) (Ul g).
Proof. ir. destruct H as [_ [_ H]]. red in H. ee.  app H8.
Qed.

Lemma rlrmmult_internal:  forall g x y, lrmodule_axioms g-> 
  inc x (Ul (rauxiliary g)) -> inc y (Ul g) -> inc (rmult g x y) (Ul g).
Proof.  ir. destruct H as [_ [_ H]]. red in H. ee. app H10.
Qed.


Lemma mplus_assoc: forall g x y z, monoid_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. destruct H as [_ [_ [a _]]]. app a. Qed.

Lemma gplus_assoc: forall g x y z, group_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app mplus_assoc. mp. Qed.

Lemma cgplus_assoc: forall g x y z, cgroup_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app gplus_assoc. mp. Qed.

Lemma rplus_assoc: forall g x y z, ring_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app cgplus_assoc. mp. Qed.

Lemma fplus_assoc: forall g x y z, sfield_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app rplus_assoc. mp. Qed.

Lemma lmplus_assoc: forall g x y z, lmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app mplus_assoc. mp. Qed.

Lemma rmplus_assoc: forall g x y z, rmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app mplus_assoc. mp. Qed.

Lemma lrplus_assoc: forall g x y z, lrmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dplus g (dplus g x y) z = dplus g x (dplus g y z).
Proof. ir. app mplus_assoc. mp. Qed.


Lemma cgplus_comm: forall g x y, cgroup_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dplus g x y) = dplus g y x.
Proof. intros. destruct H as [_ G].  app G. Qed.

Lemma rplus_comm: forall g x y, ring_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dplus g x y) = dplus g y x.
Proof. ir. app cgplus_comm. mp. Qed.

Lemma fplus_comm: forall g x y, sfield_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dplus g x y) = dplus g y x.
Proof. ir. app rplus_comm. mp. Qed.

Lemma lmplus_comm: forall g x y, lmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dplus g x y) = dplus g y x.
Proof. ir. app cgplus_comm. mp. Qed.

Lemma rmplus_comm: forall g x y, rmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dplus g x y) = dplus g y x.
Proof. ir. app cgplus_comm. mp. Qed.

Lemma lrmplus_comm: forall g x y, lrmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dplus g x y) = dplus g y x.
Proof. ir. app cgplus_comm. mp. Qed.

Lemma cfmult_comm: forall g x y, field_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dmult g x y) = dmult g y x.
Proof. ir. destruct H as [_ H]. app H. Qed.

Lemma crmult_comm: forall g x y, cring_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> (dmult g x y) = dmult g y x.
Proof. ir. destruct H as [_ H]. app H. Qed.


Lemma rmult_assoc: forall g x y z, ring_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dmult g (dmult g x y) z = dmult g x (dmult g y z).
Proof. ir. destruct H as [_ [_ [_ [a _]]]]. app a. Qed.

Lemma fmult_assoc: forall g x y z, sfield_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g)-> 
    dmult g (dmult g x y) z = dmult g x (dmult g y z).
Proof. ir.  app rmult_assoc. mp. Qed.

Lemma lmmult_assoc: forall g x y z, lmodule_axioms g
  ->inc x (Ul (lauxiliary g)) -> inc y (Ul (lauxiliary g)) -> inc z (Ul g)-> 
  lmult g x (lmult g y z) = lmult g (dmult (lauxiliary g) x y) z.
Proof. ir. destruct H as [_ [_ H]]. destruct H as  [_ [_ [a _]]]. app a. Qed.

Lemma rmmult_assoc: forall g x y z, rmodule_axioms g
  ->inc x (Ul (rauxiliary g)) -> inc y (Ul (rauxiliary g)) -> inc z (Ul g)-> 
  rmult g x (rmult g y z) = rmult g (dmult (rauxiliary g) y x) z.
Proof. ir. destruct H as [_ [_ H]]. destruct H as  [_ [_ [a _]]]. app a. Qed.

Lemma llrmmult_assoc: forall g x y z, lrmodule_axioms g
  ->inc x (Ul (lauxiliary g)) -> inc y (Ul (lauxiliary g)) -> inc z (Ul g)-> 
  lmult g x (lmult g y z) = lmult g (dmult (lauxiliary g) x y) z.
Proof. ir. destruct H as [_ [_ H]]. red in H; ee. app H7. Qed.

Lemma rlrmmult_assoc: forall g x y z, lrmodule_axioms g
  ->inc x (Ul (rauxiliary g)) -> inc y (Ul (rauxiliary g)) -> inc z (Ul g)-> 
  rmult g x (rmult g y z) = rmult g (dmult (rauxiliary g) y x) z.
Proof.  ir. destruct H as [_ [_ H]]. red in H; ee. app H8. Qed.

Lemma rdistributive_left: forall g x y z, ring_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g) ->  
    dmult g x (dplus g y z) = dplus g (dmult g x y) (dmult g x z).
Proof. ir. destruct H as [_ [_ [_ [_ [_ [_ [H _]]]]]]].  app H.  Qed.

Lemma rdistributive_right: forall g x y z, ring_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g) ->  
    dmult g (dplus g y z) x = dplus g (dmult g y x) (dmult g z x).
Proof. ir. destruct H as [_ [_ [_ [_ [_ [_ [_ H]]]]]]]. app H.  Qed.

Lemma fdistributive_left: forall g x y z, sfield_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g) ->  
    dmult g x (dplus g y z) = dplus g (dmult g x y) (dmult g x z).
Proof. ir. app rdistributive_left; mp. Qed.

Lemma fdistributive_right: forall g x y z, sfield_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> inc z (Ul g) ->  
    dmult g (dplus g y z) x = dplus g (dmult g y x) (dmult g z x).
Proof. ir. app rdistributive_right; mp. Qed.

Lemma lmdistributive_left: forall g x y z, lmodule_axioms g
  ->inc x (Ul (lauxiliary g)) -> inc y (Ul g) -> inc z (Ul g) ->  
    lmult g x (dplus g y z) = dplus g (lmult g x y) (lmult g x z).
Proof. ir. destruct H as [_ H].  destruct H as [_ [H _]]. app H. Qed.

Lemma lmdistributive_right: forall g x y z, lmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul (lauxiliary g)) -> inc z (Ul (lauxiliary g)) ->  
    lmult g (dplus (lauxiliary g) y z) x = dplus g (lmult g y x) (lmult g z x).
Proof. ir. destruct H as [_ H]. destruct H as [_ [_ [H _]]].  app H. Qed.

Lemma rmdistributive_left: forall g x y z, rmodule_axioms g
  ->inc x (Ul (rauxiliary g)) -> inc y (Ul g) -> inc z (Ul g) ->  
    rmult g x (dplus g y z) = dplus g (rmult g x y) (rmult g x z).
Proof. ir. destruct H as [_ H].  destruct H as [_ [H _]]. app H. Qed.

Lemma rmdistributive_right: forall g x y z, rmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul (rauxiliary g)) -> inc z (Ul (rauxiliary g)) ->  
    rmult g (dplus (rauxiliary g) y z) x = dplus g (rmult g y x) (rmult g z x).
Proof. ir. destruct H as [_ H]. destruct H as [_ [_ [H _]]].  app H. Qed.

Lemma llrmdistributive_left: forall g x y z, lrmodule_axioms g
  ->inc x (Ul (lauxiliary g)) -> inc y (Ul g) -> inc z (Ul g) ->  
    lmult g x (dplus g y z) = dplus g (lmult g x y) (lmult g x z).
Proof. ir. destruct H as [_ [_ H]].  red in H; ee.  app H3. Qed.

Lemma llrmdistributive_right: forall g x y z, lrmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul (lauxiliary g)) -> inc z (Ul (lauxiliary g)) ->  
    lmult g (dplus (lauxiliary g) y z) x = dplus g (lmult g y x) (lmult g z x).
Proof. ir. destruct H as [_ [_ H]]. red in H; ee. app H4. Qed.

Lemma rlrmdistributive_left: forall g x y z, lrmodule_axioms g
  ->inc x (Ul (rauxiliary g)) -> inc y (Ul g) -> inc z (Ul g) ->  
    rmult g x (dplus g y z) = dplus g (rmult g x y) (rmult g x z).
Proof. ir. destruct H as [_ [_ H]]. red in H; ee. app H5. Qed.

Lemma rlrmdistributive_right: forall g x y z, lrmodule_axioms g
  ->inc x (Ul g) -> inc y (Ul (rauxiliary g)) -> inc z (Ul (rauxiliary g)) ->  
    rmult g (dplus (rauxiliary g) y z) x = dplus g (rmult g y x) (rmult g z x).
Proof. ir. destruct H as [_ [_ H]]. red in H; ee.  app H6. Qed.

Lemma mzero_unit_right: forall g x, monoid_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. intros.  destruct H as [_ [_ [_ [H _]]]]. app H. Qed.

Lemma mzero_unit_left: forall g x, monoid_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. intros.  destruct H as [_ [_ [_ [_ H]]]]. app H. Qed.

Lemma gzero_unit_right: forall g x, group_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app mzero_unit_right. mp. Qed.

Lemma gzero_unit_left: forall g x, group_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app mzero_unit_left. mp. Qed.

Lemma cgzero_unit_right: forall g x, cgroup_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app gzero_unit_right. mp. Qed.

Lemma cgzero_unit_left: forall g x, cgroup_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app gzero_unit_left. mp. Qed.

Lemma rzero_unit_right: forall g x, ring_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app cgzero_unit_right. mp. Qed.

Lemma rzero_unit_left: forall g x, ring_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app cgzero_unit_left. mp. Qed.

Lemma fzero_unit_right: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app rzero_unit_right. mp. Qed.

Lemma fzero_unit_left: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app rzero_unit_left. mp. Qed.

Lemma lmzero_unit_right: forall g x, lmodule_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app cgzero_unit_right. mp. Qed.

Lemma lmzero_unit_left: forall g x, lmodule_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app cgzero_unit_left. mp. Qed.

Lemma rmzero_unit_right: forall g x, rmodule_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app cgzero_unit_right. mp. Qed.

Lemma rmzero_unit_left: forall g x, rmodule_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app cgzero_unit_left. mp. Qed.

Lemma lrzero_unit_right: forall g x, lrmodule_axioms g
  -> inc x (Ul g)  -> dplus g x (zero g)  = x.
Proof. ir. app cgzero_unit_right. mp. Qed.

Lemma lrzero_unit_left: forall g x, lrmodule_axioms g
  -> inc x (Ul g)  -> dplus g (zero g) x = x.
Proof. ir. app cgzero_unit_left. mp. Qed.

Lemma rone_unit_right: forall g x, ring_axioms g
  -> inc x (Ul g)  -> dmult g x (one g)  = x.
Proof. intros. destruct H as [_ [_ [_ [_ [H _]]]]]. app H. Qed.

Lemma rone_unit_left: forall g x, ring_axioms g
  -> inc x (Ul g)  -> dmult g (one g) x = x.
Proof. intros. destruct H as [_ [_ [_ [_ [_ [H _]]]]]]. app H. Qed.

Lemma fone_unit_right: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> dmult g x (one g)  = x.
Proof. ir. app rone_unit_right. mp. Qed.

Lemma fone_unit_left: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> dmult g (one g) x = x.
Proof. ir. app rone_unit_left. mp. Qed.

Lemma lmone_unit_left: forall g x, lmodule_axioms g
  -> inc x (Ul g)  -> lmult g (one (lauxiliary g)) x = x.
Proof. ir. destruct H as [_ H]. destruct H as [_ [_ [_ [_ [_ H]]]]].  app H. 
Qed.

Lemma rmone_unit_left: forall g x, rmodule_axioms g
  -> inc x (Ul g)  -> rmult g (one (rauxiliary g)) x  = x.
Proof. ir. destruct H as [_ H]. destruct H as [_ [_ [_ [_ [_ H]]]]].  app H. 
Qed.

Lemma llrmone_unit_left: forall g x, lrmodule_axioms g
  -> inc x (Ul g)  -> lmult g (one (lauxiliary g)) x = x.
Proof. ir. destruct H as [_ H]. destruct H as [_ [_ [_ [_ [_ H]]]]].  
   destruct H as [_ [_ [_ [_ [H _]]]]]. app H. 
Qed.

Lemma rlrmone_unit_left: forall g x, lrmodule_axioms g
  -> inc x (Ul g)  -> rmult g (one (rauxiliary g)) x  = x.
Proof. ir. destruct H as [_ H]. destruct H as [_ [_ [_ [_ [_ H]]]]].  
   destruct H as [_ [_ [_ [_ [_ [_ H]]]]]]. app H. 
Qed.

Lemma gopp_right: forall g x, group_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. intros. destruct H as [_ [_ [H _]]]. app H. Qed.

Lemma gopp_left: forall g x, group_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. intros. destruct H as [_ [_ [_ H]]]. app H. Qed.

Lemma cgopp_right: forall g x, cgroup_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. ir. app gopp_right. mp. Qed.

Lemma cgopp_left: forall g x, cgroup_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. ir. app gopp_left. mp. Qed.


Lemma ropp_right: forall g x, ring_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. ir. app cgopp_right. mp. Qed.

Lemma ropp_left: forall g x, ring_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. ir. app cgopp_left. mp. Qed.


Lemma fopp_right: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. ir. app ropp_right. mp. Qed.

Lemma fopp_left: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. ir. app ropp_left. mp. Qed.

Lemma lmopp_right: forall g x, lmodule_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. ir. app cgopp_right. mp. Qed.

Lemma lmopp_left: forall g x, lmodule_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. ir. app cgopp_left. mp. Qed.

Lemma rmopp_right: forall g x, rmodule_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. ir. app cgopp_right. mp. Qed.

Lemma rmopp_left: forall g x, rmodule_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. ir. app cgopp_left. mp. Qed.

Lemma lrmopp_right: forall g x, lrmodule_axioms g
  -> inc x (Ul g)  -> dplus g x (neg g x) = zero g. 
Proof. ir. app cgopp_right. mp. Qed.

Lemma lrmopp_left: forall g x, lrmodule_axioms g
  -> inc x (Ul g)  -> dplus g (neg g x) x = zero g.
Proof. ir. app cgopp_left. mp. Qed.


Lemma finv_right: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> x <> (zero g) -> dmult g x (inv g x) = one g. 
Proof. intros. destruct H as [_ [_ [H _]]]. app H. Qed.

Lemma finv_left: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> x <> (zero g) -> dmult g (inv g x) x = one g.
Proof. intros. destruct H as  [_ [H _]]. app H. Qed.

Lemma ginv_internal: forall g x, sfield_axioms g
  -> inc x (Ul g)  -> inc (inv g x) (Ul g).
Proof. intros. destruct H as  [_ [_ [_ [H _]]]]. app H. Qed.




Hint Resolve mzero_U gzero_U cgzero_U rzero_U fzero_U lmzero_U 
  rmzero_U lrmzero_U :mp.
Hint Resolve rone_U fone_U : mp.
Hint Resolve mplus_internal gplus_internal cgplus_internal  rplus_internal
  fplus_internal lmplus_internal rmplus_internal lrmplus_internal : mp.

Hint Resolve  gneg_internal cgneg_internal  rneg_internal
  fneg_internal lmneg_internal rmneg_internal lrmneg_internal : mp.
Hint Resolve rmult_internal fmult_internal lmmult_internal rmmult_internal 
  llrmmult_internal rlrmmult_internal ginv_internal : mp.

End Basic.
Export Basic.

Module Basic1.

Lemma gneg_a_a_b_is_b :forall g a b, group_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)->  dplus g (neg g a) (dplus g a b)  = b.
Proof. ir. wrr gplus_assoc. rww gopp_left. rww gzero_unit_left. mp.
Qed.

Lemma gb_a_neg_a_is_b :forall g a b, group_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)->  dplus g (dplus g b a)  (neg g a) = b.
Proof. intros. rww gplus_assoc. rww gopp_right. rww gzero_unit_right. mp.
Qed.

Lemma ga_neg_a_b_is_b :forall g a b, group_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)->  dplus g a (dplus g (neg g a) b)  = b.
Proof. ir.  wrr gplus_assoc. rww gopp_right. rww gzero_unit_left. mp.
Qed.

Lemma gb_bneg_a_a_is_b :forall g a b, group_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)->  dplus g (dplus g b (neg g a)) a = b.
Proof. intros. rww gplus_assoc. rww gopp_left. rww gzero_unit_right. mp.
Qed.

Lemma gleft_simpl: forall g a x y,
  group_axioms g -> inc a (Ul g)-> inc x (Ul g)-> inc y (Ul g)
  -> dplus g a x = dplus g a y -> x = y.
Proof. intros. rewrite <- gneg_a_a_b_is_b with g a x; try am. rw H3.
  rewrite gneg_a_a_b_is_b with g a y; try am. tv. 
Qed.
 
Lemma gright_simpl: forall g a x y,
  group_axioms g -> inc a (Ul g)-> inc x (Ul g)-> inc y (Ul g)
  -> dplus g x a = dplus g y a -> x = y.
Proof. intros. rewrite <- gb_a_neg_a_is_b with g a x; try am. rw H3.
  rewrite gb_a_neg_a_is_b with g a y; try am. tv. 
Qed.

Lemma gzero_unique_right: forall g a b,
  group_axioms g -> inc a (Ul g)-> inc b (Ul g)
  -> dplus g a b = a -> b = zero g.
Proof. ir. apply gleft_simpl with g a. am. mp. am. mp. rww gzero_unit_right.
Qed.

Lemma gzero_unique_left: forall g a b,
  group_axioms g -> inc a (Ul g)-> inc b (Ul g)
  -> dplus g b a = a -> b = zero g.
Proof. ir. apply gright_simpl with g a. am. mp. am. mp. rww gzero_unit_left.
Qed.

Lemma rzero_left_abs: forall g a,
  ring_axioms g -> inc a (Ul g) ->  dmult g (zero g) a = (zero g).
Proof. ir. set (x:= dmult g (zero g) a). assert (inc x (Ul g)). unfold x. mp.
  assert (inc (zero g) (Ul g)). mp.
  apply gzero_unique_right with x; try am. mp. unfold x.
  wrr rdistributive_right. rww rzero_unit_right.
Qed.

Lemma rzero_right_abs: forall g a,
  ring_axioms g -> inc a (Ul g) ->  dmult g a (zero g) = (zero g).
Proof. ir. set (x:= dmult g a (zero g)). assert (inc x (Ul g)). unfold x. mp.
  assert (inc (zero g) (Ul g)). mp.
  apply gzero_unique_right with x; try am. mp. unfold x. 
  wrr rdistributive_left. rww rzero_unit_right.
Qed.


Lemma fone_not_zero:  forall g, sfield_axioms g  ->
  (one g) <> (zero g).
Proof. ir. assert (Ha:=H).
  red in H.  red in H. ee.  nin H3. ee. red; ir. red in H4. app H4.
  pose (fone_unit_right  Ha H3). rwi H5 e. rwi rzero_right_abs e.
  sy. am. mp. am.
Qed.

Lemma lmzero_left_abs: forall g a,
  lmodule_axioms g -> inc a (Ul g) -> 
  lmult g (zero (lauxiliary g)) a = (zero g).
Proof. ir. set (D:= lauxiliary g). assert (ring_axioms D).  unfold D; mp.
  assert (inc (zero D) (Ul D)). mp.
   set (x:= lmult g  (zero D) a). assert (inc x (Ul g)). unfold x. mp.
   apply gzero_unique_right with x; try am. mp. unfold x.
   wrr lmdistributive_right.  fold D. rww rzero_unit_right.
Qed.

Lemma rmzero_left_abs: forall g a,
  rmodule_axioms g -> inc a (Ul g) -> 
  rmult g (zero (rauxiliary g)) a = (zero g).
Proof. ir. set (D:= rauxiliary g). assert (ring_axioms D).  unfold D; mp.
  assert (inc (zero D) (Ul D)). mp.
  set (x:= rmult g  (zero D) a). assert (inc x (Ul g)). unfold x. mp.
  apply gzero_unique_right with x; try am. mp. unfold x.
  wrr rmdistributive_right.  fold D. rww rzero_unit_right.
Qed.

Lemma llrmzero_left_abs: forall g a,
  lrmodule_axioms g -> inc a (Ul g) -> 
  lmult g (zero (lauxiliary g)) a = (zero g).
Proof. ir. app lmzero_left_abs. mp. Qed.

Lemma rlrmzero_left_abs: forall g a,
  lrmodule_axioms g -> inc a (Ul g) -> 
  rmult g (zero (rauxiliary g)) a = (zero g).
Proof. ir. app rmzero_left_abs. mp. Qed.

Lemma gright_inv_unique: forall g a x,
  group_axioms g -> inc a (Ul g)-> inc x (Ul g)
  -> dplus g x a = zero g  -> x = neg g a.
Proof. intros. rewrite <- gopp_left with (x:=a) in H2; try am.
  apply gright_simpl with g a; try am.  mp.
Qed.

Lemma gleft_inv_unique: forall g a x,
  group_axioms g -> inc a (Ul g)-> inc x (Ul g)
  -> dplus g x a = zero g  -> a = neg g x.
Proof. intros. rewrite <- gopp_right with (x:=x) in H2; try am.
  apply gleft_simpl with g x; try am. mp.
Qed.

Lemma gunit_self_inverse:  forall g, group_axioms g 
  ->  neg g (zero g) = zero g.
Proof. intros. set (x:= zero g). assert (inc (zero g) (Ul g)). mp.
  assert (dplus g x x=x). pose (H2:=gzero_unit_left H H0). 
  unfold x. am. symmetry. app gleft_inv_unique.  
Qed.

Lemma gdouble_inverse: forall g a, group_axioms g ->
   inc a (Ul g) ->  neg g (neg g a) = a.
Proof. intros. assert (dplus g (neg g a) a = (zero g)).
  app gopp_left. symmetry. app gleft_inv_unique.  mp.
Qed.

Lemma prod_inverse: forall g a b,
  group_axioms g -> inc a (Ul g)-> inc b (Ul g) ->
  neg g (dplus g a b) = dplus g (neg g b) (neg g a).
Proof. intros. symmetry.
  assert (inc (neg g a)  (Ul g)). mp.
  assert (inc (neg g b)  (Ul g)). mp.
  assert (inc (dplus g (neg g b) (neg g a)) (Ul g)).  mp.
  app gleft_inv_unique. mp.
  rewrite gplus_assoc; try am.  rww ga_neg_a_b_is_b.
  app gopp_right. 
Qed.



Lemma finv_a_a_b_is_b :forall g a b, sfield_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)-> a <> (zero g) -> 
  dmult g (inv g a) (dmult g a b)  = b.
Proof. ir. wrr fmult_assoc. rww finv_left. rww fone_unit_left. mp.
Qed.

Lemma fb_a_inv_a_is_b :forall g a b, sfield_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)->  a <> (zero g) -> 
  dmult g (dmult g b a)  (inv g a) = b.
Proof. intros. rww fmult_assoc. rww finv_right. rww fone_unit_right. mp.
Qed.

Lemma fa_inv_a_b_is_b :forall g a b, sfield_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)->  a <> (zero g) -> 
  dmult g a (dmult g (inv g a) b)  = b.
Proof. ir. wrr fmult_assoc. rww finv_right. rww fone_unit_left. mp.
Qed.

Lemma fb_binv_a_a_is_b :forall g a b, sfield_axioms g -> 
  inc a (Ul g)-> inc b (Ul g)-> a <> (zero g) -> 
   dmult g (dmult g b (inv g a)) a = b.
Proof. intros. rww fmult_assoc. rww finv_left. rww fone_unit_right. mp.
Qed.

Lemma fleft_simpl: forall g a x y,
  sfield_axioms g -> inc a (Ul g)-> inc x (Ul g)-> inc y (Ul g)-> a <> (zero g)
  -> dmult g a x = dmult g a y -> x = y.
Proof. intros. rewrite <- finv_a_a_b_is_b with g a x; try am. rw H4.
  rewrite finv_a_a_b_is_b with g a y; try am. tv. 
Qed.
 
Lemma funit_self_inverse:  forall g, sfield_axioms g 
  ->  inv g (one g) = one g.
Proof. intros. set (x:= one g). assert (inc (one g) (Ul g)). mp.
  assert (x = dmult g x x). pose (H2:=fone_unit_left H H0). sy. 
  unfold x. am. set (y := (inv g x)).
  assert (dmult g y x = dmult g y (dmult g x x)). wr H1. tv.
  wri fmult_assoc H2; try am.
  assert (dmult g y x = y). unfold x. app fone_unit_right. uf y.  mp.
  assert (dmult g y x = x). unfold y;unfold x. app finv_left. 
  app fone_not_zero.  wr H3. am. unfold y. mp. 
Qed.

Lemma fright_simpl: forall g a x y,
  sfield_axioms g -> inc a (Ul g)-> inc x (Ul g)-> inc y (Ul g)-> a <> (zero g)  
  -> dmult g x a = dmult g y a -> x = y.
Proof. intros. rewrite <- fb_a_inv_a_is_b with g a x; try am. rw H4.
  rewrite fb_a_inv_a_is_b with g a y; try am. tv. 
Qed.
 
Lemma fright_inv_unique: forall g a x,
  sfield_axioms g -> inc a (Ul g)-> inc x (Ul g)
  -> dmult g x a = one g  -> x = inv g a.
Proof. intros. apply by_cases with (a= zero g).
  ir. assert (zero g = one g). rwi H3 H2. rwi rzero_right_abs H2.  tv. 
  mp. am. symmetry in H4. elim (fone_not_zero H H4). 
  ir. rewrite <- finv_left with (x:=a) in H2; try am.
  apply fright_simpl with g a; try am.  mp.
Qed.

Lemma fleft_inv_unique: forall g a x,
  sfield_axioms g -> inc a (Ul g)-> inc x (Ul g)
  -> dmult g x a = one g  -> a = inv g x.
Proof. ir. apply by_cases with (x= zero g).
  ir. assert (zero g = one g). rwi H3 H2. rwi rzero_left_abs H2. tv. mp. 
  am. symmetry in H4. elim (fone_not_zero H H4).
  ir. rewrite <- finv_right with (x:=x) in H2; try am.
  apply fleft_simpl with g x; try am. mp.
Qed.

Lemma fdouble_inverse: forall g a, sfield_axioms g ->
   inc a (Ul g) -> a <> (zero g) -> inv g (inv g a) = a.
Proof. intros. assert (dmult g (inv g a) a = (one g)).
  app finv_left. symmetry. app fleft_inv_unique.  mp.
Qed.

Lemma sfield_integral_domain :  forall g a b,
  sfield_axioms g -> inc a (Ul g)-> inc b (Ul g) ->
  (dmult g a b) = (zero g) -> (a= (zero g)) \/ (b= (zero g)).
Proof. ir. apply by_cases with (a= zero g). ir. left. tv.
  ir. right. assert (ring_axioms g). mp.  pose (rzero_right_abs H4 H0).
  symmetry in e. rwi e H2. apply fleft_simpl with g a; try am. mp.
Qed.


Lemma fprod_inverse: forall g a b,
  sfield_axioms g -> inc a (Ul g)-> inc b (Ul g) -> a <> (zero g)-> b <> (zero g)
  ->  inv g (dmult g a b) = dmult g (inv g b) (inv g a).
Proof. intros. symmetry.
  assert (inc (inv g a)  (Ul g)). mp.
  assert (inc (inv g b)  (Ul g)). mp.
  assert (inc (dmult g (inv g b) (inv g a)) (Ul g)).  mp.
  app fleft_inv_unique. mp.
  rewrite fmult_assoc; try am.  rww fa_inv_a_b_is_b.
  app finv_right. 
Qed.

Definition lzero_divisor (g a:Bset):= (inc a (Ul g)) & 
  (exists b, (inc b (Ul g)) & b<> (zero g) & dmult g a b = zero g).
Definition rzero_divisor (g a:Bset):= (inc a (Ul g)) & 
  (exists b, (inc b (Ul g)) &  b<> (zero g) & dmult g a b = zero g).

Definition lring_domain g:= 
  forall a:Bset, inc a (Ul g) -> a<> (zero g) -> ~ (lzero_divisor g a).
Definition rring_domain g:= 
  forall a:Bset, inc a (Ul g) -> a<> (zero g) -> ~ (rzero_divisor g a).
Definition ring_domain g := (ring_axioms g)  & (lring_domain g)
  & (lring_domain g).


Lemma sfield_integral_domain_bis: forall g, sfield_axioms g-> ring_domain g.
Proof. ir. red. split. mp. split. red. ir. uf not. ir. red in H2. nin H2.
  nin H3.  ee. pose (sfield_integral_domain H H0 H3 H5). nin o.
  apply (H1 H6). apply (H4 H6).
  red. ir. uf not. ir. red in H2. nin H2.
  nin H3.  ee. pose (sfield_integral_domain H H0 H3 H5). nin o.
  apply (H1 H6). apply (H4 H6).
Qed.

Lemma rtimes_mone_is_neg_left: forall g x, ring_axioms g -> inc x (Ul g)->
  dmult g (neg g (one g))  x = neg g x.
Proof. ir. assert (inc (one g) (Ul g)). mp.
  apply gright_simpl with g x; try am. mp. mp. mp.
  rww ropp_left. set (u:=dmult g (neg g (one g)) x).
  assert (dmult g (one g) x =x). app rone_unit_left. wr H2. uf u.
  wrr rdistributive_right. rww ropp_left. app rzero_left_abs. mp.
Qed.

Lemma rtimes_mone_is_neg_right: forall g x, ring_axioms g -> inc x (Ul g)->
  dmult g x (neg g (one g))  = neg g x.
Proof. ir. assert (inc (one g) (Ul g)). mp.
  apply gright_simpl with g x; try am. mp. mp. mp.
  rww ropp_left. set (u:=dmult g x (neg g (one g))).
  assert (dmult g x (one g)  =x). app rone_unit_right. wr H2. uf u.
  wrr rdistributive_left. rww ropp_left. app rzero_right_abs. mp.
Qed.


Lemma lmtimes_mone_is_neg: forall g x, lmodule_axioms g -> inc x (Ul g)->
  lmult g (neg (lauxiliary g) (one (lauxiliary g)))  x = neg g x.
Proof. ir. set (D:= lauxiliary g). set (o:= (one D)).
  assert (ring_axioms D). red in H. ee. am.
  assert (inc (one D) (Ul D)). mp.
  apply gright_simpl with g x; try am. mp. mp. mp.
  rww lmopp_left. set (u:=lmult g (neg D o) x).
  assert (lmult g (one D) x =x). unfold D. app lmone_unit_left. wr H3. uf u.
  wrr lmdistributive_right. rww ropp_left. unfold D. app lmzero_left_abs. mp.
Qed.

Lemma rmtimes_mone_is_neg: forall g x, rmodule_axioms g -> inc x (Ul g)->
  rmult g (neg (rauxiliary g) (one (rauxiliary g)))  x = neg g x.
Proof. ir. set (D:= rauxiliary g). set (o:= (one D)).
  assert (ring_axioms D).  unfold D; mp.
  assert (inc (one D) (Ul D)). mp.
  apply gright_simpl with g x; try am. mp. mp. mp.
  rww rmopp_left. set (u:=rmult g (neg D o) x).
  assert (rmult g (one D) x =x). unfold D. app rmone_unit_left. wr H3. uf u.
  wrr rmdistributive_right. rww ropp_left. unfold D. app rmzero_left_abs. mp.
Qed.


 

 

 

 
End Basic1.

Export Basic1.

Module CommutativeGroup.

Definition gconjugate (g x y:Bset) := (dplus g x (dplus g y (neg g x))).
Definition gconjugate' (g x y:Bset) := (dplus g (dplus g x y) (neg g x)).

Lemma conj_assoc: forall g x y, group_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> gconjugate g x y = gconjugate' g x y.
Proof. intros.  unfold gconjugate.  unfold gconjugate'. rww gplus_assoc. mp.
Qed.

Lemma cg_conjugation_trivial: forall g x y, cgroup_axioms g
  ->inc x (Ul g) -> inc y (Ul g) -> gconjugate g x y = y.
Proof.  intros. unfold gconjugate.
  assert (group_axioms g). app cgroup_is_group.
  assert  ((dplus g y (neg g x)) =  (dplus g (neg g x) y)).
  app cgplus_comm.  mp. rw H3.
  wrr gplus_assoc. rww gopp_right. app gzero_unit_left. mp.
Qed.

End CommutativeGroup.

(* dans un anneau -1.x = -x dans un module idem 

*)

(*EOD *)


