Require Export jfunc.
Require Export set2.
Require Export set3.

(* $Id: set31.v,v 1.17 2009/07/20 06:33:18 grimm Exp $ *)
Set Implicit  Arguments.

Module Bproduct.

(** ** EII-5-3 Definition of the product of a family of sets *)

(** We may need the graph of a function of type [a->E] *)
Definition gbcreate a (f:a->Bset) := (IM (fun y:a => J (Ro y) (f y))).

Lemma inc_gbcreate: forall a (f:a->Bset) x,
  inc x (gbcreate f) = exists y:a, J (Ro y) (f y) =x.
Proof. ir. uf gbcreate. app iff_eq. ir. cp (IM_exists H). simpl in H0. am.
  ir. app IM_inc.
Qed.

Lemma graph_gbcreate: forall  a (f:a-> Bset), is_graph (gbcreate f).
Proof. ir. red. ir. rwi inc_gbcreate H. nin H. wr H. app pair_is_pair.
Qed.

Lemma fgraph_gbcreate: forall  a (f:a-> Bset), fgraph (gbcreate f).
Proof. ir. red. ee. app graph_gbcreate. ir. rwi inc_gbcreate H.  
  rwi inc_gbcreate H0. nin H. nin H0. wri H H1; wri H0 H1.  awi H1. 
  cp (R_inj H1). wr H0; wr H; rww H2.
Qed.

Lemma domain_gbcreate : forall a (f:a-> Bset),
  domain (gbcreate f) = a.
Proof. ir. set_extens. awi H. nin H. rwi inc_gbcreate H. nin H.
  cp (pr1_injective H). wr H0. app R_inc. app graph_gbcreate. aw.
  set (y:= Bo H). exists (f y). rw inc_gbcreate. exists y. app uneq2.
  uf y. app B_eq.  app graph_gbcreate. 
Qed.

Lemma V_gbcreate : forall a (f:a-> Bset) (x:a),
  V (Ro x) (gbcreate f) = f x.
Proof. ir. assert  (inc (J (Ro x) (f x)) (gbcreate f)). rw inc_gbcreate.
  exists x. tv. cp (fgraph_gbcreate f). cp (pr2_V H0 H). sy. awi H1. am.
Qed.


(** We have two equivalent defintions *) 
Definition productb f:=
  Zo (powerset (product (domain f) (unionb f)))
  (fun z => fgraph z & domain z = domain f 
    & forall i, inc i (domain f) -> inc (V i z) (V i f)).

Definition productt In (f:In->Bset):= productb (gbcreate f).
Definition productf sf f:=  productb (L sf f).

Lemma productb_pr: forall f x,
  fgraph f ->
  inc x (productb f) = 
  (fgraph x &  domain x = domain f & 
    forall i, inc i (domain x) -> inc (V i x) (V i f)).
Proof. ir. uf productb. ap iff_eq. ir. Ztac. am. am. rw H3. am. 
  ir. ee. Ztac. app powerset_inc.  red. ir.  app product_inc. red in H0. ee.
  app H0. cp (in_graph_V H0 H3). cp (inc_pr1_domain H0 H3). wrr H1.
  assert (inc (P x0) (domain f)). wr H1. aw. exists (Q x0). rw pair_recov. am.  
  red in H0; ee;  app H0. red in H0; ee; am.
  apply unionb_inc with (P x0). am. rw (pr2_V H0 H3). app H2. rww H1. ee.
  am. am. wr H1. am.
Qed.

Lemma productt_pr: forall In (f:In->Bset) x,
  inc x (productt f) = 
  (fgraph x &  domain x = In  & forall i:In, inc (V (Ro i) x) (f i)).
Proof. ir. uf productt. rw productb_pr. ap iff_eq. ir. ee. am. 
  rwi domain_gbcreate H0. am. rwi domain_gbcreate H0.  rwi H0 H1. ir.
  assert (inc (Ro i) In).  app R_inc. cp (H1 _ H2). rwi V_gbcreate H3. am.
  ir. ee. am.  rw domain_gbcreate. am. ir. rwi H0 H2.  set (y:= Bo H2). 
  assert (Ro y =i). uf y. app B_eq. wr H3. rw V_gbcreate. app H1. 
  app fgraph_gbcreate.
Qed.
 
Lemma productf_pr: forall sf f x,
  inc x (productf sf f) = 
  (fgraph x &  domain x = sf & 
    forall i, inc i (domain x) -> inc (V i x) (f i)).
Proof. ir. uf productf. rw productb_pr. aw. app iff_eq.
  ir. ee. am. am. ir. cp (H1 _ H2). rwi create_V_rewrite H3. am. wrr H0. ir.
  ee. am. am. ir. aw. app H1. wrr H0. app create_axioms.
Qed.

Lemma productt_extensionality:forall In (f:In->Bset) x x',
  inc x (productt f) -> inc x' (productt f) ->
  (x = x') = (forall i:In, V (Ro i) x = V (Ro i) x').
Proof. ir. app iff_eq. ir. rww H1. ir. rwi productt_pr H0. rwi productt_pr H.
  ee. ap function_extensionality. am. am. rww H2. rw H4.  ir. 
  assert (x0 = Ro (Bo H6)). rw B_eq.  tv. rw H7. app H1.
Qed.


Lemma productb_extensionality:forall f x x',
  fgraph f ->
  inc x (productb f) -> inc x' (productb f) ->
  (x = x') = (forall i, inc i (domain f) -> V i x = V i x').
Proof. ir. app iff_eq. ir. rww H2. ir. rwi productb_pr H0. rwi productb_pr H1.
  ee. ap function_extensionality. am. am. rww H3.  rw H5. am. am. am.
Qed.

Lemma productf_extensionality:forall sf f x x',
  inc x (productf sf f) -> inc x' (productf sf f) ->
  (x = x') = (forall i, inc i sf -> V i x = V i x').
Proof. ir.  ufi productf H.  ufi productf H0.
  app iff_eq. ir. rww H1. ir. rwi productb_pr H. rwi productb_pr H0.
  ee. ap function_extensionality. am. am. rww H2. rw H4. aw. aw.
  app create_axioms.   app create_axioms.
Qed.

(** Projection function to component i *)

Definition pr_i f i:=  BL(V i) (productb f) (V i f).
Definition pr_it In (f:In ->Bset):=  pr_i (gbcreate f).

Lemma source_pri: forall f i,
  source(pr_i f i) = productb f.
Proof. ir. uf pr_i. rw af_source. tv. 
Qed.

Lemma target_pri: forall f i,
  target(pr_i f i) = V i f.
Proof. ir. uf pr_i. rw af_target. tv. Qed.

Lemma axioms_pri: forall f i,
  fgraph f -> inc i (domain f) ->
  transf_axioms (V i)(productb f)(V i f).
Proof. ir. uf pr_i. red. ir. rwi productb_pr H1. ee. app H3. rww H2. am.
Qed.

Lemma function_pri: forall f i,
  fgraph f -> inc i (domain f) ->
  is_function (pr_i f i). 
Proof. ir. uf pr_i. app af_function. app axioms_pri.
Qed.

Lemma W_pri: forall f i x,
  fgraph f -> inc i (domain f) -> inc x (productb f) -> 
  W x (pr_i f i) = V i x. 
Proof. ir. uf pr_i. rw W_af_function. tv. app axioms_pri. am.
Qed.

Lemma source_prit: forall In (f:In->Bset) i,
  source(pr_it f i) = productt f.
Proof. ir. uf pr_it.  rw source_pri. tv. Qed.

Lemma target_prit: forall In (f:In->Bset) i,
  target(pr_it f (Ro i)) = f i.
Proof. ir. uf pr_it. rw target_pri. rw V_gbcreate.  tv. Qed.

Lemma axioms_prit: forall In (f:In->Bset) i,
  transf_axioms (V (Ro i))(productt f)(f i).
Proof. ir. uf pr_it. red. ir. rwi productt_pr H. ee. app H1. 
Qed.

Lemma function_prit: forall In (f:In-> Bset) (i:In),
  is_function (pr_it f (Ro i)). 
Proof. ir. uf pr_it. app function_pri. app fgraph_gbcreate. rw domain_gbcreate.
  app R_inc.
Qed.

Lemma W_prit: forall In (f:In ->Bset) (i:In) x,
  inc x (productt f) -> W x (pr_it f (Ro i)) = V (Ro i) x. 
Proof. ir. uf pr_it. rw W_pri. tv. app fgraph_gbcreate. rw domain_gbcreate.
  app R_inc. am.
Qed.

Lemma trivial_product :
  productb emptyset = singleton emptyset.
Proof. set_extens. rwi productb_pr H. ee. cp (emptyset_dichot x). nin H2. 
  rw H2. app singleton_inc. nin H2. cp (inc_pr1_domain H H2). rwi H0 H3.
  rwi domain_emptyset H3. nin H3. elim x0.  app fgraph_emptyset.
  cp (singleton_eq H). rw H0. rw productb_pr. ee. app fgraph_emptyset.
  tv. rw domain_emptyset. ir. nin H1. elim x0. app fgraph_emptyset.
Qed.

Lemma nonempty_product: forall f, 
  fgraph f ->  (forall i, inc i (domain f) -> nonempty (V i f)) ->
  nonempty (productb f).
Proof. ir. exists (L(domain f)(fun i=> rep (V i f))).
  rw productb_pr. ee. app create_axioms. aw. 
  aw. ir. rw create_V_rewrite.  app nonempty_rep. app H0. am. am.
Qed.

Lemma nonempty_product2: forall f, 
 fgraph f ->  nonempty (productb f) ->
  (forall i, inc i (domain f) -> nonempty (V i f)).
Proof. ir. nin H0. rwi productb_pr H0. ee. exists (V i y). app H3. rww H2. am.
Qed.

Lemma nonempty_productt: forall In (f:In->Bset), 
  (forall i, nonempty (f i)) ->  nonempty (productt f).
Proof. ir. uf productt. app nonempty_product. app fgraph_gbcreate. 
  rw domain_gbcreate. ir. assert (i= (Ro (Bo H0))). rww B_eq. rw H1. 
  rw V_gbcreate. app H.
Qed.

Lemma nonempty_productt2: forall In (f:In->Bset), 
  nonempty (productt f) -> (forall i, nonempty (f i)).
Proof. ir. nin H. rwi productt_pr H. ee. exists (V (Ro i) y). app H1. 
Qed.

Lemma set_of_gfunctions_pr1: forall a b z,
  inc z (set_of_gfunctions a b) =
  (fgraph z & domain z = a & sub z (product a b)).
Proof. ir. uf set_of_gfunctions. app iff_eq. ir. Ztac.  am. sy. am.
  app powerset_sub. ir. ee. Ztac. app powerset_inc.
Qed.


Lemma set_of_gfunctions_pr2: forall a b z,
  inc z (set_of_gfunctions a b) =
  (fgraph z & domain z = a & sub (range z) b).
Proof. ir. rw  set_of_gfunctions_pr1. app iff_eq. ir. ee. am. am.
  red. ir. awi H2. nin H2. cp (H1 _ H2).  rwi inc_product H3. ee. awi H5. am.
  red in H. ee. am. ir. ee. am. am. red. ir. assert (is_pair x).
  red in H; ee; app H. app product_inc.  wr H0. aw. exists (Q x). 
  rw pair_recov. am. am. red in H; ee;am. app H1. aw. exists (P x).
  rw pair_recov. am. am. red in H; ee;am. 
Qed.

Lemma sub_product_graphset:  forall f x,
  fgraph f -> sub (unionb f) x ->
  sub (productb f) (set_of_gfunctions (domain f) x).
Proof. ir. red. ir. rwii productb_pr H1. ee.  rw set_of_gfunctions_pr2. ee.
  am. am. red. ir. rwi frange_inc_rw H4.  nin H4. ee. rw H5. app H0. 
  apply unionb_inc with x2. wrr H2. app H3. am.
Qed.

Lemma eq_product_graphset:  forall f x,
  fgraph f -> (forall i, inc i(domain f) -> V i f = x) ->
  productb f = set_of_gfunctions (domain f) x.
Proof. ir. app extensionality. app sub_product_graphset. red. ir.
  cp (unionb_exists H1).  nin H2. ee. rwi H0 H3. am. am.
  red. ir. rwi set_of_gfunctions_pr2 H1. rw productb_pr. ee. am. am. ir.
  rw H0. app H3. app inc_V_range. wrr H2. am.
Qed.

Definition product1 (x a:Bset) := productt  (fun _:singleton a => x).

Lemma product1_pr: forall x a,
  product1 x a= set_of_gfunctions (singleton a) x.
Proof. ir. uf product1. uf productt.
  set (g:=  gbcreate (fun _ : singleton a => x)).
  assert (fgraph g). uf g. app fgraph_gbcreate.
  rw (eq_product_graphset H (x:=x)). uf g. rw domain_gbcreate. tv.
  ir. ufi g H0. rwi domain_gbcreate H0. uf g. set (u:=Bo H0). 
  assert (Ro u = i). uf u. app B_eq. wr H1.  rw V_gbcreate. tv.
Qed.

Lemma inc_product1: forall x a y, 
  inc y (product1 x a) = (fgraph y & domain y = singleton a
    & inc (V a y) x).
Proof. ir. rw product1_pr. rw set_of_gfunctions_pr2. 
  app iff_eq. ir. ee. am. am. app H1. app inc_V_range. rw H0. app singleton_inc.
  ir. ee. am. am. red. ir. rwi frange_inc_rw H2.  nin H2. ee. rwi H0 H2.
  cp (singleton_eq H2). rwi H4 H3. wri H3 H1. am. am.
Qed.

Definition product1_canon (x a:Bset) :=
  BL (fun i : Bset => L(singleton a) (fun _ : Bset => i))
  x (product1 x a).

Lemma transf_axioms_product1_canon: forall x a,
 transf_axioms (fun i =>  L(singleton a) (fun _ :Bset  => i)) x (product1 x a).
Proof. ir. red. ir. rw inc_product1. ee. app create_axioms. aw. aw. fprops.
Qed.

Lemma function_product1_canon: forall x a,
  is_function (product1_canon x a).
Proof. ir. uf product1_canon. app af_function. app transf_axioms_product1_canon.
Qed.

Lemma source_product1_canon: forall x a,
  source (product1_canon x a) = x.
Proof. ir. uf product1_canon. app af_source. 
Qed.

Lemma target_product1_canon: forall x a,
  target (product1_canon x a) = (product1 x a).
Proof. ir. uf product1_canon. app af_target. 
Qed.

Lemma W_product1_canon: forall x a i,
  inc i x -> W i (product1_canon x a) = 
  L(singleton a) (fun _ : Bset => i).
Proof. ir. uf product1_canon. rw W_af_function. tv. 
  app transf_axioms_product1_canon. am.
Qed.

Lemma bijective_product1_canon: forall x a,
  bijective(product1_canon x a).
Proof. ir. assert (is_function (product1_canon x a)).
  app function_product1_canon. red. ee. red. ee. am. ir. 
  rwi source_product1_canon H0.  rwi source_product1_canon H1.
  set (u := W x0 (product1_canon x a)). set (v := W y (product1_canon x a)).
  assert (V a u = V a v). uf u; uf v. rww H2. ufi u H3; ufi v H3.
  rwi W_product1_canon H3.  awi H3. rwi W_product1_canon H3. awi H3. am.
  fprops. fprops. fprops. am.
  app  surjective_pr5. ir. rwi target_product1_canon H0. 
  rw source_product1_canon. rwi inc_product1 H0. ee.
  assert (y = L (singleton a) (fun _ : Bset => (V a y))).
  ap function_extensionality. am. app create_axioms. aw. aw. 
  ir.  aw. rwi H1 H3. cp (singleton_eq H3). 
  rww H4. wrr H1.  exists (V a y). ee. am. red. 
  assert (y = W (V a y) (product1_canon x a)). rw W_product1_canon. am. am.
  set (z:= V a y). fold z in H4. rw H4. app defined_lem. 
Qed.


(** Variant of a product with two elements *) 
Definition product2 x y a b :=
  productf(doubleton a b) (variant a x y).

Definition product2_canon x y a b :=
  BL (fun z => (Lvariant a b (P z) (Q z)))
  (product x y) (product2 x y a b).

Lemma inc_product2: forall x y z a b,
  neq b a -> 
  inc z (product2 x y a b) = (fgraph z & domain z = doubleton a b &
    inc (V a z) x & 
    inc (V b z) y).
Proof. ir. uf product2. rw productf_pr.
  app iff_eq. ir. ee. am. am. 
  assert (inc a (domain z)). rw H1. fprops.
  cp (H2 _ H3). rwi variant_if_rw H4. am. tv. 
  assert (inc b (domain z)). rw H1. fprops.
  cp (H2 _ H3). rwi variant_if_not_rw H4. am. am. 
  ir. ee. am. am. ir. rwi H1 H4. cp (doubleton_or H4). nin H5.
  rw variant_if_rw. rww H5. am. rw H5. rw variant_if_not_rw. am. am.
Qed.

Lemma transf_axioms_product2_canon: forall x y a b,
  neq b a ->
  transf_axioms (fun z  => Lvariant a b (P z) (Q z))
  (product x y) (product2 x y a b).
Proof. red. ir. rw inc_product2. ee. aw. aw. tv. app axioms_variant.
  rw domain_variant.  tv. rw V_variant_a. rwi inc_product H0. ee. am. 
  rw V_variant_b. rwi inc_product H0; ee; am. am. am.
Qed.

Lemma function_product2_canon: forall x y a b,
  neq b a -> is_function (product2_canon x y a b).
Proof. ir. uf product2_canon. app af_function. 
  app transf_axioms_product2_canon.
Qed.

Lemma source_product2_canon: forall x y a b,
  source (product2_canon x y a b) = (product x y).
Proof. ir. uf product2_canon. app af_source. 
Qed.

Lemma target_product2_canon: forall x y a b,
  target (product2_canon x y a b) = (product2 x y a b).
Proof. ir. uf product2_canon. app af_target. 
Qed.

Lemma W_product2_canon: forall x y z a b,
  neq b a -> inc z (product x y) -> W z (product2_canon x y a b) = 
  Lvariant a b (P z) (Q z).
Proof. ir. uf product2_canon. rw W_af_function. tv. 
  app transf_axioms_product2_canon. am.
Qed.

Lemma bijective_product2_canon: forall x y a b,
  neq b a -> bijective(product2_canon x y a b).
Proof. ir. assert (is_function (product2_canon x y a b)).
  app function_product2_canon. red. ee. red. ee. am. ir. 
  rwi source_product2_canon H1. rwi source_product2_canon H2.
  set (u := W x0 (product2_canon x y a b)). 
  set (v := W y0 (product2_canon x y a b)).
  assert (P x0 = P y0).
  assert (V a u = V a v). uf u; uf v. rww H3. ufi u H4; ufi v H4.
  rwi W_product2_canon H4. rwi W_product2_canon H4. 
  rwi V_variant_a H4. rwi V_variant_a H4. am. am. am. am.  am.  
  assert (Q x0 = Q y0).
  assert (V b u =  V b v). uf u; uf v. rww H3. ufi u H5; ufi v H5.
  rwi W_product2_canon H5. rwi W_product2_canon H5. 
  rwi V_variant_b H5. rwi V_variant_b H5. am. am. am. am. am. am. am. 
  app pair_extensionality. 
  rwi inc_product H1; ee; am.  rwi inc_product H2; ee; am.
  app surjective_pr5. ir. rwi target_product2_canon H1. 
  rw source_product2_canon. rwi inc_product2 H1. ee.
  set (xa:= V a y0).  fold xa in H2.
  set (ya:= V b y0). fold ya in H3.
  set (z := J xa ya). assert (Ha:xa = P z). uf z. rw pr1_pair. tv. 
  assert (Hb:ya = Q z). uf z. rw pr2_pair. tv. 
  assert(Hc: inc z (product x y)). rw inc_product. ee. uf z. app pair_is_pair.
  wrr Ha. wrr Hb. 
  assert (y0 = Lvariant a b xa ya). 
  ap function_extensionality. am. uf Lvariant. aw.  uf Lvariant. 
  app create_axioms. rw domain_variant. am.
  ir. rwi H2 H5. cp H5. cp (doubleton_or H5).
  nin H7.  rw H7.  rww V_variant_a. rw H7. rww V_variant_b. 
  exists z. ee. am. red. assert (y0 = W z (product2_canon x y a b)). 
  rw W_product2_canon. wr Ha; wr Hb. 
  am. am. am.  rw H6. app defined_lem. am. 
Qed.


Definition is_singleton x := exists u, x = singleton u.

Lemma is_singleton_pr: forall x,
  is_singleton x = (nonempty x & (forall a b, inc a x -> inc b x -> a = b)).
Proof. ir. app iff_eq. ir. nin H. ee. exists x0. rw H. fprops.
  ir. rwi H H0. rwi H H1. assert (a = x0). app singleton_eq.
  assert (b = x0). app singleton_eq. rw H2; rw H3. tv.
  ir. ee. nin H. exists y. set_extens. assert (y = x0). app H0. wr H2. 
  fprops. assert (x0=y). app singleton_eq. rw H2. am.
Qed.

Lemma singleton_product: forall f,
  fgraph f -> (forall i, inc i (domain f) -> is_singleton (V i f))
  -> is_singleton (productb f).
Proof. ir. rw is_singleton_pr. ee. app nonempty_product. ir. cp (H0 _ H1).
  rwi is_singleton_pr H2. ee. am. ir. rw (productb_extensionality H H1 H2).
  ir. cp (H0 _ H3). rwi is_singleton_pr H4. 
  rwi productb_pr H1. rwi productb_pr H2. ee. rwi H8 H9; rwi H6 H7.
  app (H5 _ _ (H9 _ H3) (H7 _ H3)). am. am.
Qed.

Definition constant_graph s x :=
  L s (fun _ => x). 
Definition is_constant_graph f :=
  fgraph f &  
  (forall x x', inc x (domain f) -> inc x' (domain f) -> V x f = V x' f).

Lemma constant_graph_function: forall f,
  is_constant_function f -> is_constant_graph(graph f).
Proof. ir. red in H. red. ee. app fgraph_function. ufi W H0.
  red in H. ee. wr H2. am.
Qed.

Lemma V_constant_graph: forall s x y,
  inc y s -> V y (constant_graph s x) = x.
Proof. ir. uf constant_graph. aw.
Qed.

Lemma constant_graph_small_range: forall f,
  is_constant_graph f -> small_set(range f).
Proof. ir. red. red in H. ir. ee. rwi frange_inc_rw H0. rwi frange_inc_rw H1.
  nin H0; nin H1. ee. cp (H2 _ _ H0 H1). rw H3; rw H4. tv. am. am.
Qed.

Definition diagonal_graphp e i :=
  Zo(set_of_gfunctions i e) is_constant_graph.

Lemma diagonal_graph_pr: forall e i x,
  inc x (diagonal_graphp e i) = 
  (is_constant_graph x & domain x = i & sub (range x) e).
Proof. ir. uf diagonal_graphp. app iff_eq. ir. Ztac. am.
  rwi set_of_gfunctions_pr2 H0. ee. am. rwi set_of_gfunctions_pr2 H0. ee. am. 
  ir. ee. Ztac. rw set_of_gfunctions_pr2. red in H. intuition.
Qed.

Lemma constant_graph_is_constant: forall x y,
  is_constant_graph(constant_graph x y).
Proof. ir. red. uf constant_graph. aw. ee. app create_axioms. ir. aw.
Qed.

Definition constant_functor i e:=
  BL(fun x => constant_graph i x) e (set_of_gfunctions i e).

Lemma injective_cf : forall i e,
  nonempty i -> injective (constant_functor i e).
Proof. ir. 
  assert(transf_axioms (fun x => constant_graph i x) e 
    (set_of_gfunctions i e)).
  red. ir. rw set_of_gfunctions_pr2. uf constant_graph. aw. ee. 
  app create_axioms. tv. 
  red. ir. rwi inc_create_range H1. nin H1. ee. wrr H2.
  assert (is_function (constant_functor i e)). 
  uf constant_functor. app af_function. red. ee. am. uf constant_functor. 
  rw af_source. ir. awi H4. 
  nin H. assert (V y0 (constant_graph i x) = x). app V_constant_graph.
  wr H5. rw H4. app V_constant_graph. am. am. am. am.
Qed.

Definition product_compose f u :=
  BL (fun x => compose_graph x (graph u))
  (productb f) (productf (source u) (fun k => V (W k u) f)).

Lemma source_pc: forall f u, source (product_compose f u) = (productb f).
Proof. ir. uf product_compose. rw af_source. tv. Qed.

Lemma target_pc: forall f u, 
  target (product_compose f u) =
  (productf (source u) (fun k => V (W k u) f)).
Proof. ir. uf product_compose. app af_target. 
Qed.

Lemma compose_V: forall u v x,
  fgraph u -> fgraph v -> fgraph (compose_graph u v) -> 
  inc x (domain (compose_graph u v)) ->
  V x (compose_graph u v) = V (V x v) u.
Proof. ir.  cp (fdefined_lem  H1 H2).
  set (z:=  V x (compose_graph u v)). fold z in H3.
  awi H3. ee. nin H4. ee. cp (pr2_V H0 H4). awi H6.  wr H6.
  cp (pr2_V H H5). awi H7. am. red in H0; ee;am. red in H; ee;am.
Qed.

Lemma axioms_pc1: forall f u c,
  fgraph f -> bijective u -> target u = domain f ->
  inc c (productb f)  -> domain (compose_graph c (graph u)) = source u.
Proof. ir. rwi productb_pr H2. rw compose_domain.  ee.
  rw H3. wr H1.  uf inv_image_by_graph. set (v:= (inverse_fun u)).
  cp (inverse_bij_is_bij1 H0). fold v in H5.
  wr (source_inverse u). wr (target_inverse u). wr (graph_inverse u). fold v.
  red in H5. ee. red in H6. ee. ufi image_of_fun H7. am.  
  red in H0; ee. red in H0; ee. fprops.  ee. red in H2; ee; am.  am.
Qed.

Lemma axioms_pc2: forall f u c,
  fgraph f -> bijective u -> target u = domain f ->
  inc c (productb f)  -> fgraph (compose_graph c (graph u)).
Proof. ir.
  assert (Hb:is_graph c). rwi productb_pr H2.  ee. red in H2; ee; am. am.
  assert (Ha:is_graph (graph u)). 
  red in H0; ee. red in H0; ee. fprops.  
  red. ee. fprops. ir. awi H3.
  awi H4. nin H3; nin H4. nin H6; nin H7. ee.
  red in H0; ee; red in H0; ee. red in H0; ee.  red in H12. ufi related H12.
  rwi H5 H6. cp (H12 _ _ _ H6 H7). rwi productb_pr H2.  ee. red in H2; ee.
  assert (P (J x0 (Q x)) = P (J x1 (Q y))). rw pr1_pair. rw pr1_pair. am.
  cp (H17 _ _ H9 H8 H18). cp (pr2_injective H19).
  wr (pair_recov H3). wr (pair_recov H4). rw H5. rw H20. tv. am. am. am. am.
  am.
Qed.

Lemma axioms_pc: forall f u,
  fgraph f -> bijective u -> target u = domain f ->
  transf_axioms (fun x => compose_graph x (graph u)) 
     (productb f) (productf (source u) (fun k => V (W k u) f)).
Proof. ir. red. ir.
  cp (axioms_pc1 H H0 H1 H2).  cp (axioms_pc2 H H0 H1 H2).
  rwi productb_pr H2. ee.
  assert (Ha:is_graph (graph u)).  red in H0; ee. red in H0; ee. fprops. 
  assert (Hb:is_graph c). red in H2; ee; am. rw productf_pr.  ee. am. am.
  ir. rw compose_V. uf W. app H6. rw H5. wr H1.
  assert (V i (graph u) = W i u). uf W. tv. rw H8. app inc_W_target. 
  red in H0; ee; red in H0; ee; am. wrr H3. am. red in H0; ee; red in H0; ee.
  app fgraph_function. am. am. am.
Qed.

Lemma function_pc: forall f u,
  fgraph f -> bijective u -> target u = domain f ->
  is_function(product_compose f u).
Proof. ir. uf product_compose. app af_function. app axioms_pc.
Qed.

Lemma W_pc: forall f u x,
  fgraph f -> bijective u -> target u = domain f ->
  inc x (productb f) -> W x (product_compose f u) = compose_graph x (graph u).
Proof. ir. uf product_compose. aw. aw. tv. app axioms_pc.
Qed.

Lemma WV_pc: forall f u x i,
  fgraph f -> bijective u -> target u = domain f ->
  inc x (productb f) -> inc i (source u) ->
   V i (W x (product_compose f u)) = V  (W i u) x.
Proof. ir. rww W_pc. uf W. rw compose_V. tv. rwi productb_pr H2. ee. am. am.
  red in H0; ee. red in H0; ee. app fgraph_function.
  cp (axioms_pc2 H H0 H1 H2). am. cp (axioms_pc1 H H0 H1 H2). rw H4. am.
Qed.

Lemma bijective_pc: forall f u,
  fgraph f -> bijective u -> target u = domain f ->
  bijective (product_compose f u).
Proof. ir. assert (is_function (product_compose f u)). app function_pc.
  red. ee. red. ee. am. rw source_pc. ir. 
  rw (productb_extensionality H H3 H4). ir. wri H1 H6. cp H0. red in H0; ee.
  cp (surjective_pr2 H8 H6). nin H9. ee. rw H10.
  wr (WV_pc H H7 H1 H3 H9).  wr (WV_pc H H7 H1 H4 H9). rww H5. 
  app surjective_pr6.  rw source_pc; rw target_pc. ir.
  set (x:= L(domain f) (fun i => V (W i (inverse_fun u)) y)).
  assert (inc x (productb f)).  rw productb_pr.  uf x. aw. ee. 
  app create_axioms. tv. ir. rww create_V_rewrite. wri H1 H4. 
  cp H0.  red in H5; ee. cp (surjective_pr2 H6 H4). nin H7. ee. 
  cp (W_inverse2 H0 H7 H8). wr H9. rwi productf_pr H3. ee. rw H8. app H11.
  rw H10. am. am. exists x. ee. am.
  assert (inc (W x (product_compose f u))
    (productf (source u) (fun k  => V (W k u) f))).
  wr (target_pc). app inc_W_target. 
  set(ff:= L  (source u) (fun k  => V (W k u) f)).
  assert (fgraph ff). uf ff.  app create_axioms. 
  assert ( (productf (source u) (fun k => V (W k u) f)) = productb ff).
  uf ff. uf productf. tv. rwi H7 H3; rwi H7 H5.
  rw (productb_extensionality H6 H3 H5). ir.  ufi ff H8. rwi create_domain H8.
  rww WV_pc. uf x. 
  assert  (inc (W i u) (domain f)).  wr H1. app inc_W_target. 
  red in H0; ee; red in H0; ee; am. rww create_V_rewrite. 
  set (z := W i u). assert (z = W i u). tv. cp (W_inverse2 H0 H8 H10). wrr H11.
Qed.

(** ** EII-5-4 Partial products *)


Lemma graph_extensionality: forall r r',
  is_graph r -> is_graph r' ->
  (r = r') =  forall u v, related r u v = related r' u v.
Proof. ir. app iff_eq. ir. rw H1. tv. ir. ufi related H1.
  set_extens. assert (J (P x) (Q x) =x). app pair_recov. app H.
  wr H3. wr H1. rw H3. am. assert (J (P x) (Q x) =x). app pair_recov. app H0.
  wr H3. rw H1. rw H3. am.
Qed.

Lemma restriction_graph1 : forall x f,
  fgraph f -> sub x (domain f) ->
   L x (fun i  => V i f) = restr f x.
Proof. ir. set (g:= (L x (fun x  => V x f))). assert (fgraph g).
  uf g. app create_axioms. rw graph_extensionality. uf related.  ir.
  aw. ap iff_eq. ir.
  cp (inc_pr1_domain H1 H2). ufi g H3. rwi create_domain H3.
  cp (pr2_V H1 H2). awi H4.  rw H4. awi H3. uf g. aw. rw restr_inc_rw. ee.
  app fdefined_lem. app H0. aw. 
  ir. rwi restr_inc_rw H2.  nin H2. awi H3.
  cp (pr2_V H H2).  awi H4. assert (v = V u g). uf g. aw. rw H5.
  app fdefined_lem. uf g. rww inc_create_domain. red in H1; ee; am.
  cp (restr_axioms x H). red in H2; ee;am.
Qed.

Lemma restriction_V: forall f x i,
  fgraph f -> sub x (domain f) -> inc i x ->
  V i (restr f x) = V i f.
Proof. ir. app restr_ev.
Qed.

Lemma restriction_graph2 : forall f j,
  fgraph f -> sub j (domain f) ->
  L j (fun x  => V x f) = compose_graph f (diagonal j).
Proof. ir. rw restriction_graph1.
  rw graph_extensionality. ir. rw compose_related.
  uf related. aw. app iff_eq. ir. ee. exists u. rwi restr_inc_rw H1.
  ee. change(related (diagonal j) u u). 
  rw diagonal_related. ee. awi H2. am. tv.  am. ir. nin H1. nin  H1. 
  rwi inc_diagonal H1.  awi H1. rw restr_inc_rw. ee. rw H4. am. aw.
  fprops. red in H; ee; am.  cp (restr_axioms j H). red in H1; ee;am.
  fprops. am. am.
Qed. 

Lemma restr_domain1 : forall f x,
  fgraph f -> sub x (domain f) -> domain (restr f x) =  x.
Proof. ir. rww restr_domain. set_extens. app (intersection2_second H1).
  app intersection2_inc. app H0.
Qed.

(** Product of f, with indices restricted to j *)
Definition restriction_product f j := productb (restr f j).

Definition pr_j f j :=
  BL (restriction_graph j) (productb f)(restriction_product f j).

Lemma axioms_prj: forall f j,
  fgraph f -> sub j (domain f) ->
  transf_axioms (restriction_graph j) 
  (productb f)(restriction_product f j).
Proof. ir. red. ir. rwii productb_pr H1. ee.  uf restriction_product.
  wrr restriction_graph1.  uf restriction_graph.  rw productb_pr. 
  aw. assert (domain (restr c j) = j). app restr_domain1. rww H2.
  ee. app restr_axioms. am. ir. aw. rw restr_ev. app H3. 
  rwi H4 H5. rw H2. app H0. am. rw H2. am.  wrr H4. wrr H4. app create_axioms.
Qed.

Lemma source_prj: forall f j,
  source (pr_j f j ) = (productb f).
Proof. ir. uf pr_j. rw af_source. tv.
Qed.

Lemma target_prj: forall f j,
  target (pr_j f j ) = restriction_product f j.
Proof. ir. uf pr_j. rw af_target. tv.
Qed.

Lemma function_prj: forall f j,
  fgraph f -> sub j (domain f) -> is_function (pr_j f j).
Proof. ir. uf pr_j.  app af_function. app axioms_prj.
Qed.

Lemma W_prj: forall f j x,
  fgraph f -> sub j (domain f) -> inc x (productb f)
  -> W x (pr_j f j) = (restriction_graph j x). 
Proof. ir. uf pr_j.  rw W_af_function.  tv. app axioms_prj. am.
Qed.

Lemma WV_prj: forall f j x i,
  fgraph f -> sub j (domain f) -> inc x (productb f) -> inc i j 
  -> V i (W x (pr_j f j))  =  V i x.
Proof. ir. rww W_prj. uf pr_j. rwi productb_pr H1. ee. 
  uf restriction_graph. rw restr_ev. tv. am. rww H3. am. am.
Qed.

Lemma prolongation_exists: forall f j g,
  fgraph f -> (forall i, inc i (domain f)  -> nonempty (V i f)) ->
  fgraph g -> domain g = j -> sub j (domain f) ->
  (forall i, inc i j -> inc (V i g) (V i f)) ->
  exists h, domain h = domain f &
    fgraph h & (forall i, inc i (domain f) -> inc (V i h) (V i f)) & 
    (forall i, inc i j -> V i h  = V i g).
Proof. ir. set (h:= union2 g (unionf (complement (domain f) j)
  (fun i => (singleton (pair i (rep (V i f))))))).
  assert (is_graph h).  red. ir. ufi h H5. cp (union2_or H5). nin H6. 
  red in H1. ee. app H1. cp (unionf_exists H6). nin H7. nin H7.
  cp (singleton_eq H8). rw H9. app pair_is_pair.
  assert (domain h = domain f). set_extens. awi H6. nin H6.
  ufi h H6.  cp (union2_or H6). nin H7. 
  assert (inc x (domain g)). aw. exists x0. am. red in H1; ee; am.
  app H3. wr H2. am. cp (unionf_exists H7). nin H8. nin H8.
  cp (singleton_eq H9). cp (pr1_injective H10). wri H11 H8.
  srwi H8; ee; am. am. aw. 
  apply by_cases with (inc x j). ir. wri H2 H7. awi H7.
  nin H7. exists x0. uf h. apply union2_first. am. red in H1; ee;am. ir.
  assert (inc x  (complement (domain f) j)). srw. intuition.
  exists (rep (V x f)). uf h. app union2_second. apply unionf_inc with x.
  am. fprops. assert (fgraph h).  red. ee. am. ir.
  ufi h H7.  cp (union2_or H7).  ufi h H8.  cp (union2_or H8). nin H10.
  nin H11. red in H1. ee. app (H12 _ _ H10 H11 H9). cp (unionf_exists H11).
  nin H12. ee. cp (singleton_eq H13).
  cp (inc_pr1_domain H1 H10). rwi H2 H15. rwi H9 H15. rwi H14 H15.
  rwi pr1_pair H15. srwi H12. ee.  elim H16. am.
  cp (unionf_exists H10). nin H12. ee.  cp (singleton_eq H13). nin H11.
  cp (inc_pr1_domain H1 H11). rwi H2 H15. wri H9 H15. rwi H14 H15.
  rwi pr1_pair H15. srwi H12. ee.  elim H16. am.
  cp (unionf_exists H11). nin H15. ee. cp (singleton_eq H16). 
  rwi H14 H9; rwi H17 H9. rwi pr1_pair H9; rwi pr1_pair H9. rwi H9 H14.
  rw H14; rw H17; tv.
  assert (forall i, inc i j -> V i h  = V i g). ir.
  assert (sub g h).  uf h. red. ir. app union2_first.
  wri H2 H8. sy. app (sub_ev H7 H9 H8). 
  exists h. ee. am. am. ir. apply by_cases with (inc i j). ir. rw H8. app H4.
  am. ir. assert (inc i (complement (domain f) j)). srw.
  intuition.  assert (inc (J i (rep (V i f))) h). uf h. app union2_second.
  apply unionf_inc with i. am. fprops. cp (pr2_V H7 H12).
  rwi pr1_pair H13; rwi pr2_pair H13;  wr H13. app nonempty_rep. app H0. am.
Qed.

Lemma surjective_prj: forall f j,
  fgraph f -> (forall i, inc i (domain f)  -> nonempty (V i f)) ->
  sub j (domain f) -> surjective (pr_j f j). 
Proof. ir. app surjective_pr6. app function_prj.  rw source_prj. rw target_prj.
  ir. cp H2.  ufi restriction_product H2.
  rwi productb_pr H2. ee. wri restriction_graph1 H4. 
  rwi create_domain H4.
  assert (forall i,  inc i j -> inc (V i y) (V i f)). ir.
  rwi H4 H5. cp (H5 _ H6). rwi restriction_V H7. am. am. am. am.
  cp (prolongation_exists H H0 H2 H4 H1 H6). nin H7. ee.
  assert (inc x (productb f)). rw productb_pr. rw H7. intuition. am.  
  exists x. ee.  am. rw W_prj. app function_extensionality.
  uf restriction_graph. app restr_axioms.  uf restriction_graph.
  rw restr_domain1. am. am. rw H7. am. ir. uf restriction_graph. rw restr_ev.
  sy. app H10. wrr H4. am. rww H7. wrr H4. am. am. am. am. am.
  app restr_axioms.
Qed. 

Lemma surjective_pri: forall f k,
  fgraph f -> (forall i, inc i (domain f)  -> nonempty (V i f)) ->
  inc k (domain f) -> surjective (pr_i f k). 
Proof. ir. set (j:= singleton k). 
  assert (sub j (domain f)).  red. ir. ufi j H2. cp (singleton_eq H2). rww H3.
  assert (composable  (product1_canon (V k f) k) (pr_i f k)). red. ir.
  ee. app function_product1_canon. app function_pri. rw source_product1_canon.
  rw target_pri. tv. 
  assert (is_function (compose  (product1_canon (V k f) k) (pr_i f k))). 
  app is_function_compose.  
  assert (pr_j f j = compose (product1_canon (V k f) k) (pr_i f k)).
  ap funct_extensionality. app function_prj.  am. rw source_compose. 
  rw source_prj.  rw source_pri. tv. rw target_prj. rw target_compose. 
  rw target_product1_canon. uf product1.
  assert(restriction_graph j f = L (singleton k) (fun _ : Bset => V k f)).
  uf restriction_graph. uf j. app function_extensionality. app restr_axioms.
  app create_axioms. aw. rw restr_domain1. tv. am. am.
  ir. rwi restr_domain1 H5.  cp (singleton_eq H5). rww H6. 
  aw. rw restr_ev. tv. am. am. fprops. fprops. am. am.
  uf restriction_product. ufi restriction_graph H5. rww H5. 
  rw source_prj.  ir. rw W_compose. 
  rw W_product1_canon. rw W_prj. uf restriction_graph. 
  app function_extensionality. app restr_axioms. rwi productb_pr H5; ee; am. 
  app create_axioms. aw. rw restr_domain1. tv.  rwi productb_pr H5; ee; am. 
  tv. rwi productb_pr H5; ee.  rww H6. am.
  ir. rwi restr_domain1 H6.  aw. rw W_pri. rw restr_ev. 
  ufi j H6. cp (singleton_eq H6). rww H7.  rwi productb_pr H5; ee; am.
  rwi productb_pr H5; ee. rww H7. am. am. am. am. am. 
  rwi productb_pr H5; ee; am. rwi productb_pr H5; ee. rww H7. am. am. am. am.
  assert (inc x (source (pr_i f k))). rw source_pri. am.
  assert (is_function (pr_i f k)). app function_pri.
  cp (inc_W_target H7 H6).  rwi target_pri H8. am. am. rw source_pri. am.
  assert (surjective (compose (product1_canon (V k f) k) (pr_i f k))).
  wr H5. app surjective_prj.
  assert (bijective (product1_canon (V k f) k)). app bijective_product1_canon.
  red in H7. ee. 
  app (surj_left_compose2 H3 H6 H7).
Qed.

Lemma productb_monotone1: forall f g,
  fgraph f -> fgraph g  -> domain f = domain g ->
  (forall i, inc i (domain f) -> sub (V i f) (V i g))
  -> sub (productb f) (productb g).
Proof. ir. red. ir. rwi productb_pr H3. ee. rw productb_pr. ee. am. wrr H1.
  ir. wri H4 H2. app H2. app H5. am. am.
Qed.

Lemma productb_monotone2: forall f g,
  fgraph f -> fgraph g  -> domain f = domain g ->
  (forall i, inc i (domain f)  -> nonempty (V i f)) ->
  sub (productb f) (productb g)  -> 
  (forall i, inc i (domain f) -> sub (V i f) (V i g)).
Proof. ir. red. ir.  cp (surjective_pri H H2 H4). 
  assert (inc x (target (pr_i f i))). rw target_pri. am. 
  cp (surjective_pr2 H6 H7). nin H8.  ee. rwi source_pri H8.
  cp (H3 _ H8).  rwi W_pri H9. rw H9. rwi productb_pr H10.  ee. app H12.
  rw H11. wrr H1. am. am. am. am.
Qed.


(** ** EII-5-5 associativity of products of sets *)

Definition axioms_prod_assoc f g :=
  fgraph f  & partition_fam g (domain f).

Definition prod_assoc_map f g :=
  BL(fun z => (L (domain g) (fun l => W z (pr_j f (V l g)))))
  (productb f) 
  (productf (domain g) (fun l => (restriction_product f (V l g)))).

Lemma source_pam: forall f g, 
  source (prod_assoc_map f g) = productb f.
Proof. ir. uf prod_assoc_map. app af_source. Qed.

Lemma target_pam: forall f g, 
  target (prod_assoc_map f g) = 
  (productf (domain g) (fun l => (restriction_product f (V l g)))).
Proof. ir. uf prod_assoc_map. app af_target. Qed.

Lemma axioms_pam: forall f g,
  axioms_prod_assoc f g -> 
  transf_axioms(fun z => (L (domain g) (fun l => W z (pr_j f (V l g)))))
  (productb f) 
  (productf (domain g) (fun l => (restriction_product f (V l g)))).
Proof. ir. red.  red in H. ee. ir. rw productf_pr. aw.
  ee. app create_axioms. tv. ir. rw create_V_rewrite.
  wr target_prj. app inc_W_target. app function_prj.  red in H0. ee.
  wr H4. red. ir. apply unionb_inc with i.  am. am. am. 
Qed.

Lemma function_pam: forall f g,
  axioms_prod_assoc f g -> 
  is_function (prod_assoc_map f g). 
Proof. ir. uf prod_assoc_map. app af_function. app axioms_pam. Qed.

Lemma W_pam: forall f g x,
  axioms_prod_assoc f g -> inc x (productb f) ->
  W x (prod_assoc_map f g) =  (L (domain g) (fun l => W x (pr_j f (V l g)))).
Proof. ir. uf prod_assoc_map. rw W_af_function. tv. app axioms_pam.  am.
Qed.

Lemma injective_pam: forall f g,
  axioms_prod_assoc f g ->
  injective(prod_assoc_map f g).
Proof. ir. red.  ee. app function_pam. rw source_pam. ir. red in H; ee.
  rw (productb_extensionality H H0 H1). ir. rwi W_pam H2. rwi W_pam H2.
  red in H3. ee. wri H6 H4. cp (unionb_exists H4).  nin H7. ee.
  set (aux:= L (domain g) (fun l  => W x (pr_j f (V l g)))).
  assert (fgraph aux). uf aux. app create_axioms.
  assert (inc x0 (domain aux)). uf aux. rww create_domain.  
  assert ( W x (pr_j f (V x0 g)) = V x0 aux). uf aux. rww create_V_rewrite.
  assert ( W y (pr_j f (V x0 g)) = V x0 aux). uf aux. rw H2. 
  rww create_V_rewrite.  rwi W_prj H11. rwi W_prj H12.
  assert (V i x = V i (V x0 aux)). wr H11. uf restriction_graph. rw restr_ev.
  tv. rwi productb_pr H0; ee; am. rwi productb_pr H0; ee. rw H13. red. ir.
  wr H6. cp (unionb_inc H7 H15). am. am.  am.
  assert (V i y = V i (V x0 aux)). wr H12. uf restriction_graph. rw restr_ev. 
  tv. rwi productb_pr H1; ee; am. rwi productb_pr H1; ee. rw H14. red. ir.
  wr H6. cp (unionb_inc H7 H16). am. am. am. rw H14. am. am. wr H6.
  red. ir. cp (unionb_inc H7 H13).  am. am. am.
  wr H6. red. ir. cp (unionb_inc H7 H13). am. am. red; intuition. am.
  red; intuition. am.
Qed.

Theorem bijective_pam: forall f g,
  axioms_prod_assoc f g ->
  bijective(prod_assoc_map f g).
Proof. ir. red. ee. app injective_pam. app surjective_pr6. app function_pam.
  rw target_pam. rw source_pam. ir. assert (Hb:= H0). assert (Hc:= H).
  rwi productf_pr H0. nin H0. ee. red in H. ee. 
  assert (Hd: forall i, inc i (domain g) -> sub (V i g) (domain f)).
  red in H3. ee. wr H5. ir.  red. ir. app (unionb_inc H6 H7).
  set (h:= fun i => BL (fun z => (V z (V  i y))) (V i g) (unionb f)).
  assert(Ha:forall i, inc i (domain g) ->
    transf_axioms (fun z => V z (V i y)) (V i g) (unionb f)).
  red. ir. red in H3. ee.
  rwi H1 H2. cp (H2 _ H4). ufi restriction_product H8. 
  rwi productb_pr H8. nin H8. ee. wri restriction_graph1 H9.
  rwi create_domain H9. wri H9 H5. cp (H10 _ H5). wri restriction_graph1 H11.
  rwi create_V_rewrite H11. apply unionb_inc with c.  
  app (Hd i). wr H9.  am. am. wr H9. am. am.  app Hd. am. app Hd. 
  app restr_axioms.
  assert (forall i, inc i (domain g) -> is_function (h i)). ir.
  uf h. app af_function. app Ha. 
  set (t:=unionb f).
  assert (forall i, inc i (domain g) -> target (h i) = t). ir. uf h.
  rw af_target. tv.
  assert (forall i, inc i (domain g) -> source (h i) = (V i g)).
  uf h. ir. rw af_source. tv. 
  cp (prolongation_partition _ H3 H4 H5 H6). nin H7. clear H8. nin H7. ee.
  assert (inc (graph x) (productb f)). rw productb_pr. ee.
  app fgraph_function. 
  red in H7; ee. wr H12. am. red in H7; ee. ir.  
  change (inc (W i x) (V i f)). wri H12 H13. rwi H8 H13. red in H3. ee.
  wri H15 H13. cp (unionb_exists H13). nin H16. ee. cp (H10 _ H16).
  red in H18. nin H18. ee. rww H20. uf h. rww W_af_function.
  wri H1 H16. cp (H2 _ H16).  rwi H1 H16. ufi restriction_product H21. 
  rwi productb_pr H21.  nin H21. ee. wri restriction_graph1 H23.
  wri restriction_graph1 H22. rwi create_domain H22. rwi H22 H23.
  cp (H23 _ H17).  rwi create_V_rewrite H24. am. am. am. app Hd. am.
  app Hd. app restr_axioms.  app Ha. am.
  exists (graph x). ee. am.
  assert (inc (W (graph x) (prod_assoc_map f g))
    (target  (prod_assoc_map f g))).
  app inc_W_target. app function_pam. rwi target_pam H12. 
  rw (productf_extensionality Hb H12). ir.  rww W_pam. rww create_V_rewrite. 
  rww W_prj. cp (H10 _ H13). red in H14. ee. uf restriction_graph. 
  wri H1 H13.  cp (H2 _ H13). rwi H1 H13. ufi restriction_product H17.
  rwi productb_pr H17. ee. wri restriction_graph1 H18.  rwi create_domain H18. 
  app function_extensionality.  app restr_axioms. fprops. rw restr_domain1. am.
  fprops. red in H7. ee. wr H21. am.
  ir. rwi H18 H20. rww restr_ev.  change (V x0 (V i y) = W x0 x).
  cp (H10 _ H13). red in H21. ee. cp (H23 _ H20). rw H24. uf h. 
  rw W_af_function. tv. app Ha. am.  fprops. red in H7. ee. wr H22. am.
  am.  wrr H8.  app restr_axioms. app Hd. 
Qed.

Lemma prod_assoc_map2: forall f g a b,
  axioms_prod_assoc f g -> neq b a -> domain g = doubleton a b
  -> equipotent (productb f) 
  (product (restriction_product f (V  a g)) (restriction_product f (V b g))).
Proof. ir. apply equipotent_transitive with  
  (productf (domain g) (fun l => (restriction_product f (V l g)))).
  red.  exists (prod_assoc_map f g). ee. app bijective_pam. 
  rw source_pam. tv. rw target_pam. tv. app equipotent_symmetric.
  set (xa:=(restriction_product f (V a g))).
  set (xb:= (restriction_product f (V b g))). 
  set (z:=  (productf (domain g) (fun l  => restriction_product f (V l g)))).
  assert(z = product2 xa xb a b). uf product2. uf productf.
  assert (Lvariant a b xa xb =
    L (domain g) (fun l => restriction_product f (V l g))).
  app function_extensionality. uf Lvariant. app create_axioms. 
  app create_axioms.  rw domain_variant. aw. sy. am.  rw domain_variant. ir.
  cp (doubleton_or H2). rw create_V_rewrite. nin H3. rw H3. rww V_variant_a. 
  rw H3. rww V_variant_b. rww H1. ufi Lvariant H2. rww H2. 
  red.  exists (product2_canon xa xb a b). 
  ee. app bijective_product2_canon. rw source_product2_canon. tv.
  rw target_product2_canon.  sy. tv.
Qed.

Lemma bijective_p: forall x y,
  is_singleton y -> bijective (first_proj (product x y)).
Proof. ir. assert(is_graph (product x y)). app product_is_graph. 
  assert (is_function (first_proj (product x y))). app function_first_proj.
  nin H. red. ee. red. ee. am. rw source_first_proj. ir.
  rwi W_first_proj H4. rwi W_first_proj H4. 
  rwi inc_product H2. rwi inc_product H3. ee. rwi H H8; rwi H H6. 
  app pair_extensionality. cp (singleton_eq H8). rw H9. 
  cp (singleton_eq H6). sy; am. am. am. am. am.
  assert (inc x0 y). rw H. fprops. 
  app surjective_pr6. rw source_first_proj. rw target_first_proj.
  rw domain_product. ir. exists (J y0 x0). ee. app product_pair_inc.
  rw W_first_proj. rw pr1_pair. tv. am.  app product_pair_inc. exists x0. am.
Qed.

Lemma bijective_prj_aux1: forall f j,
  fgraph f -> sub j (domain f) ->
  (forall i, inc i (complement (domain f) j) -> is_singleton (V i f)) ->
  is_singleton (restriction_product f (complement (domain f) j)).
Proof. ir.
  uf restriction_product. app singleton_product. wr restriction_graph1. 
  app create_axioms. am. app sub_complement. wr restriction_graph1. 
  aw. ir. aw. app H1. am. app sub_complement.
Qed.

Lemma bijective_prj_aux2: forall f g,
  let alpha := Ro two_points_a  in 
    let beta :=  Ro two_points_b in
      two_points = domain g ->
      target (prod_assoc_map f g) = 
      source (inverse_fun (product2_canon  (restriction_product f (V alpha g))
  (restriction_product f (V beta g)) alpha beta)).
Proof. ir. rw target_pam.
  rw source_inverse. rw target_product2_canon.
  set (z:=  (productf (domain g) (fun l => restriction_product f (V l g)))).
  uf product2. uf productf.
  assert (Lvariantc  (restriction_product f (V alpha g))
  (restriction_product f (V beta g)) =
    L (domain g) (fun l => restriction_product f (V l g))).
  app function_extensionality. app axioms_Lvariantc.  app create_axioms. 
  aw. rw domain_Lvariantc. am. rw domain_Lvariantc. ir.
  rw create_V_rewrite. rwi two_points_pr H0. nin H0. 
  rw H0. rw V_variant_ca. tv. rw H0. rw V_variant_cb. tv. wrr H.
  ufi Lvariantc H0. fold alpha in H0.  fold beta in H0. ufi Lvariant H0.
  rw H0.  ufi productf z. tv.
Qed.

Lemma bijective_prj: forall f j,
  fgraph f -> sub j (domain f) ->
  (forall i, inc i (complement (domain f) j) -> is_singleton (V i f)) ->
  bijective (pr_j f j).
Proof. ir. set (k:= (complement (domain f) j)).
  assert (is_singleton (restriction_product f k)). uf k. app bijective_prj_aux1.
  set(g:= partition_with_complement (domain f) j).
  ufi partition_with_complement  g.
  assert (axioms_prod_assoc f g). red.  ee. am. 
  cp (is_partition_with_complement H0).  uf g. am.
  set (f1:= prod_assoc_map f g). assert (bijective f1). uf f1. 
  app bijective_pam. 
  set (alpha := Ro two_points_a).
  set (beta := Ro two_points_b).
  set (xa:=(restriction_product f (V alpha g))).
  set (xb:= (restriction_product f (V beta g))).
  set (f2:= inverse_fun((product2_canon xa xb alpha beta))).
  assert (domain g = two_points). uf g. rww domain_Lvariantc.
  assert (V alpha g = j). uf g. rw V_variant_ca. tv. 
  assert (V beta g = k). uf g. rw V_variant_cb. tv. 
  assert(neq beta alpha). red. red. uf alpha; uf beta. 
  ir. cp (R_inj H8). discriminate H9.
  assert (bijective f2). uf f2. ap inverse_bij_is_bij1.
  ap  bijective_product2_canon. am.
  assert (target f1 = source f2). uf f1; uf f2. uf xa; uf xb.
  uf alpha; uf beta. app bijective_prj_aux2. sy. am.
  set (f3:= first_proj (product xa xb)).
  assert (is_singleton xb). uf xb. 
  uf restriction_product. app singleton_product. wr restriction_graph1. 
  app create_axioms. am. rww H7. uf k. app sub_complement. 
  wr restriction_graph1. aw. ir. aw. app H1. 
  fold k.  wrr H7.  am.  rw H7. uf k. app sub_complement.
  assert (bijective f3).  uf f3. app bijective_p. 
  assert (target f2 = source f3). uf f2; uf f3. rw target_inverse.
  rw source_first_proj. rw source_product2_canon. tv.
  assert (composable f3 f2). red. ee. red in H12; ee; red in H12; ee; am. 
  app bij_is_function. sy. am.
  assert (bijective (compose f3 f2)).  app bij_compose.
  assert (composable (compose f3 f2) f1). red. ee.  app bij_is_function.
  app bij_is_function. rw source_compose. sy. am.
  assert (bijective (compose (compose f3 f2) f1)).  app bij_compose.
  assert (pr_j f j =compose (compose f3 f2) f1).
  ap funct_extensionality. app function_prj.  app bij_is_function.
  rw source_compose. uf f1.
  rw source_prj.  rw source_pam. tv. rw target_prj. rw target_compose.
  rw target_compose. uf f3. rw target_first_proj. rw domain_product. wrr H6.
  red in H11.  nin H11. rw H11. exists x. fprops.
  rw source_prj. ir. rw W_compose. rw W_compose. uf f3. rw W_first_proj.
  set (y:= W x f1). assert (y= W x f1).  tv. ufi f1 H19. rwi W_pam H19.
  set (z:= W y f2). assert (z = W y f2). tv. 
  assert (inc y (target f1)).  uf y. app inc_W_target. 
  red in H4; ee; red in H4; ee; am.  
  assert (inc z (target f2)). uf z. ap inc_W_target.  app bij_is_function.
  wr H10. am.
  assert (inc (J y z) (graph f2)). rw H20. ap defined_lem. app bij_is_function.
  wr H10. am. ufi f2 H23. rwi graph_inverse H23. rwi inverse_graph_pr H23. ee. 
  rwi pr1_pair H24. rwi pr2_pair H24. 
  assert (inc z (source( product2_canon xa xb alpha beta))).
  ufi f2 H22. rwi target_inverse H22. am.
  assert (is_function (product2_canon xa xb alpha beta)).
  app function_product2_canon.
  cp (pr2_W H26 H24). rwi pr1_pair H27. rwi pr2_pair H27.
  rwi W_product2_canon H27.
  assert (P z = V alpha y). rw H27. rw V_variant_a. tv. rw H28. rw H19. 
  rw create_V_rewrite. rw H6. app refl_equal. rw H5. uf alpha. app R_inc. am.
  rwi source_product2_canon H25. am. 
  assert (is_function (product2_canon xa xb alpha beta)).
  app function_product2_canon.  fprops. am. am. app product_is_graph.
  assert(source f3 = (product xa xb)). uf f3. rw source_first_proj. tv.
  wr H19. wr H13. app inc_W_target.  app bij_is_function. wr H10.
  app inc_W_target. app bij_is_function. am. wr H10.
  app inc_W_target.  app bij_is_function. am. uf f1. rw source_pam. am.
  rw H18. am.
Qed.

(** ** EII-5-6 distributivity formulae *)

Theorem distrib_union_inter: forall f,
  fgraph f -> nonempty (domain f) ->
  (forall l, inc l (domain f) -> fgraph (V l f)) ->
  (forall l, inc l (domain f) -> nonempty (domain (V l f))) ->
  unionf (domain f) (fun l => intersectionb (V l f)) =
  intersectionf (productf (domain f) (fun l => (domain (V l f))))
  (fun g => (unionf (domain f) (fun l => V (V l g) (V l f)))).
Proof. ir. set_extens. app intersectionf_inc. uf productf. app nonempty_product.
  app create_axioms. aw. ir. rw create_V_rewrite.  app H2. am.
  ir. cp (unionf_exists  H3). nin H5. ee. rwi productf_pr H4. ee.
  apply unionf_inc with x0. am. 
  cp (H1 _ H5).  wri H7 H5. cp (H8 _ H5). cp (intersectionb_forall H6 H10).
  am. apply by_cases with 
  (inc x (unionf (domain f) (fun l => intersectionb (V l f)))).
  tv. ir.
  assert (nonempty  (productf (domain f) (fun l =>
    (Zo (domain (V l f)) (fun i =>  ~ (inc x (V i (V l f)))))))).
  uf productf. ap nonempty_product. app create_axioms. aw. ir.
  rw create_V_rewrite.
  set (jl:=(Zo (domain (V i f)) (fun i0 => ~ inc x (V i0 (V i f))))).
  cp (emptyset_dichot jl). nin H6. elim H4. apply unionf_inc with i. am.
  app intersectionb_inc. cp (H2 _ H5).  nin H7.  awi H7.   nin H7.  
  exists (J y x0). am. cp (H1 _ H5).  nin H8;ee;am. ir.
  apply by_cases with (inc i0 jl).  ir. rwi H6 H8. nin H8. elim x0. ir.
  ufi jl H8. apply by_cases with (inc x (V i0 (V i f))). ir. am. ir.
  elim H8. Ztac. am. am.  nin H5. 
  assert (inc y (productf (domain f) (fun l => domain (V l f)))).
  rw productf_pr. rwi productf_pr H5. ee. am. am. ir. cp (H7 _ H8). Ztac. am.
  cp (intersectionf_forall H3 H6). simpl in H7. cp (unionf_exists H7). nin H8.
  ee. rwi productf_pr H5. ee. wri H10 H8.  cp (H11 _ H8). Ztac. 
  elim H14. am.
Qed.


Lemma complementary_union1: forall sf f x,
  nonempty sf ->  (forall i, inc i sf -> sub (f i) x) ->
  complement x (unionf sf f) = intersectionf sf (fun i=> complement x (f i)).
Proof. ir. set (g:= fun (u:sf) => f (Ro u)).  cp (nonempty_equiv H).
  uf unionf. uf intersectionf. fold g.
  assert (forall i, sub (g i) x).  uf g. ir. app (H0 (Ro i)). app R_inc.
  rw (complementary_union _ H1 H2). uf g. tv.
Qed.

Lemma complementary_intersection1:  forall sf f x,
  nonempty sf ->  (forall i, inc i sf -> sub (f i) x) ->
  complement x (intersectionf sf f) = unionf sf (fun i=> complement x (f i)).
Proof. ir. set (g:= fun (u:sf) => f (Ro u)).  cp (nonempty_equiv H).
  uf unionf. uf intersectionf. fold g.
  assert (forall i, sub (g i) x).  uf g. ir. app (H0 (Ro i)). app R_inc.
  rw (complementary_intersection _ H1 H2). uf g. tv.
Qed.

Theorem distrib_inter_union: forall f,
  fgraph f -> nonempty (domain f) ->
  (forall l, inc l (domain f) -> fgraph (V l f)) ->
  (forall l, inc l (domain f) -> nonempty (domain (V l f))) ->
  intersectionf (domain f) (fun l => unionb (V l f)) =
  unionf (productf (domain f) (fun l => (domain (V l f))))
  (fun g => (intersectionf (domain f) (fun l => V (V l g) (V l f)))).
Proof. ir. set (b:=unionf (domain f) (fun l => (unionf (domain (V l f))
  (fun i => (V i (V l f)))))).
  set(ff := L(domain f)(fun l => L(domain (V l f))(fun i=> 
    (complement b (V i (V l f)))))).
  assert (fgraph ff). uf ff; app create_axioms.
  assert (domain ff = domain f). uf ff. aw.  
  assert(nonempty (domain ff)).  rww H4.
  assert (forall l, inc l (domain ff) -> fgraph (V l ff)). ir. uf ff. 
  rw create_V_rewrite. app create_axioms.   wrr H4.
  assert (forall l, inc l (domain ff) -> domain (V l f) =(domain (V l ff))).
  ir. uf ff. rw create_V_rewrite. aw. wrr H4. 
  assert (forall l, inc l (domain ff) -> nonempty (domain (V l ff))).
  ir. wrr H7. app H2. wrr H4. 
  cp (distrib_union_inter H3 H5 H6 H8).
  assert(Ha: forall l, forall i,
    inc l (domain ff) -> inc i (domain (V l ff)) -> sub (V i (V l ff)) b).
  ir. uf ff. rw create_V_rewrite.  rw create_V_rewrite. app sub_complement. 
  rww H7. wrr H4. 
  assert(Hb: forall l, forall i,
    inc l (domain f) -> inc i (domain (V l f)) -> sub (V i (V l f)) b).
  ir. uf b. red. ir. apply unionf_inc with l. am. apply unionf_inc with i.
  am. am.
  assert(Hc: forall l, forall i,  inc l (domain f) -> inc i (domain (V l f))
    -> complement b (V i (V l ff)) = V i (V l f)).
  ir. uf ff. rww create_V_rewrite. rww create_V_rewrite. cp (Hb _ _ H10 H11).  
  rww double_complement.
  set (x1:=  unionf (domain ff) (fun l => intersectionb (V l ff))).
  set(x2:=intersectionf (domain f) (fun l => unionb (V l f))).
  assert(complement b x1 = x2). set (sf:= domain ff). assert (nonempty sf).
  uf sf. am. 
  assert (forall i, inc i sf -> sub ((fun l => intersectionb (V l ff)) i) b).
  ir. red. ir. ufi sf H11. cp (H6 _ H11). cp (H8 _ H11).  nin H14.
  cp (intersectionb_forall  H12 H14). app (Ha _ _ H11 H14).
  cp (complementary_union1 _ H10 H11). ufi sf H12. fold x1 in H12.
  rwi H4 H12. assert (forall i, inc i  (domain f) -> 
  complement b (intersectionb (V i ff))  = unionb (V i f)). ir.
  set (sff:= domain (V i ff)). assert (nonempty sff). uf sff. app H8. rww H4.
  set (af:= fun j => V j (V i ff)).
  assert(complement b (intersectionf sff af) = unionb (V i f)).
  assert (forall j, inc j sff -> sub (af j) b). ir.
  uf af. red. ir. wri H4 H13. ufi sf H15.  app (Ha _ _ H13 H15).
  cp (complementary_intersection1 _ H14 H15). rw H16. uf af. 
  assert (L sff (fun i0  => complement b (V i0 (V i ff))) = V i f).
  app function_extensionality. app create_axioms. app H1. rw create_domain. 
  rww H7. rww H4. 
  rw create_domain. ir. rw create_V_rewrite. app Hc. ufi sff H17. 
  rww H7. rww H4. am.  uf unionb. uf sff. rw H7. app unionf_extensionality.
  ir. wr H17. rw create_V_rewrite. tv. uf sff. am. rww H4.
  assert (V i ff = L sff af). ap function_extensionality.  app H6.
  rw H4. am. app create_axioms. rw create_domain. tv.
  uf af. ir. rw create_V_rewrite. tv.  am.  rww H16. 
  assert (intersectionf sff af = intersectionb (L sff af)).
  uf intersectionb. rw create_domain.  app intersectionf_extensionality.
  ir. wr H16. uf af. tv. wr H17. am.
  rw H12. uf x2. app intersectionf_extensionality.
  set (x3 :=intersectionf (productf (domain ff) (fun l => domain (V l ff)))
         (fun g  => unionf (domain ff) (fun l => V (V l g) (V l ff)))).
  set (x4:= unionf (productf (domain f) (fun l => domain (V l f)))
     (fun g  => intersectionf (domain f) (fun l => V (V l g) (V l f)))).
  fold x1 in H9. fold x3 in H9. wr H10. rw H9. clear H10. uf x3. uf x4.
  set (sfa:=  (productf (domain ff) (fun l => domain (V l ff)))).
  assert (nonempty sfa). uf sfa.  uf productf. app nonempty_product.
  app create_axioms. aw. ir. rw create_V_rewrite. app H8.  am.
  assert (forall i, inc i sfa -> 
    sub ((fun g => unionf (domain ff) (fun l => V (V l g) (V l ff))) i) b).
  ir. red. ir. cp (unionf_exists  H12). nin H13. ee. ufi sfa H11.
  rwi productf_pr H11. ee. cp  H13. wri H15 H17.  cp (H16 _ H17).
  app (Ha _ _ H13 H18).
  cp (complementary_intersection1 _ H10 H11). rw H12. clear H12. wr H4. 
  assert (sfa = (productf (domain ff) (fun l => domain (V l f)))).
  uf sfa. uf productf. app uneq. app function_extensionality1. ir. sy. app H7.
  wr H12. app unionf_extensionality. ir.
  assert( unionb (L (domain ff) (fun l  => V (V l i) (V l ff))) =
  unionf (domain ff) (fun l => V (V l i) (V l ff))). uf unionb.
  rw create_domain. app unionf_extensionality. ir. rw create_V_rewrite. tv.
  am. wr H14. clear H14. wri H4 H0.
  assert (forall w, inc w (domain ff) ->sub ((fun l=> V (V l i) (V l ff)) w) b).
  ir. ufi sfa H13. rwi productf_pr H13. nin H13. ee. wri H15 H14.
  cp (H16 _ H14). rwi H15 H14. cp (Ha _ _ H14 H17). am.
  cp (complementary_union1 _ H0  H14).
  assert ( unionf (domain ff) (fun l  => V (V l i) (V l ff)) = 
    unionb (L (domain ff) (fun l  => V (V l i) (V l ff)))).
  uf unionb. rw create_domain. app unionf_extensionality. ir.
  rw create_V_rewrite. tv. am. wr H16. rw H15.
  app intersectionf_extensionality. ir. rw Hc. tv. wrr H4.
  ufi sfa H13. rwi productf_pr H13. nin H13. ee. 
  rw H7.  app H19. rw  H18. am.  am. 
Qed.

Lemma fgraph_rec_Lvariantc: forall f g,
  fgraph f -> fgraph g -> 
  (forall l, inc l (domain (Lvariantc f g)) -> 
    fgraph (V l (Lvariantc f g))). 
Proof. ir. rwi domain_Lvariantc H1. rwi two_points_pr H1. nin H1. rw H1.
  rw V_variant_ca. am. rw H1. rw V_variant_cb. am.
Qed.

Lemma nonempty_rec_dom_Lvariantc: forall f g,
  nonempty (domain f) -> nonempty (domain g) ->
  forall l, inc l (domain (Lvariantc f g)) -> 
    nonempty (domain (V l (Lvariantc f g))). 
Proof. ir. rwi domain_Lvariantc H1. rwi two_points_pr H1. nin H1. rw H1.
  rw V_variant_ca. am. rw H1. rw V_variant_cb. am.
Qed.

Lemma product_Lvariantc: forall f g,
  productf (domain (Lvariantc f g)) (fun l => domain (V l (Lvariantc f g))) =
  product2 (domain f) (domain g) (Ro two_points_a) (Ro two_points_b). 
Proof. ir. uf productf. uf product2. uf productf. rw domain_Lvariantc. 
  app uneq. rw two_points_pr2. app function_extensionality1. 
  ir. rwi two_points_pr H.  nin H. rw H. rw V_variant_ca. 
  rww variant_if_rw. rw H. rw V_variant_cb. rww variant_if_not_rw.
  ap two_points_distinctb.
Qed.

Lemma distrib_union2_inter: forall f g,
  fgraph f -> fgraph g -> nonempty (domain f) -> nonempty (domain g) ->
  union2 (intersectionb f)(intersectionb g) =
  intersectionf(product (domain f)(domain g)) (fun z => 
    (union2 (V (P z) f) (V (Q z) g))).
Proof. ir. 
  assert (fgraph (Lvariantc f g)). uf Lvariantc. app axioms_variant.
  assert (domain (Lvariantc f g) = two_points). app domain_Lvariantc.  
  cp (nonempty_domain_Lvariantc f g).
  cp (fgraph_rec_Lvariantc H H0).
  cp (nonempty_rec_dom_Lvariantc H1 H2).
  cp (distrib_union_inter H3 H5 H6 H7).
  assert(unionf (domain (Lvariantc f g)) 
    (fun l  => intersectionb (V l (Lvariantc f g))) =
    union2 (intersectionb f) (intersectionb g)).
  rw H4. uf pair. cp  (union_of_twosets_aux (Ro two_points_a) (Ro two_points_b)
  (fun l  => intersectionb (V l (Lvariantc f g)))). 
  rwi two_points_pr2 H9. rw H9.  rw V_variant_ca. rw V_variant_cb. tv. 
  wr H9. rw H8. clear H8; clear H9.
  cp (product_Lvariantc f g). rw H8.
  set (wg:= L(product2 (domain f) (domain g) (Ro two_points_a)
    (Ro two_points_b))
     (fun g0  => unionf (domain  (Lvariantc f g))
       (fun l => V (V l g0) (V l  (Lvariantc f g))))).
  set (wf := product2_canon (domain f) (domain g) 
    (Ro two_points_a)  (Ro two_points_b)).
  assert (bijective wf). uf wf. app bijective_product2_canon. 
  app two_points_distinctb.
  assert (fgraph (graph wf)). app fgraph_function. app bij_is_function.
  assert (fgraph wg). uf wg. app create_axioms.
  assert(range (graph wf) = domain wg). uf wg. rw create_domain. 
  red in H9. ee. rw surjective_pr3. uf wf. rw target_product2_canon. tv.
  am. assert (nonempty (domain wg)). uf wg. rw create_domain.
  uf product2. uf productf. app nonempty_product. app create_axioms.
  aw. ir. rw create_V_rewrite. cp (doubleton_or H13). nin H14. rw H14.
  rw variant_if_rw. am. tv. rw variant_if_not_rw. am. rw H14. 
  app two_points_distinctb.
  am.  assert (nonempty wg).  nin H13. awi H13. nin H13. exists (J y x). am.
  red in H11. ee. am.
  cp (intersectionb_rewrite H10  H12 H14).
  set (tmpa:= (fun g0  => unionf (domain (Lvariantc f g))
    (fun  l => V (V l g0) (V l (Lvariantc f g))))).
  assert (intersectionf
     (product2 (domain f) (domain g) (Ro two_points_a) (Ro two_points_b)) tmpa
     = intersectionb wg). 
  uf wg. uf intersectionb. aw. app intersectionf_extensionality. 
  ufi wg H13. awi H13. am. ir. rw create_V_rewrite. tv. am. rw H16.
  clear H16; clear tmpa.
  assert (fcomposable wg (graph wf)). red. ee. am. am.
  rw H12. fprops.
  assert (domain (fcompose wg (graph wf)) =  product (domain f) (domain g)).
  rw fcomposable_domain. red in H9. ee. red in H9; ee. red in H9; ee. wr H20.
  uf wf. rw source_product2_canon. tv. am.
  cp (alternate_compose H16).
  rw H15. uf intersectionb. rw H18. rw H17.
  app intersectionf_extensionality.  nin H1; nin H2. exists (J y y0).
  app product_pair_inc. ir. rw fcompose_ev.
  assert (V i (graph wf) = W i wf). uf W. tv. rw H20. clear H20.
  cp  (union_of_twosets_aux (Ro two_points_a) (Ro two_points_b)
    (fun l => V (V l (W i wf)) (V l  (Lvariantc f g)))). uf wg. 
  rw create_V_rewrite. rw H4. rwi two_points_pr2 H20. rw H20. 
  rw V_variant_ca. rw V_variant_cb. uf wf. rw W_product2_canon. rw V_variant_a. 
  rw V_variant_b. tv.  app two_points_distinctb.  app two_points_distinctb. 
  am. 
  assert (product2 (domain f) (domain g)  (Ro two_points_a) (Ro two_points_b)
    = target wf). uf wf. rw target_product2_canon. tv.
  rw H21. app inc_W_target. app bij_is_function. rww H17.
Qed.

Lemma distrib_inter2_union: forall f g,
  fgraph f -> fgraph g -> nonempty (domain f) -> nonempty (domain g) ->
  intersection2 (unionb f)(unionb g) =
  unionf(product (domain f)(domain g)) (fun z => 
    (intersection2 (V (P z) f) (V (Q z) g))).
Proof. ir. 
  assert (fgraph (Lvariantc f g)). uf Lvariantc. app axioms_variant.
  assert (domain (Lvariantc f g) = two_points). app domain_Lvariantc.  
  cp (nonempty_domain_Lvariantc f g).
  cp (fgraph_rec_Lvariantc H H0).
  cp (nonempty_rec_dom_Lvariantc H1 H2).
  cp (distrib_inter_union H3 H5 H6 H7).
  assert(intersectionf (domain (Lvariantc f g)) 
    (fun l  => unionb (V l (Lvariantc f g))) =
    intersection2 (unionb f) (unionb g)).
  rw H4.
  cp (intersection_of_twosets_aux  (Ro two_points_a) (Ro two_points_b)
    (fun l => unionb (V l (Lvariantc f g)))). 
  rwi two_points_pr2 H9. rw H9. rw V_variant_ca. rw V_variant_cb. tv. 
  wr H9. rw H8. clear H8; clear H9.  cp (product_Lvariantc f g). rw H8.
  set (wg:= L(product2 (domain f) (domain g)  (Ro two_points_a) 
    (Ro two_points_b))
     (fun g0 => intersectionf (domain  (Lvariantc f g))
       (fun l  => V (V l g0) (V l  (Lvariantc f g))))).
  set (wf := product2_canon (domain f) (domain g) 
    (Ro two_points_a) (Ro two_points_b)).
  assert (bijective wf). uf wf. app bijective_product2_canon. 
  app two_points_distinctb. 
  assert (fgraph (graph wf)). app fgraph_function. app bij_is_function.
  assert (fgraph wg). uf wg. app create_axioms.
  assert(range (graph wf) = domain wg). uf wg. rw create_domain. 
  red in H9. ee. rw surjective_pr3. uf wf. rw target_product2_canon. tv.
  am. assert (nonempty (domain wg)). uf wg. rw create_domain.
  uf product2. uf productf. app nonempty_product. app create_axioms.
  aw. ir. rw create_V_rewrite. cp (doubleton_or H13). nin H14. rw H14.
  rw variant_if_rw. am. tv. rw variant_if_not_rw. am. rw H14. 
  app two_points_distinctb. am. assert (nonempty wg). nin H13. awi H13. 
  nin H13. exists (J y x). am. red in H11. ee. am.
  cp (unionb_rewrite H10  H12).
  set (tmpa:= (fun g0 => intersectionf (domain (Lvariantc f g))
    (fun l => V (V l g0) (V l (Lvariantc f g))))).
  assert (unionf
     (product2 (domain f) (domain g)  (Ro two_points_a) (Ro two_points_b)) tmpa
     = unionb wg). 
  uf wg. uf unionb. aw. app unionf_extensionality. 
  ir. rw create_V_rewrite. tv. am. rw H16.  clear H16; clear tmpa.
  assert (fcomposable wg (graph wf)). red. ee. am. am.  rw H12. fprops.
  assert (domain (fcompose wg (graph wf)) =  product (domain f) (domain g)).
  rw fcomposable_domain. red in H9. ee. red in H9; ee. red in H9; ee. wr H20.
  uf wf. rw source_product2_canon. tv. am.
  cp (alternate_compose H16). rw H15. uf unionb. rw H18. rw H17.
  app unionf_extensionality. ir. rw fcompose_ev.
  assert (V i (graph wf) = W i wf). uf W. tv. rw H20. clear H20.
  cp  (intersection_of_twosets_aux  (Ro two_points_a) (Ro two_points_b)
    (fun l => V (V l (W i wf)) (V l  (Lvariantc f g)))). uf wg. 
  rw create_V_rewrite. rw H4. uf J. wr two_points_pr2. rw H20. rw V_variant_ca. 
  rw V_variant_cb. uf wf. rw W_product2_canon. rw V_variant_a. rw V_variant_b. 
  tv.  app two_points_distinctb.  app two_points_distinctb. am.
  assert (product2 (domain f) (domain g)  (Ro two_points_a) (Ro two_points_b) 
    = target wf). uf wf. rw target_product2_canon. tv. 
  rw H21. app inc_W_target.  app bij_is_function.  rww H17.
Qed.

Lemma empty_function: forall f, L emptyset f = emptyset.
Proof. ir. app function_extensionality.
  app create_axioms. red.  ir. ee. app emptyset_is_graph.
  ir.  elim (emptyset_pr H). aw. aw. ir. elim (emptyset_pr H). 
Qed.

Lemma trivial_product1: forall f, productf emptyset f = singleton emptyset.
Proof. ir. uf productf. rw empty_function. app trivial_product.
Qed.
   
Lemma unionf_singleton: forall x f, unionf (singleton x) f =  f x.
Proof. ir.  set_extens. cp (unionf_exists H). nin H0. ee. 
  cp (singleton_eq H0). wrr H2. assert (inc x (singleton x)). 
  fprops. app (unionf_inc f H0 H).
Qed.
   
Lemma intersectionf_singleton: forall x f, intersectionf (singleton x) f =  f x.
Proof. ir. assert (inc x (singleton x)). fprops.
  set_extens. app (intersectionf_forall H0 H). app intersectionf_inc.
  exists x. am. ir. cp (singleton_eq H1). rww H2.
Qed.

Lemma unionf_emptyset: forall f, unionf emptyset f =  emptyset.
Proof. ir. app empty_unionf.
Qed.

Lemma nonempty_product3: forall sf f i,
  inc i sf -> f i = emptyset ->  productf sf f = emptyset.
Proof. ir. cp (emptyset_dichot (productf sf f)). nin H1. am.
  nin H1. rwi productf_pr H1. nin H1. ee. rwi H2 H3. cp (H3 _ H).  rwi H0 H4.
  elim (emptyset_pr H4).
Qed.

Theorem distrib_prod_union: forall f,
  fgraph f ->
  (forall l, inc l (domain f) -> fgraph (V l f)) ->
  productf (domain f) (fun l => unionb (V l f)) =
  unionf (productf (domain f) (fun l => (domain (V l f))))
  (fun g => (productf (domain f) (fun l => V (V l g) (V l f)))).
Proof. ir. cp (emptyset_dichot (domain f)). nin H1. rw H1.
  rw trivial_product1. rw trivial_product1. rw unionf_singleton. 
  rw trivial_product1. tv. 
  apply by_cases with 
  (forall l, inc l (domain f) -> nonempty (domain (V l f))). ir.
  set_extens.
  assert (nonempty (productf (domain f) (fun l =>
    (Zo (domain (V l f)) (fun i => (inc (V l x) (V i (V l f)))))))).
  uf productf. ap nonempty_product. app create_axioms. aw. ir. 
  rw create_V_rewrite. rwi productf_pr H3. ee. rwi H5 H6.
  cp (H6 _ H4). cp (H0 _ H4). cp (unionb_exists H7). nin H9. ee.
  exists x0. Ztac. am. nin H4. rwi productf_pr H4. ee.
  apply unionf_inc with y. rw productf_pr. ee. am. am. ir.  cp (H6 _ H7).
  Ztac. am. rwi productf_pr H3. ee. rw productf_pr. ee. am. am. ir.
  rwi H7 H9; wri H5 H9;  cp (H6 _ H9). Ztac. am. cp (unionf_exists H3).
  nin H4. ee. rwi productf_pr H5. ee. rw productf_pr. ee. am. am.
  ir. cp (H7 _ H8). apply unionb_inc with (V i x0). 
  rwi productf_pr H4. ee. app H11. rw H10. wrr H6. am. 
  ir. (* special cases *)
  set (HH:=exists l, inc l (domain f) & domain (V l f) = emptyset).
  assert( ~ (forall l, inc l (domain f) -> nonempty (domain (V l f))) -> HH). 
  ap excluded_middle'. ir. cp (emptyset_dichot (domain (V l f))).
  nin H5. elim H3. uf HH. exists l. ee. am. am. am. 
  cp (H3 H2). ufi HH H4.  nin H4. clear H3; clear HH. clear H2. ee.
  assert (unionb (V x f) = emptyset). uf unionb. rw H3. app unionf_emptyset. 
  cp (nonempty_product3 (sf:=domain f) (fun l  => domain (V l f))).
  simpl in H5. rw (H5 _ H2 H3).
  cp (nonempty_product3 (sf:=domain f) (fun l  => unionb (V l f))).
  simpl in H6. rw (H6 _ H2 H4). rw unionf_emptyset.  tv.
Qed.

Theorem distrib_prod_intersection: forall f,
  fgraph f -> nonempty (domain f) ->
  (forall l, inc l (domain f) -> fgraph (V l f)) ->
  (forall l, inc l (domain f) -> nonempty (domain (V l f))) ->
  productf (domain f) (fun l => intersectionb (V l f)) =
  intersectionf (productf (domain f) (fun l => (domain (V l f))))
  (fun g => (productf (domain f) (fun l => V (V l g) (V l f)))).
Proof. ir. 
  assert(nonempty (productb (L (domain f) (fun l => domain (V l f))))).
  uf productf. app nonempty_product. app create_axioms. aw.  ir.
  rw create_V_rewrite. app H2.  am.
  set_extens. app intersectionf_inc.  ir. rwi productf_pr H4. rw productf_pr.
  ee. am. am. ir. cp (H7 _ H8). rwi productf_pr H5. ee.
  rwi H10 H11. wri H6 H11. cp (H11 _ H8). app intersectionb_forall.
  nin H3. cp (intersectionf_forall  H4 H3).
  simpl in H5. rwi productf_pr H5. ee.
  rw productf_pr. ee.  am. am. ir. app intersectionb_inc. 
  rwi H6 H8. cp (H2 _ H8).  nin H9. awi H9.  nin H9.  exists (J y0 x0). 
  am. cp (H1 _ H8). red in H10; ee;am. ir.
  assert (nonempty (productf (domain f) (fun l =>
    (Zo (domain (V l f)) (fun j => l = i  -> j = i0))))).
  uf productf. ap nonempty_product. app create_axioms. aw. ir.
  rw create_V_rewrite.
  apply by_cases with (i1 = i). ir. exists i0. Ztac.  rww H11. ir.
  cp (H2 _ H10). nin H12. exists y0. Ztac. ir. elim H11. am. am. 
  nin H10. rwi productf_pr H10. ee. 
  assert (inc y0 (productf (domain f) (fun l  => domain (V l f)))).
  ir. rw productf_pr. ee. am. am. ir. cp (H12 _ H13). Ztac. am.
  cp (intersectionf_forall H4 H13).  simpl in H14.
  rwi productf_pr H14. ee. cp (H16 _ H8).
  rwi H6 H8;  wri H11 H8; cp (H12 _ H8).  Ztac. rwi H20 H17. am. tv.
Qed.

Lemma partition_product: forall f,
  fgraph f -> nonempty (domain f) ->
  (forall l, inc l (domain f) -> fgraph (V l f)) ->
  (forall l, inc l (domain f) -> (partition_fam (V l f) (unionb (V l f)))) ->
  partition_fam(L(productf (domain f) (fun l => domain (V l f)) ) 
    (fun g => (productf (domain f) (fun l => V (V l g) (V l f)))))
  ( productf (domain f) (fun l => unionb (V l f))).
Proof. ir. red. ee. app create_axioms. red.  aw. ir. 
  rww create_V_rewrite. rww create_V_rewrite. apply by_cases with (i=j).
  ir. left. am. ir. right. red.
  cp (emptyset_dichot (intersection2 (productf (domain f)
    (fun l  => V (V l i) (V l f)))
  (productf (domain f) (fun l  => V (V l j) (V l f))) )).  nin H6. am.
  assert (i = j).  rwi productf_pr H3; rwi productf_pr H4. ee.
  app function_extensionality. rw H7. am. rw H9. ir.
  rwi H9  H10. cp (H10 _ H11).  clear H10. rwi H7 H8. cp (H8 _ H11). clear H8.
  nin H6. cp (intersection2_first H6); cp (intersection2_second H6).
  cp (H2 _ H11). red in H14. ee.  red in H15.
  assert (nonempty (intersection2  (V (V x i) (V x f))  (V (V x j) (V x f)))).
  exists (V x y). app intersection2_inc.
  rwi productf_pr H8. ee. rwi H17 H18. cp (H18 _ H11). am.
  rwi productf_pr H13. ee. rwi H17 H18. cp (H18 _ H11). am.
  cp (H15 _ _ H12 H10). nin H18. am.  red in H18.
  nin H17. rwi H18 H17. elim (emptyset_pr H17). elim H5. am. sy.
  rw distrib_prod_union. uf unionb. rw create_domain. 
  app unionf_extensionality. ir. rw create_V_rewrite. tv. am. am. am. 
Qed.

Lemma distrib_prod2_union: forall f g,
  let a := Ro two_points_a in let b:=Ro two_points_b in
    fgraph f -> fgraph g -> 
    nonempty (domain f) -> nonempty (domain g) ->
    product2 (unionb f)(unionb g) a b =
    unionf(product (domain f)(domain g))
      (fun z =>  (product2 (V (P z) f) (V (Q z) g) a b)).
Proof. ir.
  cp (axioms_Lvariantc f g).
  assert (domain (Lvariantc f g) = two_points). app domain_Lvariantc.  
  cp (nonempty_domain_Lvariantc f g).
  cp (fgraph_rec_Lvariantc H H0).
  cp (distrib_prod_union H3 H6).
  assert(Ha: doubleton a b = two_points). uf a; uf b; app two_points_pr2.
  assert(productf (domain (Lvariantc f g)) 
    (fun l  => unionb (V l (Lvariantc f g))) =
    product2 (unionb f) (unionb g) a b).
  rw H4. uf product2. uf productf. rw Ha.
  assert(L two_points  (fun l => unionb (V l (Lvariantc f g))) =
     Lvariantc (unionb f) (unionb g)).
  app function_extensionality. app create_axioms. app axioms_Lvariantc.
  rw create_domain. rw domain_Lvariantc. tv. 
  rw create_domain. ir. rwi two_points_pr H8. nin H8. rw H8. aw. 
  rw V_variant_ca. rw V_variant_ca. tv. app R_inc. aw.
  rw H8. rw V_variant_cb.  rw V_variant_cb. tv. rw H8. app R_inc. 
  rw H8. uf Lvariantc. uf Lvariant. fold a; fold b; rw Ha. tv.
  rwi H8 H7. rw H7. clear H7; clear H8.
  cp (product_Lvariantc f g). rw H7.
  set (wg:= L(product2 (domain f) (domain g) a b)
     (fun g0  => productf (domain  (Lvariantc f g))
       (fun l => V (V l g0) (V l  (Lvariantc f g))))).
  set (wf := product2_canon (domain f) (domain g) a b).
  assert (bijective wf). uf wf. app bijective_product2_canon. 
  app two_points_distinctb.
  assert (fgraph (graph wf)). app fgraph_function. app bij_is_function.
  assert (fgraph wg). uf wg. app create_axioms.
  assert(range (graph wf) = domain wg). uf wg. aw. 
  red in H6. ee. rw surjective_pr3. uf wf. rw target_product2_canon. tv.
  red in H8; ee; am.
  assert (nonempty (domain wg)). uf wg. aw.
  uf product2. uf productf. app nonempty_product. app create_axioms.
  aw. ir. aw. cp (doubleton_or H12). nin H13. rw H13. rw variant_if_rw. am. tv.
  rw H13. rw variant_if_not_rw. am. app two_points_distinctb.
  cp (unionb_rewrite H9 H11). 
  set (tmpa:= (fun g0  =>
      productf (domain (Lvariantc f g))
        (fun l  => V (V l g0) (V l (Lvariantc f g))))).
  assert (unionf (product2 (domain f) (domain g) a b) tmpa = unionb wg). 
  uf wg. uf unionb. aw. app unionf_extensionality. 
  ir. rw create_V_rewrite. tv. am. fold a; fold b;rw H14.  
  clear H14; clear tmpa. rw H13.
  assert (fcomposable wg (graph wf)). red. ee. am. am.
  rw H11. app sub_refl. 
  assert (domain (fcompose wg (graph wf)) =  product (domain f) (domain g)).
  rw fcomposable_domain. red in H8; ee; red in H8; ee; red in H8; ee. wr H18.
  uf wf. rw source_product2_canon. tv. am. 
  cp (alternate_compose H14). uf unionb. rw H16. rw H15.
  app unionf_extensionality. ir. rw fcompose_ev.
  assert (V i (graph wf) = W i wf). uf W. tv. rw H18. clear H18.
  uf wg. rw create_V_rewrite. uf product2. rw H4. 
  assert (L two_points (fun l  => V (V l (W i wf)) (V l (Lvariantc f g))) =
   Lvariant a b (V (P i) f) (V (Q i) g)).
  app function_extensionality. app create_axioms.  uf Lvariant. 
  app create_axioms. uf Lvariant. aw. sy; app Ha.  rw create_domain. ir. 
  aw. uf wf. rw W_product2_canon. rwi two_points_pr H18; nin H18. rw H18.
  rww V_variant_a.  rww V_variant_a. rww V_variant_ca. rw H18.
  rww V_variant_b. rww V_variant_b. rww V_variant_cb. app two_points_distinctb.
  app two_points_distinctb. app two_points_distinctb. am.
  uf productf. uf productf. rw H18. tv. 
  assert (product2 (domain f) (domain g) a b = target wf). uf wf. 
  rw target_product2_canon. tv. rw H18. app inc_W_target. app bij_is_function.
  rww H15.
Qed.

Lemma distrib_prod2_inter: forall f g,
  let a := Ro two_points_a in let b:=Ro two_points_b in
    fgraph f -> fgraph g -> 
    nonempty (domain f) -> nonempty (domain g) ->
    product2 (intersectionb f)(intersectionb g) a b =
    intersectionf(product (domain f)(domain g)) (fun z => 
      (product2 (V (P z) f) (V (Q z) g) a b)).
Proof. ir.
  cp (axioms_Lvariantc f g).
  assert (domain (Lvariantc f g) = two_points). app domain_Lvariantc.  
  cp (nonempty_domain_Lvariantc f g).
  assert(Ha: doubleton a b = two_points). uf a; uf b; app two_points_pr2.
  assert(Hb: b <> a). app two_points_distinctb.
  cp (fgraph_rec_Lvariantc H H0).
  cp (distrib_prod_intersection H3 H5 H6).
  assert(productf (domain (Lvariantc f g)) 
    (fun l => intersectionb (V l (Lvariantc f g))) =
    product2 (intersectionb f) (intersectionb g) a b).
  rw H4. uf pair. uf product2. uf productf.
  assert(L two_points (fun l => intersectionb (V l (Lvariantc f g))) =
     (Lvariant a b  (intersectionb f) (intersectionb g))).
  app function_extensionality. app create_axioms. uf Lvariant. 
  app create_axioms. uf Lvariant. aw. sy. app Ha. rw create_domain. ir. 
  rww create_V_rewrite. rwi two_points_pr H8. nin H8. rw H8. 
  rw V_variant_a. rw V_variant_ca. tv. rw H8. rw V_variant_b. rw V_variant_cb. 
  tv.  am. rw H8. uf Lvariant. tv. rwi H8 H7. rw H7. clear H7; clear H8.
  cp (product_Lvariantc f g). rw H7.
  set (wg:= L(product2 (domain f) (domain g) a b)
     (fun g0  => productf (domain  (Lvariantc f g))
       (fun l => V (V l g0) (V l  (Lvariantc f g))))).
  set (wf := product2_canon (domain f) (domain g) a b).
  assert (bijective wf). uf wf. app bijective_product2_canon.
  assert (fgraph (graph wf)). app fgraph_function. app bij_is_function.
  assert (fgraph wg). uf wg. app create_axioms.
  assert(range (graph wf) = domain wg). uf wg. aw. 
  rw surjective_pr3. uf wf. rw target_product2_canon. tv.  red in H8; ee; am.
  assert (nonempty (domain wg)). uf wg. aw.
  uf product2. uf productf. app nonempty_product. app create_axioms.
  aw. ir. rw create_V_rewrite. cp (doubleton_or H12). nin H13. rw H13. 
  rw variant_if_rw. am. tv. rw H13. rw variant_if_not_rw. am. am. am.
  assert (nonempty wg). nin H12. awi H12. nin H12. exists (J y x). am.
  red in H10; ee; am.
  cp (intersectionb_rewrite H9  H11 H13). 
  set (tmpa:= (fun g0 => productf (domain (Lvariantc f g))
        (fun l  => V (V l g0) (V l (Lvariantc f g))))).
  assert (intersectionf  (product2 (domain f) (domain g) a b) tmpa
     = intersectionb wg). 
  uf wg. uf intersectionb. aw. app intersectionf_extensionality.
  ufi wg H12.  awi H12. am.
  ir. rw create_V_rewrite. tv. am. fold a; fold b;rw H15.  
  clear H15; clear tmpa. rw H14.
  assert (fcomposable wg (graph wf)). red. ee. am. am.
  rw H11. app sub_refl. 
  cp (alternate_compose H15).
  assert (domain (fcompose wg (graph wf)) =  product (domain f) (domain g)).
  rw fcomposable_domain. red in H8; ee; red in H8; ee; red in H8; ee. wr H20.
  uf wf. rw source_product2_canon. tv. am. uf intersectionb. rw H16; rw H17.
  app intersectionf_extensionality. nin H1; nin H2. exists (J y y0).
  app product_pair_inc. ir. rw fcompose_ev.
  assert (V i (graph wf) = W i wf). uf W. tv. rw H19. clear H19.
  uf wg. rw create_V_rewrite. uf product2. rw H4. 
  assert (L two_points (fun l => V (V l (W i wf)) (V l (Lvariantc f g))) =
   Lvariant a b (V (P i) f) (V (Q i) g)).
  app function_extensionality. app create_axioms.  uf Lvariant.
  app create_axioms. uf Lvariant. aw. sy; app Ha. rw create_domain. ir. 
  aw. uf wf. rw W_product2_canon. rwi two_points_pr H19. nin H19. rw H19.
  rww V_variant_a.  rww V_variant_a. rww V_variant_ca. 
  rw H19. rww V_variant_b. rww V_variant_b. rww V_variant_cb.  am. am.
  uf productf. uf productf. rw H19. tv. 
  assert (product2 (domain f) (domain g) a b = target wf). 
  uf wf. rw target_product2_canon. tv. rw H19. app inc_W_target. 
  app bij_is_function. rww H17. app nonempty_rec_dom_Lvariantc. 
Qed.

Lemma distrib_product2_union: forall f g,
  fgraph f -> fgraph g -> 
  product (unionb f)(unionb g) =
  unionf(product (domain f)(domain g)) (fun z => 
    (product (V (P z) f) (V (Q z) g))).
Proof. ir. cp (emptyset_dichot (domain f)). nin H1. rw H1. srw. sy.
  rw empty_unionf. uf unionb.  rw H1.  rw unionf_emptyset. srw. tv. tv. 
  cp (emptyset_dichot (domain g)). nin H2. rw H2. srw. rw empty_unionf.
  uf unionb.  rw H2. rw unionf_emptyset. srw. tv. tv.
  cp (distrib_prod2_union H H0 H1 H2). ufi product2 H3.
  set_extens.  
  set (qx := Lvariantc (P x) (Q x)).
  assert (inc qx (productb (Lvariantc  (unionb f) (unionb g)))).
  rw productb_pr. uf qx. ee. app axioms_Lvariantc.
  rw domain_Lvariantc. rw domain_Lvariantc. tv.  rw domain_Lvariantc. ir.
  rwi two_points_pr H5. nin H5. rw H5. rw V_variant_ca.  rw V_variant_ca.
  rwi inc_product H4; ee; am. rw H5. rw V_variant_cb. rw V_variant_cb. 
  rwi inc_product H4; ee; am. app axioms_Lvariantc. 
  ufi productf H3. ufi Lvariantc H5. ufi Lvariant H5.  rwi H3 H5.
  cp (unionf_exists H5). nin H6. ee. apply unionf_inc with x0. am.
  rwi productb_pr H7. nin H7. ee. 
  assert (inc (Ro two_points_a) (domain qx)). uf qx. rw domain_Lvariantc. 
  app R_inc. cp (H9 _ H10). 
  awi H11. rwi variant_if_rw H11. ufi qx H11. rwi V_variant_ca H11.
  assert (inc (Ro two_points_b) (domain qx)). uf qx. rw domain_Lvariantc. 
  app R_inc. cp (H9 _ H12). awi H13. rwi variant_if_not_rw H13.
  ufi qx H13. rwi V_variant_cb H13. app product_inc.
  rwi inc_product H4; ee; am. app two_points_distinctb. fprops.
  tv. fprops. app create_axioms. 
  cp (unionf_exists H4). nin H5. ee. rwi inc_product H6. ee.
  set (qx := Lvariantc (P x) (Q x)).
  assert (inc qx (productb (Lvariantc  (unionb f) (unionb g)))). 
  ufi productf H3. uf Lvariantc. uf Lvariant.
  rw H3. apply unionf_inc with x0. am. 
  rw productb_pr. uf qx. ee. app axioms_Lvariantc. 
  aw. rw domain_Lvariantc. rw two_points_pr2. tv. rw domain_Lvariantc. ir. 
  rw create_V_rewrite. rwi two_points_pr H9. nin H9. rw H9. rw variant_if_rw.
  rw V_variant_ca. am. tv. tv. rw H9. rw variant_if_not_rw.
  rw V_variant_cb. am. app two_points_distinctb. rw two_points_pr2. am.
  app create_axioms.
  rwi productb_pr H9. nin H9. ee. rwi domain_Lvariantc H10.  
  assert (inc (Ro two_points_a) (domain qx)). rw H10.  app R_inc.
  cp (H11 _ H12). awi H13. rwi V_variant_ca H13. ufi qx H13. 
  rwi V_variant_ca H13.
  assert (inc (Ro two_points_b) (domain qx)). rw H10. app R_inc.
  cp (H11 _ H14).  ufi qx H15. awi H15.
  rwi V_variant_cb H15. rwi V_variant_cb H15.  app product_inc. 
  app axioms_Lvariantc. 
Qed.

Lemma distrib_product2_inter: forall f g,
  fgraph f -> fgraph g -> nonempty (domain f) -> nonempty (domain g) ->
  product (intersectionb f)(intersectionb g) =
  intersectionf(product (domain f)(domain g)) (fun z => 
    (product (V (P z) f) (V (Q z) g))).
Proof. ir. cp (distrib_prod2_inter H H0 H1 H2).  ufi product2 H3.
  set_extens. 
  set (qx :=  Lvariantc (P x) (Q x)).
  assert (inc qx (productb (Lvariantc  (intersectionb f) (intersectionb g)))).
  rwi inc_product H4. ee. rw productb_pr. ee. uf qx. app axioms_Lvariantc. 
  uf qx. rw domain_Lvariantc. rw domain_Lvariantc. tv.  aw. uf qx.  
  rw domain_Lvariantc. ir.  rwi two_points_pr H7. nin H7. rw H7.
  rw V_variant_ca. rw V_variant_ca. am.  rw H7. rw V_variant_cb. 
  rw V_variant_cb. am. app axioms_Lvariantc.  ufi productf H3. 
  ufi Lvariantc H5. ufi Lvariant H5.
  rwi H3 H5. app intersectionf_inc. nin H1. nin H2. exists (J y y0).
  app product_pair_inc. ir. cp (intersectionf_forall H5 H6). simpl in H7.
  rwi productb_pr H7. ee. rwi create_domain H8.
  assert (inc (Ro two_points_a) (domain qx)). rw H8. fprops.
  cp (H9 _ H10).  ufi qx H11. awi H11. rwi variant_if_rw H11. 
  rwi V_variant_ca H11. 
  assert (inc (Ro two_points_b) (domain qx)). rw H8. fprops.
  cp (H9 _ H12). ufi qx H13. awi H13. rwi variant_if_not_rw H13. 
  rwi V_variant_cb H13.  app product_inc. rwi inc_product H4. ee. am. 
  app two_points_distinctb. fprops.  tv. fprops. app create_axioms.  
  set (qx :=  Lvariantc (P x) (Q x)).
  assert (Ha:nonempty (product (domain f) (domain g))).
  nin H1. nin H2. exists (J y y0).  app product_pair_inc.
  assert (inc qx (productb (Lvariantc (intersectionb f) (intersectionb g)))).
  ufi productf H3. uf Lvariantc. uf Lvariant. rw H3. app intersectionf_inc. ir. 
  cp (intersectionf_forall H4 H5). 
  simpl in H6. rw productb_pr. ee. uf qx.  app axioms_Lvariantc. uf qx. aw.
  rw domain_Lvariantc. rw two_points_pr2. tv. ir. uf qx.  unfold qx in H7.  
  rwi domain_Lvariantc H7. rwi two_points_pr H7. nin H7. rw H7. aw. 
  rw V_variant_ca. rw variant_if_rw. rwi inc_product H6; ee; am. tv. 
  fprops. rw H7. aw. aw. rw V_variant_cb. rw variant_if_not_rw. 
  rwi inc_product H6; ee; am.  app two_points_distinctb. 
  fprops. app create_axioms. rwi productb_pr H5. ee.  
  rwi domain_Lvariantc H6. 
  assert (inc (Ro two_points_a) (domain qx)). rw H6. app R_inc.
  cp (H7 _ H8).  ufi qx H9. rwi V_variant_ca H9.  rwi V_variant_ca H9. 
  assert (inc (Ro two_points_b) (domain qx)). rw H6. app R_inc.
  cp (H7 _ H10).  ufi qx H11. rwi V_variant_cb H11.  rwi V_variant_cb H11. 
  app product_inc.  nin Ha.
  cp (intersectionf_forall H4 H12). simpl in H13. rwi inc_product H13.
  ee. am. app axioms_Lvariantc.
Qed.

Theorem distrib_inter_prod: forall f sa sb,
  fgraph f -> nonempty sb ->
  intersectionf sb (fun k => productf sa (fun i=> V (J i k) f)) =
  productf sa (fun i => intersectionf sb (fun k=> V (J i k) f)).
Proof. ir. set_extens. nin H0. cp (intersectionf_forall H1 H0). ee.
  simpl in H2. rwi productf_pr H2. ee. rw productf_pr. ee. am. am. ir.
  app intersectionf_inc.  exists y. am. ir.
  cp (intersectionf_forall H1 H6). ee.  simpl in H7. rwi productf_pr H7. ee.
  app (H9  _ H5).  rwi productf_pr H1. ee. app intersectionf_inc. ir.
  rw productf_pr. ee.  am. am. ir. cp (H3 _ H5).
  app (intersectionf_forall H6 H4).
Qed.
  
Lemma productf_extension : forall sf1 f1 sf2 f2,
  L sf1 f1 = L sf2 f2 -> productf sf1 f1 = productf sf2 f2.
Proof. ir. uf productf. rw H. tv.
Qed.


Lemma distrib_prod_inter2_prod: forall f g,
  fgraph f -> fgraph g -> domain f = domain g -> 
  nonempty (domain f) -> 
  intersection2 (productb f) (productb g) =
  productf (domain f) (fun i => intersection2 (V i f) (V i g)).
Proof. ir.
  set (sa:= domain f). set (sb := two_points).
  set (h:= L(product sa sb) 
    (fun i => Yo (Q i = Ro two_points_a) (V (P i) f) (V (P i) g))).
  assert (fgraph h). uf h. app create_axioms.
  assert (nonempty sb). uf sb.  exists (Ro two_points_a). app R_inc.
  cp (distrib_inter_prod sa H3 H4). ufi sb  H5. wri two_points_pr2 H5.
  rwi intersection_of_twosets_aux H5.
  set (fa:=productf sa (fun i => V (J i (Ro two_points_a)) h)). 
  set (fb:=productf sa (fun i => V (J i (Ro two_points_b)) h)).
  fold fa in H5. fold fb in H5. 
  assert (fa = productb f). uf fa. uf productf. app uneq. 
  app function_extensionality. app create_axioms. aw. aw. ir. 
  rw create_V_rewrite.  uf h. rw create_V_rewrite.  aw.
  rw Y_if_rw. tv. tv. app product_pair_inc. uf sb. app R_inc. am.
  assert (fb = productb g). uf fb. uf productf. app uneq. 
  app function_extensionality.  app create_axioms. aw. aw. ir.
  rw create_V_rewrite.  uf h. rw create_V_rewrite. aw.
  rw Y_if_not_rw. tv. app two_points_distinctb. 
  app product_pair_inc. uf sb.  app R_inc. am. wr H6; wr H7. uf productf. 
  rw H5. uf productf. app uneq. app function_extensionality1. ir. 
  rw intersection_of_twosets_aux. uf h.  rw create_V_rewrite.
  rw create_V_rewrite. aw. rww Y_if_rw. rww Y_if_not_rw.
  app two_points_distinctb. app product_pair_inc. uf sb.  ap R_inc. 
  app product_pair_inc. uf sb.  ap R_inc. 
Qed.

Lemma distrib_inter_prod_inter: forall f g,
  let  a:= Ro two_points_a in let b:= Ro two_points_b in
    fgraph f -> fgraph g -> domain f = domain g -> 
    nonempty (domain f) -> 
    product2 (intersectionb f) (intersectionb g) a b =
    intersectionf (domain f) (fun i => product2 (V i f) (V i g) a b).
Proof. ir.
  set (sb:= domain f). set (sa := two_points).
  set (h:= L(product sa sb)
    (fun i => Yo (P i = a) (V (Q i) f) (V (Q i) g))).
  assert (fgraph h). uf h. app create_axioms.
  assert (nonempty sb). uf sb.  am.
  cp (distrib_inter_prod sa H3 H4). sy.
  assert (intersectionf sb (fun k  => productf sa (fun i => V (J i k) h))
    = intersectionf sb  (fun i  =>  product2 (V i f) (V i g) a b)).
  app intersectionf_extensionality. ir. uf product2. 
  app productf_extension. app function_extensionality1. ir.
  uf a; uf b; uf sa; sy; app two_points_pr2.
  ir. ufi sa H7.  rwi two_points_pr H7; nin H7. rw H7.
  rw variant_if_rw. uf h.  rw create_V_rewrite. aw. rw Y_if_rw. tv. tv.
  app product_pair_inc. uf sa.  ap R_inc. tv. rw H7.
  rw variant_if_not_rw. uf h.  rw create_V_rewrite. aw. rw Y_if_not_rw. tv. 
  app two_points_distinctb. app product_pair_inc. uf sa. ap R_inc. 
  app two_points_distinctb.  wr H6. rw H5.
  uf product2. app productf_extension. app function_extensionality1.
  uf a; uf b; uf sa; sy; app two_points_pr2. ir. 
  ufi sa H7.  rwi two_points_pr H7; nin H7. rw H7.
  rw variant_if_rw. uf h. uf intersectionb. app intersectionf_extensionality.
  ir. rw create_V_rewrite. aw. rw Y_if_rw. tv. tv.
  app product_pair_inc. uf sa.  ap R_inc. tv. 
  rw H7.  rw variant_if_not_rw.  uf h. uf intersectionb. wr H1.
  app intersectionf_extensionality. ir.  rw create_V_rewrite. aw.
  rw Y_if_not_rw. tv. app two_points_distinctb.  app product_pair_inc. uf sa. 
  ap R_inc. app two_points_distinctb. 
Qed.

(** Product of two products of family of sets *)

Definition prod_of_function x x':=
  L(domain x)(fun i => J (V i x) (V i x')).

Definition prod_of_products_canon f f':=
  BL(fun w => prod_of_function (P w) (Q w))
  (product (productb f) (productb f'))
  (productf (domain f)(fun i => product (V i f) (V i f'))).



Definition prod_of_product_aux f f' :=
  fun i =>  (product (W i f) (W i f')).

Definition prod_of_prod_target f f' :=
  fun_image(source f)(prod_of_product_aux f f').

Definition prod_of_products f f' :=
  BL (prod_of_product_aux f f')(source f)(prod_of_prod_target f f').

Lemma inc_prod_of_prod_target: forall f f' x,
  inc x (prod_of_prod_target f f') = 
  (exists i, inc i (source f) & product (W i f) (W i f') = x).
Proof. ir. uf prod_of_prod_target. aw.
Qed.

Lemma function_prod_of_products: forall f f',
  is_function (prod_of_products f f').
Proof. ir. uf prod_of_products. app af_function. red. ir. 
  rw inc_prod_of_prod_target. exists c. ee. am. uf prod_of_product_aux. tv.
Qed.

Lemma source_prod_of_products: forall f f',
  source (prod_of_products f f') = source f.
Proof. ir. uf prod_of_products. rw af_source. tv.
Qed.

Lemma target_prod_of_products: forall f f',
  target (prod_of_products f f') =  prod_of_prod_target f f'.
Proof. ir. uf prod_of_products. rw af_target. tv.
Qed.

Lemma W_prod_of_products: forall f f' i,
  inc i (source f) ->
  W i (prod_of_products f f') =  product (W i f) (W i f').
Proof. ir. uf prod_of_products. rw W_af_function. uf prod_of_product_aux. tv.
  red. ir. rw inc_prod_of_prod_target. exists c. ee. am. 
  uf prod_of_product_aux. tv. am.
Qed.

Lemma prod_of_products_fam_pr: forall f f' x,
  is_function x -> source x = source f ->
  target x = union (prod_of_prod_target f f') ->
  inc (graph x) (productb (graph (prod_of_products f f'))) =
  forall i, inc i (source f) -> 
    (is_pair (W i x) &  inc (P (W i x)) (W i f) &  inc (Q (W i x)) (W i f')).
Proof. ir. 
  assert (Ha:fgraph (graph (prod_of_products f f'))).
  app fgraph_function.  app function_prod_of_products. 
  app iff_eq. ir.  rwi productb_pr H2.
  assert (domain (graph x) = source x). red in H;ee; sy; am.
  cp (W_prod_of_products  f f' H3). ufi W H5.
  nin H2. nin H4. wri H0 H3. nin H6. cp (H6 _ H3). rwi H5 H7.
  rwi inc_product H7. uf W. am. am.
  ir. rw productb_pr. ee. app fgraph_function.  
  assert (is_function  (prod_of_products f f')). app function_prod_of_products.
  red in H3; ee. wr H5. red in H; ee. wr H7. rw H0. rw source_prod_of_products.
  tv. ir. red in H. ee. wri H5 H3. rwi H0 H3.
  cp (W_prod_of_products  f f' H3). ufi W H6. rw H6.
  cp (H2 _ H3). ufi W H7. ee. app product_inc.  am.
Qed.

Lemma axioms_prod_of_function:forall x x' f f',
  is_function f -> is_function f' -> source f = source f' ->
  inc (graph x) (productb (graph f)) ->  inc (graph x') (productb (graph f')) ->
  transf_axioms (fun i => J (W i x) (W i x')) 
     (source f) (union (prod_of_prod_target f f')).
Proof. ir.
  assert (forall c, inc c (source f) ->
    inc (J (W c x) (W c x')) (product (W c f) (W c f'))).
  ir. rwi productb_pr H2. rwi productb_pr H3. ee.
  assert (inc (W c x) (W c f)). red in H. ee.  rwi H10 H4. uf W. app H8. rww H7.
  assert (inc (W c x') (W c f')).  red in H0. rwi H1 H4. uf W. ee.
  rwi H11 H4. app H6. rww H5. app product_pair_inc.  fprops.  fprops. red. ir. 
  apply union_inc with (product (W c f) (W c f')).
  app H4. rw inc_prod_of_prod_target. exists c. intuition.
Qed.

Lemma W_prod_of_function:forall x x' f f' i,
  is_function f -> is_function f' -> source f = source f' ->
  inc x (productb (graph f)) ->  inc x' (productb (graph f')) ->
  inc i (source f) -> 
  V i (prod_of_function x x') =  J (V i x) (V i x').
Proof. ir.  uf prod_of_function. rw create_V_rewrite. tv. 
  rwi productb_pr H2. ee. rw H5. red in H. ee. wrr H8. app fgraph_function.
Qed.

Lemma function_prod_of_function: forall x x' f f',
  is_function f -> is_function f' -> source f = source f' ->
  inc x (productb (graph f)) ->  inc x' (productb (graph f')) ->
  inc (prod_of_function x x')
  (productb (graph (prod_of_products f f'))).
Proof. ir. 
  assert (forall c, inc c (source f) ->
    inc (J (V c x) (V c x')) (product (W c f) (W c f'))).
  ir. rwi productb_pr H2. rwi productb_pr H3. ee.
  assert (inc (V c x) (W c f)). rwi H7 H8.  red in H.  ee. rwi H10 H4.  uf W. 
  app H8.
  assert (inc (V c x') (W c f')). rwi H1 H4.  red in H0. ee. rwi H11 H4. 
  rwi H5 H6. uf W. app H6.  app product_pair_inc. app fgraph_function.
  app fgraph_function. rw productb_pr. uf prod_of_function. ee.
  app create_axioms. aw. assert (is_function (prod_of_products f f')).
  app function_prod_of_products. red in H5. ee. wr H7.
  rw source_prod_of_products. rwi productb_pr H2. ee. rw H8. sy.
  red in H; ee; am. fprops. rw create_domain. ir. rww create_V_rewrite.
  rwi productb_pr H2. ee. rwi H6 H5.  red in H. ee. wri H9 H5. 
  cp (W_prod_of_products f f' H5). unfold W in H10. rw H10.
  ufi W H4. app H4. fprops.  assert (is_function (prod_of_products f f')).
  app function_prod_of_products. fprops.
Qed.

Lemma source_popc:forall f f', 
  source (prod_of_products_canon f f') =
  (product (productb f) (productb f')).
Proof. ir. uf prod_of_products_canon. app af_source. 
Qed.

Lemma target_popc_aux:forall f f', 
 is_function f -> is_function f' -> source f = source f' ->
 productb(L (domain (graph f))
   (fun i => product (V i (graph f)) (V i (graph f')))) =
 productb(graph (prod_of_products f f')).
Proof. ir. app uneq. 
  assert (is_function (prod_of_products f f')). app function_prod_of_products.
  app function_extensionality. app create_axioms.
  fprops. aw. red in H;ee. red in H2. ee. wr H4; wr H6. 
  rw source_prod_of_products. tv. aw.  ir. rww create_V_rewrite.
  red in H. ee.  wri H5 H3. cp (W_prod_of_products f f' H3). sy. ufi W H6. am.
Qed.

Lemma target_popc:forall f f', 
  is_function f -> is_function f' -> source f = source f' ->
  target (prod_of_products_canon (graph f) (graph f')) =
  (productb (graph (prod_of_products f f'))).
Proof. ir. uf prod_of_products_canon. rw af_target.  uf productf.
  app target_popc_aux.
Qed.

Lemma axioms_popc:forall f f', 
 is_function f -> is_function f' -> source f = source f' ->
  transf_axioms(fun w => prod_of_function (P w) (Q w))
  (product (productb (graph f)) (productb (graph f')))
  (productb (graph (prod_of_products f f'))).
Proof. ir. red. ir. rwi inc_product H2. ee.
  app function_prod_of_function. 
Qed.

Lemma W_popc:forall f f' w, 
  is_function f -> is_function f' -> source f = source f' ->
  inc w (product (productb (graph f)) (productb (graph f'))) ->
  W w (prod_of_products_canon (graph f) (graph f')) = 
  prod_of_function (P w) (Q w).
Proof. ir. uf prod_of_products_canon. rww W_af_function. uf productf.
  rww target_popc_aux. app axioms_popc.
Qed.

Lemma bijection_popc:forall f f', 
  is_function f -> is_function f' -> source f = source f' ->
  bijective (prod_of_products_canon (graph f) (graph f')).
Proof. ir. set (g:= prod_of_products_canon (graph f) (graph f')).
  assert (is_function g). uf g.
  uf prod_of_products_canon.  app af_function. 
  uf productf. rww target_popc_aux. app axioms_popc.
  red. ee. red. ee. am. ir. ufi g H3. rwi source_popc H3.  
  ufi g H4. rwi source_popc H4. ufi g H5. rwii W_popc H5. rwii W_popc H5. 
  rwi inc_product H3; rwi inc_product H4. ee.
  assert (forall i, inc i (source f) ->
    J (V i (P x)) (V i (Q x)) = J (V i (P y)) (V i (Q y))). ir.
  rewrite <- W_prod_of_function with (f:=f)(f':=f').
  rewrite <- W_prod_of_function with (f:=f)(f':=f'). rww H5. am. am. am.
  am. am. am. am. am. am. am. am. am.
  assert (P x = P y). rewrite productb_extensionality with (f:=(graph f)).
  ir. red in H. ee. wri H13 H11. cp (H10 _ H11). app (pr1_injective H14). 
  fprops. am. am.
  assert (Q x = Q y). rewrite productb_extensionality with (f:=(graph f')).
  ir. red in H0. ee. wri H14 H12. wri H1 H12. cp (H10 _ H12). 
  app (pr2_injective H15). fprops. am. am.
  app pair_extensionality.
  app surjective_pr6.  uf g. rww target_popc. rw source_popc. ir. 
  rwi productb_pr H3. ee. 
  assert (is_function (prod_of_products f f')). app function_prod_of_products.
  red in H6. ee. wri H8 H4. rwi source_prod_of_products H4.
  set (xa:= BL(fun i => P(V i y)) (source f)(union (target f))).
  set (xb:= BL(fun i => Q(V i y)) (source f)(union (target f'))).
  assert(transf_axioms (fun i  => P (V i y)) (source f) (union (target f))).
  red. ir. cp (W_prod_of_products f f' H9). ufi W H10.
  wri H4 H9. cp (H5 _ H9). rwi H10 H11. rwi inc_product H11. ee.
  apply union_inc with (V c (graph f)). am.
  assert (sub (range (graph f)) (target f)). app range_correspondence.  
  red in H;ee;am. app H14. app inc_V_range.  fprops. red in H; ee. wr H16. 
  wr H4. am.
  assert(transf_axioms (fun i  => Q (V i y)) (source f')(union (target f'))).
  red. ir. wri H1 H10. cp (W_prod_of_products f f' H10). ufi W H11.
  wri H4 H10. cp (H5 _ H10). rwi H11 H12. rwi inc_product H12. ee.
  apply union_inc with (V c (graph f')). am.
  assert (sub (range (graph f')) (target f')). app range_correspondence.  
  red in H0;ee;am. app H15. app inc_V_range.  fprops. red in H0; ee. wr H17. 
  wr H1. wr H4. am.
  assert (is_function xa). uf xa. app af_function. 
  assert (is_function xb). uf xb. app af_function. rww H1.
  assert (inc (graph xa) (productb (graph f))). rw productb_pr.
  ee. fprops. red in H11; ee. wr H14. red in H; ee. wr H16. uf xa. aw.
  ir. assert(domain (graph xa) = source xa). red in  H11; ee; sy; am.
  assert (W i xa = P (V i y)). uf xa. rw W_af_function. tv. am.
  assert (source f = source xa). uf xa; rw af_source. tv. rw H15. wrr H14.
  ufi W H15. rw H15. rwi H14 H13. 
  assert (source f = source xa). uf xa; rw af_source. tv. wri H16 H13.
  wri H4 H13. cp (H5 _ H13). 
  rwi H4 H13. cp (W_prod_of_products f f' H13). ufi W H18. rwi H18 H17.
  rwi inc_product H17. ee. am. fprops. 
  assert (inc (graph xb) (productb (graph f'))). rw productb_pr.
  ee. fprops. red in H12; ee. wr H15. red in H0; ee. wr H17. uf xb. aw.
  ir. assert(domain (graph xb) = source xb). red in  H12; ee; sy; am.
  assert (W i xb = Q (V i y)). uf xb. rw W_af_function. tv. rww H1.
  assert (source f = source xb). uf xb; rw af_source. tv. rw H16. wrr H15.
  ufi W H16. rw H16. rwi H15 H14. 
  assert (source f = source xb). uf xb; rw af_source. tv. wri H17 H14.
  wri H4 H14. cp (H5 _ H14). 
  rwi H4 H14. cp (W_prod_of_products f f' H14). ufi W H19. rwi H19 H18.
  rwi inc_product H18. ee. am. fprops. 
  exists (pair (graph xa) (graph xb)). ee. app product_pair_inc. rww W_popc.
  aw. uf prod_of_function. ap function_extensionality. am.  
  app create_axioms. aw. red in H11. ee. wr H16. uf xa. rw af_source. am.
  ir. rw create_V_rewrite. 
  assert (V x (graph xa) = W x xa). uf W. tv. rw H16. uf xa. 
  rw W_af_function.  assert (V x (graph xb) = W x xb). uf W. tv. rw H17. 
  uf xb. rw W_af_function. sy. app pair_recov.
  cp (H5 _ H15). 
  rwi H4 H15. cp (W_prod_of_products f f' H15). ufi W H19. rwi H19 H18.
  rwi inc_product H18. ee. am. rw H1.  am. wr H4. am. am. wrr H4.
  red in H11. ee. wr H17. uf xa. rw af_source. wr H4. am.
  app product_pair_inc. assert (is_function (prod_of_products f f')).
  app function_prod_of_products. fprops.
Qed.

(** ** EII-5-7 extensions of mappings to products *)
Definition ext_map_prod_aux x  f := fun i=> V (V i x)  (f i).
Definition ext_map_prod In src trg f :=
  BL (fun x => L In (ext_map_prod_aux x f))
    (productf In src ) (productf In trg ).

Definition ext_map_prod_axioms In src trg f :=
  forall i, inc i In -> (fgraph (f i) & domain (f i) = src i &
    sub (range (f i)) (trg i)).

Lemma taxioms_ext_map_prod: forall In src trg f,
  ext_map_prod_axioms In src trg f ->
  transf_axioms (fun x => L In (ext_map_prod_aux x f))
    (productf In src) (productf In trg).
Proof. ir. red. ir.  red in H. rw productf_pr. ee. app create_axioms.
  rw create_domain. tv. aw. ir. rww create_V_rewrite. uf ext_map_prod_aux.
  rwi productf_pr H0. ee.  cp (H _ H1). ee. app H6. app inc_V_range. rw H5.
  rwi H2 H3. app H3.
Qed.

Lemma source_ext_map_prod: forall  In src trg f,
  source (ext_map_prod In src trg f) = (productf In src ).
Proof. ir. uf ext_map_prod. rw af_source. tv. Qed.

Lemma target_ext_map_prod: forall  In src trg f,
  target (ext_map_prod In src trg f) = (productf In trg).
Proof. ir. uf ext_map_prod. rw af_target. tv. Qed.

Lemma function_ext_map_prod: forall In src trg f,
  ext_map_prod_axioms In src trg f -> is_function ( ext_map_prod In src trg f).
Proof. ir. uf ext_map_prod. app af_function. app taxioms_ext_map_prod.
Qed.

Lemma W_ext_map_prod: forall In src trg f x,
  ext_map_prod_axioms In src trg f -> 
  inc x (productf In src) ->
  W x (ext_map_prod In src trg f) = L In (ext_map_prod_aux x f).
Proof. ir. uf ext_map_prod. rww W_af_function. app taxioms_ext_map_prod.
Qed.

Lemma WV_ext_map_prod: forall In src trg f x i,
  ext_map_prod_axioms In src trg f -> 
  inc x (productf In src) -> inc i In ->
  V i (W x (ext_map_prod In src trg f)) = V (V i x)  (f i).
Proof. ir.  rww W_ext_map_prod. rw create_V_rewrite. uf ext_map_prod_aux. tv.
  am.
Qed.

Lemma composable_ext_map_prod: forall  In p1 p2 p3  g f h,
  ext_map_prod_axioms In p1 p2 f ->
  ext_map_prod_axioms In p2 p3 g -> 
  (forall i, inc i In -> h i = fcompose (g i) (f i)) ->
  (forall i, inc i In -> fcomposable (g i) (f i)) ->
  ext_map_prod_axioms In p1 p3 h.
Proof. ir. ee. red in H; red in H0; red. ir.
  cp (H _ H3). cp (H0 _ H3). clear H; clear H0. ee. rw H1.
  app fcompose_axioms. am. rw H1. wr H6. app fcomposable_domain. app H2. am. 
  rw H1. red. ir. rwi frange_inc_rw H8. nin H8. ee. rwi fcompose_ev H9.
  app H5. rw H9.  rw frange_inc_rw. exists  (V x0 (f i)).  ee. rw H0.
  app H7. rw frange_inc_rw.  exists x0.  ee. rwi fcomposable_domain  H8. am.
  app H2. tv. am. tv. am. am. app fcompose_axioms. am.
Qed.

Lemma compose_ext_map_prod: forall  In p1 p2 p3  g f h,
  ext_map_prod_axioms In p1 p2 f ->
  ext_map_prod_axioms In p2 p3 g ->
  (forall i, inc i In -> h i = fcompose (g i) (f i)) ->
  (forall i, inc i In -> fcomposable (g i) (f i)) ->
  compose (ext_map_prod In p2 p3 g)  (ext_map_prod In p1 p2 f) =
  (ext_map_prod In p1 p3 h).
Proof. ir. 
  assert (composable (ext_map_prod In p2 p3 g) 
    (ext_map_prod In p1 p2 f) ).
  red. ee.  app function_ext_map_prod. app function_ext_map_prod.
  rw source_ext_map_prod.  rw target_ext_map_prod. tv.
  assert (is_function (compose(ext_map_prod In p2 p3 g)
    (ext_map_prod In p1 p2 f))).
  app is_function_compose. app funct_extensionality. 
  app function_ext_map_prod.  app (composable_ext_map_prod h  H H0 H1 H2). 
  ir. rwi source_compose H5. rww W_compose.
  set (f1:= (ext_map_prod In p1 p2 f)).
  set (t:= W x f1).
  assert(t =  L In (ext_map_prod_aux x f)). uf t. uf f1. rw W_ext_map_prod. tv.
  am. rwi source_ext_map_prod H5. am.
  set (g1:=  (ext_map_prod In p2 p3 g)).
  assert (W t g1 =  L In (ext_map_prod_aux t g)). uf g1. rw W_ext_map_prod. tv.
  am.  uf t. assert (target f1 = productf In p2). uf f1. 
  rw target_ext_map_prod. tv. wr H7. app inc_W_target. uf f1.
  app function_ext_map_prod.
  set (h1:=  (ext_map_prod In p1 p3 h)).
  assert (W x h1 =  L In (ext_map_prod_aux x h)). uf h1. rw W_ext_map_prod. tv.
  app (composable_ext_map_prod h  H H0 H1 H2).  
  rwi source_ext_map_prod H5. am.
  rw H7; rw H8. app function_extensionality1. ir. rw H6. uf  ext_map_prod_aux.
  rww create_V_rewrite. set (y:= (V x0 x)). rw H1. rw fcompose_ev. tv.
  rw fcomposable_domain. uf y. red in H. ee. cp (H _ H9). ee. rw H11.
  rwi source_ext_map_prod H5. rwi productf_pr H5. ee. app H14. rww H13.
  app H2. am.
Qed.


Lemma injective_ext_map_prod: forall  In p1 p2 f,
  ext_map_prod_axioms In p1 p2 f ->
  (forall i, inc i In -> injective_graph (f i)) ->
  injective (ext_map_prod In p1 p2 f).
Proof.  ir.  red. ir. ee. app function_ext_map_prod. ir. 
  rwi source_ext_map_prod H2.
  rwi source_ext_map_prod H1. rw (productf_extensionality H1 H2). ir.
  cp (WV_ext_map_prod H H1 H4). rwi H3 H5.
  cp (WV_ext_map_prod H H2 H4). rwi H5 H6. cp (H0 _ H4). red in H7. ee.
  app H8. red in H. cp (H i H4). ee. rw H10. rwi productf_pr H1. ee. app H13.
  rww H12. red in H. cp (H i H4). ee. rw H10. rwi productf_pr H2. ee. app H13.
  rww H12.
Qed.


Lemma surjective_ext_map_prod: forall  In p1 p2 f,
  ext_map_prod_axioms In p1 p2 f ->
  (forall i, inc i In -> range (f i) = p2 i) ->
  surjective (ext_map_prod In p1 p2 f).
Proof. ir. app surjective_pr6. app function_ext_map_prod. 
  rw source_ext_map_prod. rw target_ext_map_prod.  ir.
  assert (Ha:= H1).
  rwi productf_pr H1. ee. rwi H2 H3.
  assert (forall i, inc i In -> exists z, inc z (p1 i) & V i y =V z (f i)).
  ir.  cp (H _ H4). ee. cp (H0 _ H4). cp (H3 _ H4). wri H8 H9. 
  rwi frange_inc_rw H9. wrr H6. am.
  set (x:= L In (fun i => choose (fun z => inc z (p1 i) & V i y = V z (f i)))).
  assert (inc x (productf In p1)). rw productf_pr. ee. uf x. app create_axioms.
  uf x. rw create_domain. tv. uf x.  rw create_domain. ir. rw create_V_rewrite.
  set (p:=fun z  => inc z (p1 i) & V i y = V z (f i)). 
  assert (ex p). uf p. app H4. cp (choose_pr H6). set (w:= choose p).
  fold w. fold w in H7.  ufi p H7. ufi w H7. ee.  am. am. exists x. ee. am.
  assert (inc (W x (ext_map_prod In p1 p2 f))(productf In p2)).
  cp (target_ext_map_prod In p1 p2 f). wr H6. app inc_W_target.
  app function_ext_map_prod. rw (productf_extensionality Ha H6). ir.
  rww WV_ext_map_prod.
  set (p:=fun z  => inc z (p1 i) & V i y = V z (f i)). 
  assert (ex p). uf p. app H4. cp (choose_pr H8). set (w:= choose p).
  fold w. fold w in H9.  ufi p H9. ufi w H9.
  assert (V i x = w). uf x. rw create_V_rewrite. tv. am. ee. rw H11.  rww H10.
Qed.

Lemma fun_set_to_prod1: forall F f i,
  fgraph F -> inc i (domain F) ->
  is_function f -> target f = productb F -> 
  (is_function (compose (pr_i F i) f) &
   source (compose (pr_i F i) f) = source f &
   target (compose (pr_i F i) f) = V i F&
   (forall x, inc x (source f) ->  W x  (compose (pr_i F i) f) = V i (W x f))).
Proof. ir.  assert (composable (pr_i F i) f). red. ee. app function_pri.
  am.  rw source_pri. sy; am. ee. app is_function_compose. rw source_compose.
  tv. rw target_compose. rw target_pri. tv. ir. rw W_compose. rw W_pri.  tv.
  am. am. wr H2. app inc_W_target. am. am.
Qed.

Definition fun_set_to_prod src F :=
  BL(fun f => 
       L(domain F)( fun i=> (graph (compose (pr_i F i)
          (corresp src (productb F) f)))))
     (set_of_gfunctions src (productb F))
    (productb (L (domain F) (fun i=> set_of_gfunctions src (V i F)))).

Lemma fun_set_to_prod2: forall src  F f gf,
  fgraph F -> inc gf (set_of_gfunctions src (productb F)) ->
  f = (corresp src (productb F) gf) ->
  (is_function f & target f = productb F & source f = src).
Proof. ir.  cp (set_of_gfunctions_inc H0). nin H2. 
  assert (x = f). rw H1. sy. ee. wr H3; wr H4;wr H5. app corr_propc.
  wr H3. ee. am. am. am.
Qed.

Lemma fun_set_to_prod3: forall src  F,
  fgraph F -> transf_axioms(fun f => 
       L(domain F)( fun i=> (graph (compose (pr_i F i)
          (corresp src (productb F) f)))))
     (set_of_gfunctions src (productb F))
    (productb (L (domain F) (fun i=> set_of_gfunctions src (V i F)))).
Proof. ir. red. ir. set (f:=(corresp src (productb F) c)).
  assert (f = corresp src (productb F) c). tv.
  cp (fun_set_to_prod2 H H0 H1). ee. rw productb_pr. aw. ee. app create_axioms.
  tv. ir. rw create_V_rewrite. rw create_V_rewrite. 
  cp (fun_set_to_prod1 H H5 H2  H3). ee.
  set (h:= (compose (pr_i F i) f)).  wr H4;  wr H7; wr H8; fold h.
  app inc_set_of_gfunctions. am. am. app create_axioms.
Qed.

Lemma fun_set_to_prod4: forall src  F,
  fgraph F -> ( is_function (fun_set_to_prod src F)
   & source (fun_set_to_prod src F) = (set_of_gfunctions src (productb F))
   & target (fun_set_to_prod src F) 
    = (productb (L (domain F) (fun i=> set_of_gfunctions src (V i F))))).
Proof. ir.  uf fun_set_to_prod. ee. app af_function. app fun_set_to_prod3.
  rw af_source. tv. rw af_target. tv.
Qed.

Definition fun_set_to_prod5 F f :=
  ext_map_prod (domain F) (fun i=> source f)(fun i=> V i F) 
  (fun i => (graph (compose (pr_i F i) f))).

Lemma fun_set_to_prod6: forall F f, 
  fgraph F -> is_function f -> target f = productb F -> 
  (is_function (fun_set_to_prod5 F f) & 
    composable (fun_set_to_prod5 F f) (constant_functor (domain F)(source f) )& 
    compose (fun_set_to_prod5 F f) (constant_functor (domain F)(source f)) =f).
Proof. ir. set (tf := fun_set_to_prod5 F f). 
  assert (is_function tf). uf tf. uf fun_set_to_prod5. 
  app function_ext_map_prod.  red. ir. cp (fun_set_to_prod1 H H2 H0 H1).
  ee. fprops. red in H3. ee. wr H8; rw H4. tv. wr H5. red in H3; ee. fprops.
  assert (productb F = target tf). uf tf. uf fun_set_to_prod5.  
  rw target_ext_map_prod. uf productf.  app uneq. app function_extensionality. 
  app create_axioms. aw. ir. rw create_V_rewrite. tv. am. 
  assert(composable tf (constant_functor (domain F)(source f))).
  red. ee. am. uf constant_functor. app af_function. red. ir.
  rw set_of_gfunctions_pr2. uf constant_graph. aw. ee. 
  app create_axioms. tv. red. ir. rwi inc_create_range H5. nin H5. ee. wrr H6.
  uf constant_functor. rw af_target. uf tf. uf fun_set_to_prod5.
  rw source_ext_map_prod. uf productf.
  set (k:= L (domain F) (fun _ : Bset => source f)). 
  assert (domain F = domain k). uf k. aw. rw H4. app eq_product_graphset.
  uf k. app create_axioms. ir. uf k. rw create_V_rewrite. tv. rww H4. 

  ee. am. am. sy. app funct_extensionality.  app is_function_compose.
  rw target_compose. uf tf. uf fun_set_to_prod5.  rw target_ext_map_prod.
  uf productf.  rw H1. app uneq. app function_extensionality. app create_axioms.
  aw. ir.  rw create_V_rewrite. tv. am. ir. rw W_compose.
  uf constant_functor. rw W_af_function. 
  assert (inc (W x f) (productb F)).  wr H1. app inc_W_target.
  assert (inc (W (constant_graph (domain F) x) tf) (productb F)). rw H3.
  app inc_W_target. uf tf. uf fun_set_to_prod5. rw source_ext_map_prod.
  rw productf_pr.  uf constant_graph. ee. app create_axioms. aw. 
  rw create_domain. ir.  rw create_V_rewrite. am. am. 
  rw (productb_extensionality H  H6 H7). ir.  uf tf. uf fun_set_to_prod5.
  rw W_ext_map_prod. rw create_V_rewrite.  uf ext_map_prod_aux.
  uf constant_graph. rw create_V_rewrite.
  assert (V x (graph (compose (pr_i F i) f)) =  W x (compose (pr_i F i) f)).
  uf W; tv. rw H9. rw W_compose.
  rw W_pri. tv. am. am. am. red. ee. app function_pri. am. rw source_pri.
  sy. am. am. am. am. red. ir. cp (fun_set_to_prod1 H H9 H0 H1). ee. 
  fprops.  red in H10; ee. wr H15.  rw H11. tv. red in H10. wr H12.
  app range_correspondence.  ee. am. rw productf_pr. uf constant_graph. ee. 
  app create_axioms. aw. rw create_domain. ir. rw create_V_rewrite. am. am. 
  red.  ir.  uf constant_graph. uf set_of_gfunctions. Ztac. app powerset_inc.
  red. ir. ufi L H7. awi H7. nin H7.  ee.  wr H8. 
  app product_pair_inc.  ee. app create_axioms. aw. am. am. 
  uf constant_functor. aw.
Qed.

Lemma fun_set_to_prod7: forall src F f g ,
  fgraph F -> 
  (forall i, inc i (domain F) -> inc (f i) (set_of_gfunctions src (V i F))) ->
  g = ext_map_prod (domain F) (fun i=> src)(fun i=> V i F) f ->
  (forall i, inc i (domain F) ->
    f i = graph (compose (pr_i F i) (compose g (constant_functor (domain F) src) ))).
Proof. ir.  assert (is_function (constant_functor (domain F) src)).
  uf constant_functor. app af_function. red. ir. rw set_of_gfunctions_pr2. 
  uf constant_graph. aw. ee. app create_axioms. tv. 
  red. ir. rwi inc_create_range H4. nin H4. ee. wrr H5.
  assert (is_function g). rw H1. app function_ext_map_prod. red. ir.
  cp (H0 _ H4). cp (set_of_gfunctions_inc H5).  nin H6. ee. wr H9. fprops.
  wr H9. red in H6; ee.  wr H11. am. wr H8. wr H9. app range_correspondence.
  red in H6; ee. am. assert (composable g (constant_functor (domain F) src)).
  red. ee. am. am. uf constant_functor. rw af_target. rw H1. 
  rw source_ext_map_prod. uf productf. 
  set (k:= (L (domain F) (fun _ : Bset => src))). assert (domain k = domain F).
  uf k. aw. wr H5. assert (fgraph k). uf k. app create_axioms.
  assert (forall i, inc i(domain k) -> V i k = src). ir. uf k.
  rw create_V_rewrite. tv.  wr H5.  am. app (eq_product_graphset H6 H7).
  assert (is_function (compose g (constant_functor (domain F) src))).
  app is_function_compose.
  assert (composable (pr_i F i) (compose g (constant_functor (domain F) src))).
  red. ee. app function_pri. am. rw target_compose. rw source_pri.
  rw H1. rw target_ext_map_prod. uf productf. app uneq. 
  app function_extensionality. app create_axioms. aw. ir. rw create_V_rewrite.
  tv. am. assert(is_function (compose (pr_i F i) 
  (compose g (constant_functor (domain F) src)))).
  app is_function_compose. cp (H0 _ H2). cp (set_of_gfunctions_inc H9). 
  nin H10. ee. 
  assert (x = compose (pr_i F i) (compose g (constant_functor (domain F) src))).
  app funct_extensionality. ir. rww W_compose. rww W_compose.
  uf constant_functor. rw W_af_function. rw W_pri.
  rw H1. rw W_ext_map_prod. rw create_V_rewrite. uf ext_map_prod_aux.
  uf W. wr H13. assert (x0 = (V i (constant_graph (domain F) x0))).
  rw V_constant_graph. tv. am. wr H15. tv. am.  red. ir.
  cp (H0 _ H15). cp (set_of_gfunctions_inc H16). nin H17. ee. wr H20. fprops.
  wr H20. red in H17; ee. wrr H18. sy; am. wr H19. red in H17. ee. wr H20. 
  app range_correspondence. rw productf_pr.  ee. uf constant_graph.
  app create_axioms. uf constant_graph. aw. uf constant_graph. aw. ir.
  rw create_V_rewrite. wrr H11. am. am. am. uf constant_graph. 
  assert (target g =  (productb F)). rw H1.  rw target_ext_map_prod. 
  uf productf. app uneq. 
  app function_extensionality. app create_axioms. aw. ir. rw create_V_rewrite.
  tv. rwi create_domain H15.   am. wr H15. app inc_W_target. rw H1.
  rw source_ext_map_prod. rw productf_pr. ee. app create_axioms.
  rw create_domain. tv.  rw create_domain. ir. rw create_V_rewrite.  wrr H11.
  am. red. ir. uf constant_graph.
  set (k:= BL(fun _ : Bset => c)(domain F) src). 
  assert (is_function  k). uf k. app af_function. red.  ir. am.
  app (inc_set_of_gfunctions H16). wrr H11. uf constant_functor.
  rw af_source. wr H11. am. rw source_compose. uf constant_functor.
  rw af_source. wr H11. am. wr H14. sy. am.
Qed.


Lemma fun_set_to_prod8: forall src  F,
  fgraph F -> bijective (fun_set_to_prod src F).
Proof. ir. cp (fun_set_to_prod4 src H). ee.
  red. ee. red. ee. am. rw H1. ir. cp (set_of_gfunctions_inc H3).  
  cp (set_of_gfunctions_inc H4).  nin H6; nin H7.  ee. wr H13; wr H10. 
  assert (x0 = x1). ufi fun_set_to_prod H5.
  cp (fun_set_to_prod3 (src:=src) H). 
  rwii W_af_function H5. rwii W_af_function H5. clear H14.  
  app funct_extensionality.  rw H11; rw H8; tv. rw H12; rw H9; tv. ir. 
  assert (inc (W x2 x0) (productb F)). wr H12. app inc_W_target.
  assert (inc (W x2 x1) (productb F)). wr H9. app inc_W_target. rw H8. wrr H11. 
  rw (productb_extensionality H H15 H16). ir.
  cp (fun_set_to_prod1 H H17 H7 H9). cp (fun_set_to_prod1 H H17 H6 H12).
  ee. wrr H25. wrr H22.
  assert (compose (pr_i F i) x0 = compose (pr_i F i) x1). 
  assert (x0 = corresp src (productb F) x).
  sy. wr H11; wr H12; wr H13.  rww corr_propc. wri H26 H5.
  assert (x1 = corresp src (productb F) y).
  sy. wr H8; wr H9; wr H10.  rww corr_propc. wri H27 H5.
  assert (V i  (L (domain F) (fun i => graph (compose (pr_i F i) x0))) =
    V i (L (domain F) (fun i  => graph (compose (pr_i F i) x1)))).
  rw H5. tv. rwi create_V_rewrite H28. rwi create_V_rewrite H28. 
  wr (corr_propc(compose (pr_i F i) x0)).
  wr (corr_propc(compose (pr_i F i) x1)). 
  rw H23; rw H20; rw H24; rw H21; rw H28. rw H11; rw H8; tv. am. am. rww H26.
  rw H8; wr H11. am.  rww H14.
  app surjective_pr6. uf fun_set_to_prod. rw af_target. rw af_source. ir.
  rwi productb_pr H3. ee. rwi create_domain H4.
  set (x:= BL (fun u=> L (domain F) (fun i=> V u (V i y))) src (productb F)).
  assert (is_function x). uf x. app af_function. red. ir. rw productb_pr.
  ee.  app create_axioms. aw. aw. ir.  rw create_V_rewrite. 
  rwi H4 H5. cp (H5 _ H7). rwi create_V_rewrite H8. 
  cp (set_of_gfunctions_inc H8). nin H9. ee. wr H11.
  assert (V c (V i y) = W c x0). uf W. rww H12. rw H13. app inc_W_target. 
  rww H10. am. am. am. 
  assert (inc (graph x) (set_of_gfunctions src (productb F))).
  assert (source x = src). uf x. rw af_source. tv. wr H7.
  assert (target x = (productb F)). uf x. rw af_target. tv. wr H8.
  app inc_set_of_gfunctions.
  exists (graph x). ee. am. rw W_af_function. app function_extensionality.
  app create_axioms.  aw. ir. rw create_V_rewrite. cp (H5 _ H8).
  rwi create_V_rewrite H9.
  assert (x = corresp src (productb F) (graph x)).
  assert (source x = src). uf x. rw af_source. tv. wr H10.
  assert (target x = (productb F)). uf x. rw af_target. tv. wr H11.
  sy. app corr_propc. cp (set_of_gfunctions_inc H9). nin H11. ee.
  wr H10. assert (x1 = compose (pr_i F x0) x). app funct_extensionality.
  app is_function_compose. red. ee. app function_pri. wrr H4. am. 
  rw source_pri. uf x. simpl. tv. ir.  rw W_compose. rw W_pri. uf x.
  rw W_af_function. rw create_V_rewrite. uf W. rw H14. tv. wrr H4.
  red. ir.  rw productb_pr.  ee. app create_axioms. aw. aw. ir. 
  rw create_V_rewrite. rwi H4 H5. cp (H5 _ H17). rwi create_V_rewrite H18.
  cp (set_of_gfunctions_inc H18). nin H19. ee.
  wr H22. wr H21. change (inc (W c x3) (target x3)).  app inc_W_target. rww H20.
  am. am. am. wrr H12. am. wrr H4. 
  assert (target x = productb F). uf x.  rw af_target. tv. wr H16. 
  app inc_W_target. assert(source x = source x1). uf x.  rw af_source. sy. am.
  rww H17. red. ee. app function_pri. wrr H4. am. rw source_pri. uf x.
  rw af_target. tv. uf x. rw af_source. wrr H12. wr H15. sy. am. wrr H4. 
  wrr H4. app (fun_set_to_prod3 H (src:=src)). am. app create_axioms.
Qed.

End Bproduct. 

Export Bproduct.   

