(** * Theory of Sets  EIII-6  Infinite sets
  Copyright INRIA (2009-2013) Apics; Marelle Team (Jose Grimm).
*)

(* $Id: sset10.v,v 1.92 2015/03/05 17:27:36 grimm Exp $ *)

Require Import ssreflect ssrfun ssrbool eqtype ssrnat.
Require Export sset9.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Module InfiniteSets.
(** ** EIII-6-1 The set of natural integers *)

(** ** EIII-6-2 Definition of mappings by induction *)

Definition induction_defined1 (h: fterm2) a p :=
   induction_defined0 (fun n x => Yo (n <c p) (h n x) a) a.

Definition induction_defined_set s a E:= 
  change_target_fun (induction_defined s a) E.

Definition induction_defined_set0 h a E:= 
  change_target_fun (induction_defined0 h a) E.
 
Definition induction_defined_set1 h a p E:= 
  change_target_fun (induction_defined1 h a p) E.
 
Lemma induction_defined_pr1 h a p (f := induction_defined1 h a p):
    natp p ->
    [/\  source f = Nat, surjection f, 
      Vf f \0c = a,
      (forall n,  n <c p -> Vf f (csucc n) = h n (Vf f n)) &
      (forall n, natp n -> ~ (n <=c p) -> Vf f n = a)].
Proof.
move=> pN.
move: (induction_defined_pr0 (fun n x => Yo (n <c p) (h n x) a) a).
rewrite /f /induction_defined1.
set g := (induction_defined0 _ a).
move=> [pa pb pc pd]; split => //;  move => n np.
  by rewrite (pd _ (NS_le_int (proj1 np) pN)); Ytac0.
move => cnp.
have nnz: (n <> \0c) by dneg nz; rewrite nz; fprops.
move: (cpred_pr np nnz) => [pnN nsp].
rewrite nsp (pd _ pnN) Y_false //; apply /(cleSltP pnN); ue.
Qed.
 
Lemma integer_induction1 h a p: natp p ->
  exists! f, [/\ source f = Nat, surjection f,
    Vf f \0c = a ,
     (forall n,  n <c p -> Vf f (csucc n) = h n (Vf f n)) &
     (forall n, natp n -> ~ (n <=c p) -> Vf f n = a)].
Proof. 
move=> pN; exists (induction_defined1 h a p); split.
  by apply: induction_defined_pr1.
have [sx sjx x0 xs xl] := (induction_defined_pr1 h a pN).
move => y  [sy sjy y0 ys yl].
apply function_exten4=>//; first by ue.
rewrite sx; apply: Nat_induction; first by ue.
move=> n nN eq.
case: (p_or_not_p (n <c p)) => np.  
  by rewrite  (xs _ np) (ys _ np) eq.
have snp: (~ (csucc n) <=c p) by  move /cleSltP; fprops. 
have snN:= (NS_succ nN).
by rewrite  (xl _ snN snp) (yl _ snN snp).
Qed.

Lemma integer_induction_stable E g  a: 
  inc a E -> (forall x, inc x E -> inc (g x) E) ->
  sub (target (induction_defined g a)) E.
Proof. 
move=> eA agE t tt.
have [sf sjf f0 fn]:=(induction_defined_pr g a).
have [x xsf <-] := ((proj2 sjf) _ tt).
move:x xsf; rewrite sf;apply: Nat_induction; first by ue.
by move=> n nN WE; rewrite fn //; apply: agE.
Qed.

Lemma integer_induction_stable0 E h a: 
  inc a E -> (forall n x, inc x E -> natp n -> inc (h n x) E) ->
  sub (target (induction_defined0 h a)) E.
Proof. 
move=> aE ahE t tt.
have [sf sjf f0 fn] := (induction_defined_pr0 h a).
have [x xsf <-] :=(proj2 sjf _ tt).
move: x xsf;rewrite sf;apply: Nat_induction; first by ue.
by move=> n nN WE; rewrite fn //; apply: ahE.
Qed.

Lemma integer_induction_stable1 E h a p: 
  natp p ->
  inc a E -> (forall n x, inc x E -> n <c p -> inc (h n x) E) ->
  sub (target (induction_defined1 h a p)) E.
Proof. 
move=> pN aE ahE t tt.
have [sf sjf f0 fs lf] := (induction_defined_pr1 h a pN).
have [x xsf <-] :=(proj2 sjf _ tt).
move: x xsf; rewrite sf;apply: Nat_induction; first by ue.
move=> n nN WE.
case:(p_or_not_p (n <c p)) => np.  
  by rewrite (fs _ np); apply: ahE.
rewrite lf; fprops; move /cleSltP; fprops.
Qed.

Lemma change_target_pr f E (g:= change_target_fun f E): 
   function f -> sub (target f) E
  -> (function_prop g (source f) E 
     /\ forall x, inc x (source f) -> Vf g x = Vf f x).
Proof.
move => ff stf; rewrite /g/change_target_fun; aw; split => //.
  have h := (sub_trans (f_range_graph ff) stf).
  have [_ pc pb] := ff.
  split;aw; exact: function_pr.
move=> x xsf; rewrite /Vf; aw.
Qed.

Lemma induction_defined_pr_set  E g a (f := induction_defined_set g a E):
    inc a E -> (forall x, inc x E -> inc (g x) E) -> 
     [/\ function_prop f Nat E, Vf f \0c = a & 
     forall n, natp n -> Vf f (csucc n)  = g (Vf f n)].
Proof. 
move => aE ag. 
have st:= (integer_induction_stable aE ag).
have [sf [ff _] f0 fs] := (induction_defined_pr g a).
move: (change_target_pr  ff st); rewrite -/(induction_defined_set g a E) -/f.
rewrite sf; move => [pa pd].
split => //; first by rewrite -f0 pd //; exact: NS0.
by move => n nb; rewrite (pd _ (NS_succ nb)) (fs _ nb) -pd.
Qed.

Lemma induction_defined_pr_set0 E h a (f := induction_defined_set0 h a E):
    inc a E -> (forall n x, inc x E -> natp n -> inc (h n x) E) ->
     [/\ function_prop f Nat E, Vf f \0c = a & 
     forall n, natp n -> Vf f (csucc n) = h n (Vf f n)].
Proof. 
move=> aE ag. 
have st := (integer_induction_stable0 aE ag).
have [sf [ff _] f0 fs] := (induction_defined_pr0 h a).
move: (change_target_pr  ff st); rewrite -/(induction_defined_set0 h a E) -/f.
rewrite sf; move => [pa pd].
split => //; first by rewrite -f0 pd //; exact: NS0.
by move => n nb; rewrite (pd _ (NS_succ nb)) (fs _ nb) -pd.
Qed.

Lemma induction_defined_pr_set1  E h a p (f := induction_defined_set1 h a p E):
    natp p ->
    inc a E -> (forall n x, inc x E -> n <c p -> inc (h n x) E) ->
    [/\ function_prop f Nat  E, Vf f \0c = a, 
      (forall n, n <c p -> Vf f (csucc n) = h n (Vf f n)) &
      (forall n, natp n -> ~ (n <=c p) -> Vf f n = a)].
Proof.
move=> pN aE ag. 
have st := (integer_induction_stable1 pN aE ag).
have [sf [ff _] f0 fs1 fs2] := (induction_defined_pr1 h a pN).
move: (change_target_pr  ff st); rewrite -/(induction_defined_set1 h a E) -/f.
rewrite sf; move => [pa pd].
split => //; first by  rewrite -f0 pd //; exact: NS0.
  move => n np.
  have nN: natp n by Nat_tac.
  by rewrite (pd _ nN) (pd _ (NS_succ nN)) fs1. 
by move => n nN np; rewrite (pd _ nN) (fs2 _ nN np). 
Qed.

(** ** EIII-6-3 Properties of infinite cardinals *)

Lemma cardinal_Nat: cardinal Nat = Nat.
Proof. apply: card_card; apply: CS_omega. Qed.

Lemma card_Nat_nz: cardinal Nat <> \0c.
Proof. 
by move => h; move:NS1; rewrite /natp - cardinal_Nat h; move/in_set0.
Qed.

Lemma infinite_gt1 x: infinite_c x -> \1c <c x.
Proof. 
by move=> h; apply: finite_lt_infinite => //; fprops.
Qed.

Lemma infinite_ge2 x: infinite_c x -> \2c <=c x.
Proof. 
move => h; apply: finite_le_infinite => //; fprops.
Qed.


Lemma infinite_greater_countable1 E:
  infinite_set E ->  (cardinal Nat) <=c (cardinal E).
Proof.
move=>iE;move;split;fprops.
rewrite cardinal_Nat; apply/omega_P1 => //.
apply: OS_cardinal; fprops.
Qed.

Lemma infinite_greater_countable E:
  infinite_set E -> exists F, sub F E /\ cardinal F = cardinal Nat.
Proof. 
move=> /infinite_greater_countable1 /eq_subset_cardP1 [F e ef]. 
exists F;split => //; by symmetry; apply /card_eqP.
Qed.

Lemma equipotent_N2_N: (coarse Nat) \Eq Nat.
Proof.
pose g2 a := (binom (csucc a) \2c).
pose mi n m := n +c (g2 (n +c m)).
have g2N: forall a, natp a -> natp (g2 a).
   move => a aN; apply: NS_binom =>//; fprops.
have miN: forall n m, natp n -> natp m -> natp (mi n m).
  move=> n m nN mN; rewrite /mi; fprops.
have ha: forall n m, natp n -> natp m -> 
    (g2 (n +c m)) <=c (mi n m).
  move=> n m nN mN;  rewrite /mi; set t:=  (g2 (n +c m)). 
  rewrite csumC; apply: Nsum_M0le; apply:(g2N _ (NS_sum nN mN)).
have hc: forall n, natp n -> g2 (csucc n) = (g2 n) +c (csucc n).
  by move=> n nN; rewrite /g2 (binom_2plus0 (NS_succ nN)). 
have hb: forall n m, natp n -> natp m -> 
  (mi n m) <c (g2 (csucc (n +c m))).
  move=>  n m nN mN.
  move: (NS_sum nN mN) => sN.
  rewrite (hc _ sN) /mi; set t:= (g2 (n +c m)).
  have aux: natp t by rewrite /t; apply: g2N; fprops.
  rewrite csumC; apply: (csum_Mlelt  aux (NS_succ sN)); fprops.
  apply /(cltSleP sN); apply: Nsum_M0le; fprops.
exists (Lf (fun z => (mi (P z) (Q z))) (coarse Nat) Nat); hnf;aw;split => //.
apply: lf_bijective; first by move => t /setX_P  [_ pN qN]; apply:miN.
  move=> u v /setX_P [pu puN quN] /setX_P [pv pvN qvN] sm.
  suff ss: (P u) +c (Q u) = (P v) +c (Q v).
    move: sm; rewrite /mi ss; set (w:=  g2 ((P v) +c (Q v))).
    have wN: natp w by apply: (g2N _ (NS_sum pvN qvN)).
    move=> ss2; move: (csum_eq2r wN puN pvN ss2) => sp.
    rewrite sp in ss; apply: (pair_exten pu pv sp).  
    apply: (csum_eq2l pvN quN qvN ss).
  move: (ha _ _ puN quN)(hb _ _ puN quN)(ha _ _ pvN qvN)(hb _ _ pvN qvN).
  rewrite sm; move=> r1 r2 r3 r4.
  move: (cle_ltT r1 r4)(cle_ltT r3 r2)=> r5 r6.
  have ls2: forall z, natp z -> \2c <=c  (csucc (csucc z)).
    move => z zN; rewrite - succ_one - succ_zero.
    apply:cleSS; apply:cleSS; fprops.
  have n2z:=card2_nz.
  suff gi: forall x y, natp x -> natp y  -> 
      (g2 x) <c (g2 (csucc y)) -> x <=c  y.
    move: (NS_sum puN quN) (NS_sum pvN qvN) => sun svn.
    apply:  (cleA (gi _ _ sun svn r5)(gi _ _ svn sun r6)).
  move=> x y xN yN; rewrite /g2.
  move: (NS_succ xN) (NS_succ yN) (NS_succ (NS_succ yN)) => qa qb qc.
  rewrite - (binom_monotone2 NS2 qa qc n2z (ls2 _ xN) (ls2 _ qb)). 
  by move /(cltSleP qb)/(cleSSP (CS_Nat xN)(CS_Nat yN)).
move=> y yN.
have g2z: g2 \0c = \0c.
  rewrite /g2 succ_zero (binom_bad NS1 NS2) //.  
  rewrite - succ_one; apply: (cltS NS1); fprops.
case: (equal_or_not y \0c) => yz.
  move: NS0 => bz.
  rewrite yz; exists (J \0c \0c); first by rewrite /coarse;fprops. 
  by rewrite pr1_pair pr2_pair /mi (Nsum0l bz) g2z (Nsum0l bz).
have pc: exists2 k, natp k & y <c (g2 k). 
  exists (csucc y);fprops.
  have ysy:= (cltS yN).
  rewrite /g2 binom_2plus0; last by fprops.
  have syN:=(NS_succ yN).
  have ax: (csucc y) <=c (binom (csucc y) \2c +c (csucc y)).
    rewrite csumC; apply: (Nsum_M0le _  syN). 
  apply:clt_leT ysy ax.
case: (wleast_int_prop pc); first by rewrite g2z; move=> /clt0. 
move => [x [xN psx npx]].
have bp:= (hc _ xN).
have p1: y <=c ((g2 x) +c x).
  move: psx; rewrite bp (csum_nS _ xN).
  by move /(cltSleP (NS_sum (g2N _ xN) xN)).
have p2: (g2 x) <=c y.
  have cg2: cardinalp (g2 x) by move: (g2N _ xN); fprops.
  have cy: cardinalp y by fprops.
  by case: (cleT_el cg2 cy).
have g2Nx:= (g2N _ xN).
move: (cdiff_pr p2); set n:= (y -c (g2 x)).
move=> p3;rewrite -p3 in p1.
have nN:natp n by apply: NS_diff.
have lenx:= (csum_le2l g2Nx nN xN p1).
move: (cdiff_pr lenx); set m:= (x -c n).
move=> p4;rewrite -p4 in p3.
have mN: natp m by apply: NS_diff.
exists (J n m); first by apply: setXp_i. 
by rewrite /mi pr1_pair pr2_pair csumC p3.
Qed.

Theorem equipotent_inf2_inf E: infinite_c E ->
  E ^c \2c = E.
Proof. 
move=> ciE; move: (ciE) => [cE ioE].
have cEE:cardinal E = E by apply: card_card.
have isE:infinite_set E by hnf; rewrite cEE.
set (base := sub_functions E (coarse E)).
pose pr z :=  injection z /\ range (graph z) = coarse (source z).
have [psi [psibase prpsi ispsi]]:
  exists psi, [/\ inc psi base, pr psi & infinite_set (source psi)].
  have [F [FE cFN]] := (infinite_greater_countable isE).
  have [f [bf sf tf]]: (F \Eq (coarse F)). 
    move:cFN => /card_eqP eFN; eqsym.
    eqtrans (coarse Nat); [rewrite /coarse; fprops | eqtrans Nat ].
      apply: equipotent_N2_N.
    by eqsym.
  have ff: function f by fct_tac. 
  have pm: (sub (target f) (coarse E)) by rewrite tf; apply: setX_Slr.
  move: (change_target_pr ff pm); set y := (change_target_fun _ _).
  move => [[fy sy ty] yv].
  exists y; rewrite /base sy sf; split => //.
      by apply /sfun_set_P; split => //;  rewrite sy sf.
    split; last by rewrite sy sf - tf -(surjective_pr3 (proj2 bf)) corresp_g.
    split => // u v; rewrite sy => uf vf. 
    by rewrite yv // yv //; apply: (bij_inj bf).
  hnf; rewrite cFN;apply: infinite_Nat.
set (odr := opp_order (extension_order E (coarse E))).
set (ind := Zo base (fun z => pr z /\ sub (graph psi) (graph z))).
have [oe se]:= (extension_osr E (coarse E)).
have [oo soo] := (opp_osr oe). 
have so: sub ind (substrate odr) by rewrite soo se;apply: Zo_S.
have [oi soi] := (iorder_osr oo so).
have ii: inductive (induced_order odr ind). 
  move => X; rewrite (iorder_sr oo so) => Xind. 
  have aux1 := (sub_trans Xind so).
  rewrite (iorder_trans _ Xind)/total_order; aw; move => [iot tot]. 
  case: (emptyset_dichot X)=> neX.  
    exists psi; split; first by aw; apply: Zo_i =>//; fprops. 
    by rewrite neX; move=> u /in_set0.
  have [rx rxX]:= neX.
  have Hla:forall i, inc i X -> function i.
    by move => i /Xind  /Zo_hi  [[[ok _] _] _].
  have Hlb:forall i, inc i X -> target i = coarse E.
    by move => i /Xind /Zo_S /sfun_set_P [_ _ h].
  have Hlc: forall x y, inc x X -> inc y X -> gle odr x y \/ gle odr y x.
    move=> x y xX yX; case: (tot _ _  xX yX) => h;move: (iorder_gle1 h); fprops.
  set si := Lg X source.
  have Hd: forall i j, inc i (domain si) -> inc j (domain si) -> 
      agrees_on ((Vg si i) \cap (Vg si j)) i j.
    rewrite /si; bw;move=> i j iX jX; bw.
    split; [ apply: subsetI2l |  by apply: subsetI2r |].
    move=> a /setI2_P [asi asj].
    by case: (Hlc _ _ iX jX)=> h; [ | symmetry ];apply: (extension_order_pr h). 
  have He:forall i, inc i (domain si) -> function_prop i (Vg si i) (coarse E).
    rewrite /si; bw; move=> i iX; bw; split;fprops.
  move: (extension_covering He Hd).
  set x := common_ext _ _ _; rewrite/si; bw;move=> [[fx sx tx] agx _ etc].
  have xb: inc x base.
    rewrite /base; bw; apply /sfun_set_P; split => // t; rewrite sx.
    move /setUb_P; bw; move=> [y yX]; move: (Xind _ yX). 
    move => /Zo_S /sfun_set_P [_ pa _ ] ; bw; apply: pa.
  have Hg: forall i, inc i X -> sub (graph i) (graph x). 
    move=> i iX; rewrite sub_function =>//; last by apply: Hla.
    move: (etc _ iX); rewrite /agrees_on; bw;move=> [p1 p2 p3]; split => //.
    move=> a asi; symmetry; exact: p3.
  have sgp:(sub (graph psi) (graph x)).
    apply: sub_trans (proj2 (Zo_hi (Xind _ rxX))) (Hg _ rxX).
  have injx: injection x.
     split=>//; move => u v;rewrite sx.
     move => /setUb_P1  [a aX usa]/setUb_P1 [b bX vsb].
     have [z [zX usz vsz]]:
        (exists z, [/\ inc z X, inc u (source z) & inc v (source z)]).
      by case: (Hlc _ _  aX bX);
         move /igraph_pP /extension_orderP => [_ _ /extends_Ssource h];
          [exists  b | exists a ]; split => //; apply: h.
     move: (etc _ zX); rewrite /agrees_on; bw;move => [_ _ aux].
     rewrite (aux _ usz) (aux _ vsz).
     by move: (Zo_hi (Xind _ zX)) => [[[_ iz] _] _]; apply: iz.
  have rgx: (range (graph x) = coarse (source x)). 
    set_extens a. 
      move /(range_fP (proj1 injx)) => [b bsx aW]; move: bsx. 
      rewrite sx; move /setUb_P1=> [c cy bsc].
      move: (etc _ cy); rewrite/agrees_on;bw;move => [_ _ aux].
      rewrite (aux _ bsc) in aW.
      move: (Xind _ cy) => /Zo_hi [[ic rc] _].
      have: inc a (coarse (source c)).
        rewrite -rc aW; Wtac; move: bsc; bw.
      move /setX_P => [pa Ps Qs]; apply /setX_P.
      split => //; apply /setUb_P1; ex_tac.
    move => /setX_P [pa]; rewrite sx. 
    move /setUb_P1 => [b bX psb] /setUb_P1 [c cX qsc].
    have [z [zX usz vsz]]:
        (exists z, [/\ inc z X, inc (P a) (source z) & inc (Q a) (source z)]).
      by case: (Hlc _ _  bX cX);
         move /igraph_pP /extension_orderP => [_ _ /extends_Ssource h];
         [exists  c | exists b ]; split => //; apply: h.
    have :inc a (coarse (source z)) by apply /setX_i.
    move: (Xind _ zX) => /Zo_hi  [[ic rc] _].
    rewrite -rc; move  /(range_fP (proj1 ic)) => [d dsz ->].
    move: (etc _ zX); rewrite/agrees_on;bw;move => [_ _ aux].
    rewrite -(aux _ dsz); apply /(range_fP fx); exists d => //.
    rewrite sx; apply /setUb_P1; ex_tac.
  have xi: (inc x ind) by apply: Zo_i.
  exists x;split; first by aw.
  move=> z zX; move: (Xind _ zX) => zi.
  move: (zi) => /Zo_S zb.
  apply /iorder_gleP => //; apply /igraph_pP /extension_orderP;split => //.
  move: zb => /sfun_set_P [_ _ tz]; hnf; rewrite tx tz; split; fprops.
move: (Zorn_lemma oi ii) => [x].
rewrite /maximal (iorder_sr oo so); move => [xi aux].
have maxp: forall u, inc u ind -> extends u x -> u = x.
  move=> u ui ue;apply: aux; apply /iorder_gleP => //.
  move: (Zo_S ui) (Zo_S xi)=> ub xb.
  apply /igraph_pP /extension_orderP;split => //.
clear aux ii.
move: xi => /Zo_P [] /sfun_set_P [fx ssxE txE] [[ijx rgx] sgx].
set (F := source x); set b:= cardinal F. 
have binf:infinite_c b.
  have ss: (sub (source psi) (source x)).
    have fspi: function psi by move: prpsi=> [ip _]; fct_tac.
    move: (domain_S sgx);  rewrite?  f_domain_graph //.
  move: (sub_smaller ss); rewrite -/F -/b => p1. 
  by apply: (ge_infinite_infinite _ p1); apply /infinite_setP.
have bsq: (b = b *c b). 
  have a1:  (F \Eq (coarse F)) by rewrite /F -rgx; apply: equipotent_range.
  rewrite cprod2_pr1 /b; apply /card_eqP; eqtrans (F \times F).
case: (equal_or_not b E) => bE; first by rewrite -bE cpowx2 =>//. 
have cb: cardinalp b by rewrite /b; fprops.
have bltE: (b <c E) by split ; [rewrite - cEE; exact: (sub_smaller ssxE) |]. 
have Hc: b *c \2c = b +c b by rewrite cprodC two_times_n.
have cb2b: b <=c (b *c \2c)  by rewrite Hc; apply: csum_M0le.
have: (b *c \2c) <=c  (b *c b).
  apply: cprod_Meqle; fprops; apply: (infinite_ge2 binf).
rewrite - bsq => b2b2. 
move: (cleA cb2b b2b2); rewrite Hc => b2b {b2b2 cb2b Hc}.
have [/cardinal_le_aux1 lbEF _]: (cardinal b <c (cardinal(E -s F))). 
  case: (cleT_el (CS_cardinal (E -s F)) (CS_cardinal b)); last by done.
  move /(csum_Meqle b). 
  by rewrite (card_card cb) (cardinal_setC ssxE) -b2b cEE => /(cltNge bltE).
move:lbEF; move/(eq_subset_cardP b _) => [Y syc /card_eqP cY].
set (Z:= F \cup Y). 
set (a1:= coarse F); set (a2 := F \times Y); set (a3 := Y \times Z).
move: (card_card cb) => ccb.
case: (infinite_dichot1 _ binf). 
suff: (Y= emptyset).
  move => eq; move: cY; rewrite eq cardinal_set0 ccb => ->; exact finite_0.
have dFY: disjoint F Y.
  by apply: disjoint_pr; move=> u uF uy; move: (syc _ uy) => /setC_P [_].
move: (csum2_pr5 dFY). 
rewrite -/Z -/b - (csum2_pr2a _ Y) - (csum2_pr2b _ Y) - cY ccb - b2b => cZ.
have aux: forall u v, cardinal u *c (cardinal v) = u *c v.
  by move => u v; apply: cprod2_pr2; fprops; rewrite double_cardinal. 
have ca1: cardinal (a2 \cup a3) = b. 
  have di: disjoint a2 a3.
    apply: disjoint_pr; move => t /setX_P [_ af _] /setX_P [_ ay _].
    by move: (syc _ ay) => /setC_P [_].
  by rewrite (csum2_pr5 di) - (csum2_pr2b a2) - (csum2_pr2a a2)
    - 2! cprod2_pr1 -aux -(aux Y) cZ - cY ccb -bsq -b2b.
have pzz:  (coarse Z  = (coarse F) \cup (a2 \cup a3)).
  set_extens t.
    move /setX_P => [pt]; case /setU2_P => r1.
      by case /setU2_P => r2; [| apply: setU2_2]; apply: setU2_1 ;apply: setX_i.
    move => xx; apply: setU2_2; apply: setU2_2 ;apply: setX_i => //.
  case /setU2_P; first by apply: setX_Sll; apply: subsetU2l.
  case /setU2_P => /setX_P [pa pb pc]; apply /setX_P;split => //;
    try (by apply: setU2_1); (by apply: setU2_2).
have dzz: (disjoint (coarse F) (a2 \cup a3)). 
  hnf in dFY.
  apply: disjoint_pr; move=> u /setX_P [_ pF qF];case /setU2_P.
     move /setX_P => [_ _ qY]; empty_tac1 (Q u).
     move /setX_P => [_ pY _]; empty_tac1 (P u).
have: (Y \Eq (a2 \cup a3)) by apply /card_eqP; rewrite - cY ca1.
move=> [x0 [bx0 sx0 tx0]].
set (g:= triple Z (coarse E) ((graph x) \cup (graph x0))).
have Hga: source g = Z by rewrite /g;aw. 
have Hgb:graph g = ((graph x) \cup (graph  x0)) by rewrite /g;aw.
have fx0: function x0 by fct_tac.
have Hgd: range (graph g) = coarse (source g).
  move: bx0 => [_ sjx0].
  rewrite Hga Hgb range_setU2 rgx pzz  -tx0 -(surjective_pr3 sjx0) //.
have Hhf: target g = coarse E by rewrite /g; aw.
have Hge:domain (graph g) = (source g).
  rewrite Hgb Hga domain_setU2; aw; rewrite sx0 //.
have ze: sub Z E.
  by move=> t /setU2_P []; [apply: ssxE | move /syc => /setC_P []].
have fg: function g. 
  rewrite /g;apply: function_pr.
      apply: fgraph_setU2; fprops; aw;rewrite sx0; exact dFY.
    rewrite - Hgb Hgd /coarse Hga; apply: setX_Slr =>//. 
  by rewrite -Hgb Hge Hga.
have ig: injection g.
   have aux1: forall a b c, inc (J a b) (graph x) -> inc (J c b) (graph x0) ->
    False.  
     move=> u v w J1 J2; empty_tac1 v.
     rewrite /F - rgx; ex_tac.
     rewrite -tx0; apply: (p2graph_target fx0 J2).
  apply: injective_pr_bis =>//; rewrite /related Hgb.
  move=> u v w; case  /setU2_P => h1; case /setU2_P => h2.
    apply: (injective_pr3 ijx h1 h2).
    case: (aux1 _ _ _ h1 h2).
    case: (aux1 _ _ _ h2 h1).
    move: bx0 => [ijx0 _]; apply: (injective_pr3 ijx0 h1 h2).
have sg: sub (graph x) (graph g) by rewrite Hgb;apply: subsetU2l. 
have gi: inc g ind. 
  apply: Zo_i; first by  apply /sfun_set_P; rewrite Hga; split => //.
  split => //; apply: sub_trans sgx sg.
have egx: extends g x by rewrite /extends; split => //; rewrite txE Hhf; fprops.
apply /set0_P; move=> y yY; move: (syc _ yY) => /setC_P [_ yF].
by case: yF; rewrite/F - (maxp _  gi egx) Hga; apply: setU2_2.
Qed.

Lemma csquare_inf a: infinite_c a -> a *c a = a.
Proof.
move=> ai.
have ca: (cardinalp a) by move: ai => [ca _].
by rewrite - cpowx2 //; apply: equipotent_inf2_inf.
Qed.

Lemma cpow_inf a n: infinite_c a -> natp n ->
  n <> \0c -> a ^c n = a.
Proof.
move=> ia nN nzn. 
have ca: (cardinalp a) by move: ia => [ca _].
have aux:= (csquare_inf ia).
have [qN ->] :=(cpred_pr nN nzn).
move: qN; set (q:= cpred n); move:q.
apply: Nat_induction.
  rewrite succ_zero; apply: cpowx1 =>//.
by move=> m mN; rewrite (cpow_succ _ (NS_succ mN)); move=> ->.
Qed.

Lemma cpow_inf1 a n: infinite_c a -> natp n ->
  (a ^c n) <=c  a.
Proof.
move=> ia nN.
case: (equal_or_not n \0c) => h.
  rewrite h cpowx0; apply: (finite_le_infinite finite_one ia). 
rewrite  (cpow_inf ia nN h);apply:cleR; exact: (proj1 ia).
Qed.

Lemma finite_family_product a f: fgraph f ->
  finite_set (domain f) -> infinite_c a ->
  (forall i, inc i (domain f) -> (Vg f i) <=c a) ->
  card_nz_fam f ->
  (exists2 j,  inc j (domain f) & (Vg f j) = a) ->
  cprod f = a.
Proof.
move=> fgf fsd ifa alea alnz [j0 j0d vj0].
set (g:= cst_graph (domain f) a). 
have fgg: fgraph g by rewrite /g /cst_graph; fprops. 
have df: domain f = domain g by rewrite /g/cst_graph; bw.
have le1:forall x, inc x (domain f) -> (Vg f x) <=c (Vg g x). 
  by move=> t tdf; rewrite /g/cst_graph; bw; apply: alea.
move: (cprod_increasing  df le1); rewrite/g cpow_pr2.
set (n:= cardinal (domain f)). 
have nN: natp n by fprops.
have nz: n <> \0c by apply: cardinal_nonemptyset1; exists j0 =>//.
rewrite - cpow_prb (cpow_inf ifa nN nz)=> le0. 
set (j:= singleton j0).
have alc: (forall x, inc x (domain f) -> cardinalp (Vg f x)). 
  by move=> x xdf; move: (alea _ xdf) => [h _].
move: (alc _ j0d); rewrite vj0 => ca.
have sjd: (sub j (domain f)) by  move=> t /set1_P => ->.
move: (cprod_increasing1 alc alnz sjd). 
by rewrite /j cprod_trivial4 // vj0 (card_card ca) => /(cleA le0).
Qed.

Lemma cprod_inf a b: b <=c a ->
  infinite_c a -> b <> \0c -> a *c b = a.
Proof. 
move=> leba ia nzb.
move :(leba) => [cb ca _]. 
move:(cprod_Meqle a leba); rewrite  (csquare_inf ia) => h.
exact: (cleA h (cprod_M1le ca nzb)).
Qed.

Lemma cprod_inf6 a b: cardinalp a -> cardinalp b ->
  (infinite_c a \/ infinite_c b) -> a <> \0c -> b <> \0c ->
  a *c b = cmax a b.
Proof.
move => ca cb H anz bnz.
case (cleT_ee ca cb)=> le1.
  rewrite (cmax_xy le1).
  have ib: infinite_c b by case H => // h; move:(ge_infinite_infinite h le1).
  by rewrite cprodC (cprod_inf le1 ib anz).
rewrite (cmax_yx le1).
have ib: infinite_c a by case H => // h; move:(ge_infinite_infinite h le1).
by rewrite (cprod_inf le1 ib bnz).
Qed.

Lemma cprod_inf1 a b: b <=c a ->
  infinite_c a -> (a *c b) <=c a.
Proof.
move =>leba ia; move :(leba) => [_ ca _]. 
case: (equal_or_not b \0c) => h.
  rewrite h; rewrite cprod0r; apply: (finite_le_infinite finite_0 ia).
rewrite (cprod_inf leba ia h); fprops. 
Qed.

Lemma cprod_inf2 a b: finite_c b ->
  infinite_c a -> (a *c b) <=c a.
Proof.
move => fb ia; exact: (cprod_inf1 (finite_le_infinite fb ia) ia).
Qed.


Lemma cprod_inf4 a b c:
   a <=c c -> b <=c c -> infinite_c c -> a *c b <=c c.
Proof.
move=> ac bc ci.
wlog lab: a b ac bc / a <=c b.
   move => H; case: (cleT_ee (proj31 ac) (proj31 bc)) => ab; first by apply: H.
   by rewrite cprodC; apply: H.
case: (Nat_dichot (proj31 bc)) => fc.
  apply: (Nat_le_infinite (NS_prod (NS_le_int lab fc) fc) ci).
move: (cprod_inf1 lab fc); rewrite cprodC => h; exact:(cleT h bc).
Qed.

Lemma cprod_inf5 a b c:
    a <c c -> b <c c -> infinite_c c -> a *c b <c c.
Proof.
move => l1 l2 ic.
move: (proj31_1 l1) (proj31_1 l2) => ca cb.
have [d [da [db dc]]]: exists d, a <=c d /\ b <=c d /\ d <c c.
  case: (cleT_ee ca cb) => cab; [ exists b | exists a];fprops.
case: (finite_dichot (proj31_1 dc)) => fd.
  move: (le_finite_finite fd da) (le_finite_finite fd db).
  move /NatP => aN /NatP bN.
  by apply: finite_lt_infinite => //; apply /NatP; apply: NS_prod.
exact (cle_ltT (cprod_inf4 da db fd) dc).
Qed.

Lemma inf_cpow x y: infinite_c x -> y <> \0c -> infinite_c (x ^c y).
Proof.
move=> ix ynz; exact: (ge_infinite_infinite ix (cpow_Mle1 (proj1 ix) ynz)).
Qed.

Lemma inf_cpow2 x y: infinite_c x ->  infinite_c y ->
   infinite_c (x ^c y).
Proof. move=> ix iy; apply: (inf_cpow ix (infinite_nz iy)). Qed.

Lemma inf_cpow3 x y : \2c <=c x -> infinite_c y -> infinite_c (x ^c y).
Proof.
move => sa sb.
case: (Nat_dichot (proj32 sa)) => bN; last by apply:inf_cpow2.
move:(clt_leT (cantor (proj1 sb)) (cpow_Mleeq y sa card2_nz)) => [h _].
exact:(ge_infinite_infinite sb h).
Qed.

Lemma cpow_succ_gen a b: cardinalp b -> a ^c (csucc b) = (a ^c b) *c a.
Proof.
move => cb.
case: (Nat_dichot cb) => bN; first by rewrite cpow_succ.
move/infinite_cP: (bN) => [_ <-].
rewrite - cpow_pra - cprod2_pr2b.
case: (cleT_el CS2 (CS_cardinal a)); last first.
  case/clt2 => ->.
    rewrite cprod0r cpow0x // => bz.
    apply: infinite_dichot1 bN; rewrite bz; apply: finite_0.
  by rewrite cpow1x (cprod1l CS1).
move => h.
move:(inf_cpow3 h bN) => ha.
have anz: cardinal a <> \0c.  
  by move => az; move: clt_02; rewrite - az => /(cleNgt h).
have Hb: cardinal a <=c cardinal a ^c b.
  apply:(cpow_Mle1 (CS_cardinal a) (infinite_nz bN)).
by rewrite(cprod_inf Hb ha anz).
Qed.

Lemma notbig_family_sum a f: 
  infinite_c a -> (cardinal (domain f)) <=c a -> 
  (forall i, inc i (domain f) -> (Vg f i) <=c a) ->
  (csum f) <=c a.
Proof. 
move=> ifa leda alea.
set (g:= Lg (domain f) (fun _ =>a)). 
have dg : domain f = domain g by rewrite /g; bw.
have ale: forall x, inc x (domain f) -> (Vg f x) <=c (Vg g x). 
  by move=> x xdf; rewrite /g; bw; apply: alea. 
move: (cprod_inf1 leda ifa) (csum_increasing dg ale).
rewrite csum_of_same  cprod2_pr2b => r1 r2; apply: cleT r2 r1.
Qed.

Lemma notbig_family_sum1 a f:
  infinite_c a -> (cardinal (domain f)) <=c a -> 
  (forall i, inc i (domain f) -> (Vg f i) <=c a) ->
  (exists2 j,  inc j (domain f) & (Vg f j) = a) ->
  csum f = a.
Proof. 
move=> ifa leda alea [j0 j0d vj0].
have ca:  (cardinalp a) by move: ifa => [ca _].
apply:cleA; first exact: (notbig_family_sum ifa leda alea).
have: sub (singleton j0) (domain f) by move=> t /set1_P ->. 
have: (forall x, inc x (domain f) -> cardinalp (Vg f x)).
  by move=> x xdf; move: (alea _ xdf) => [ok _].
move=>p1 p2;move: (csum_increasing1 p1 p2).  
by rewrite csum_trivial4 // vj0 // (card_card ca).
Qed.

Lemma csum_inf1 a: infinite_c a -> a +c a = a.
Proof. 
move=> ia.
have: (a +c a = a *c \2c) by rewrite cprodC  two_times_n//. 
rewrite (cprod_inf (infinite_ge2 ia) ia) //; fprops.
Qed.

Lemma csum_inf a b: b <=c a ->
  infinite_c a -> a +c b = a.
Proof. 
move=> leba ia; move :(leba) => [cb ca _]. 
apply: cleA; last by apply: csum_M0le.
by rewrite - {2} (csum_inf1 ia); apply: csum_Meqle. 
Qed.

Lemma csum_inf6 a b: cardinalp a -> cardinalp b ->
  (infinite_c a \/ infinite_c b) ->  a +c b = cmax a b.
Proof.
move => ca cb H.
case (cleT_ee ca cb)=> le1.
  rewrite (cmax_xy le1).
  have ib: infinite_c b by case H => // h; move:(ge_infinite_infinite h le1).
  by rewrite csumC (csum_inf le1 ib).
rewrite (cmax_yx le1).
have ib: infinite_c a by case H => // h; move:(ge_infinite_infinite h le1).
by rewrite (csum_inf le1 ib).
Qed.

Lemma csum_inf2 a b c: cardinalp c -> infinite_c a ->
  b <c a -> a = b +c c -> a = c.
Proof.
move => cc ica [ba nba] sv.
have cb := proj31 ba.
case: (cleT_ee cb cc) => h.
   case: (Nat_dichot cc) => ic.
     have fc := (NS_le_int h ic).
     have fa: finite_c a by apply /NatP; rewrite sv; apply: NS_sum.
     case: (infinite_dichot1 fa ica).
   by move: (csum_inf h ic); rewrite csumC sv.
case: (Nat_dichot cb) => ib.
  move: (NS_le_int h ib) => fb.
  have fa: finite_c a by  apply /NatP;  rewrite sv; apply: NS_sum.
  case: (infinite_dichot1 fa ica).
by move: (csum_inf h ib); rewrite - sv => eq; case: nba.
Qed.

Lemma card_setC1_inf E x:
  infinite_set E -> cardinal E = cardinal (E -s1 x).
Proof.
move=> iE; case: (p_or_not_p (inc x E)) => xE; last by rewrite setC1_eq.
move:(csucc_pr1 E x); rewrite setC1_K // => ces.
have cF := (CS_cardinal (E -s1 x)).
case: (finite_dichot cF). 
  move/(finite_succP cF); rewrite - ces=> h; case: (infinite_dichot2 h iE).
by move /infinite_cP; rewrite  - ces; move=> [_ res].
Qed.

Lemma csum_Mltlt a b c d : a <c b -> c <c d -> a +c c <c b +c d.
Proof.
move: a b c d.
suff: forall a b c d, b <=c d -> a <c b -> c <c d -> a +c c <c b +c d.
  move => H a b c d le1 le2.
  move: (proj32_1 le1)(proj32_1 le2) => cb cd.
  case: (cleT_ee cb cd) => le3; first by apply: H.
  by rewrite (csumC a) (csumC b); apply: H.
move => a b c d lebd lt1 lt2.
move: (proj32_1 lt2) => cd.
case: (finite_dichot cd).
  move /NatP => dN.
  have bN: natp b by Nat_tac.
  have aN: natp a by Nat_tac.
  have cN: natp c by Nat_tac.
  exact: (csum_Mlelt bN dN (proj1 lt1) lt2).
move => ifd.
rewrite (csumC b) (csum_inf lebd ifd).
have ad:= clt_leT lt1 lebd.
move: (proj31_1 lt1) (proj31_1 lt2) => ca cc.
suff: forall u v, u <=c v -> u <c d -> v <c d -> u +c v <c d.
  move => H; case: (cleT_ee ca cc) => le3; first by apply: H.
  by rewrite (csumC a); apply: H.
move => u v le1 lt3 lt4.
move: (proj31_1 lt4) => cv.
case: (finite_dichot cv).
  move /NatP => dN.
  have uN: natp u by Nat_tac.
  apply: finite_lt_infinite => //; apply /NatP; fprops.
by move => iv; rewrite csumC (csum_inf le1 iv).
Qed.

Lemma csum2_pr6_inf1 a b X: infinite_c X ->
  cardinal a <=c X -> cardinal b <=c X -> 
  cardinal (a \cup b) <=c X.
Proof.
move => pa pb pc.
move: (csum_Mlele pb pc); rewrite (csum_inf1 pa) csum2_pr2a csum2_pr2b.
move: (csum2_pr6 a b); apply:cleT.
Qed.

Lemma csum2_pr6_inf2 a b X: infinite_c X ->
  cardinal a <c X -> cardinal b <c X -> 
  cardinal (a \cup b) <c X.
Proof.
move => pa pb pc.
move: (csum_Mltlt pb pc); rewrite (csum_inf1 pa) csum2_pr2a csum2_pr2b.
move: (csum2_pr6 a b); apply:cle_ltT.
Qed.


Lemma infinite_compl A B:
   infinite_set B -> cardinal A <c cardinal B ->
   cardinal (B -s A) = cardinal B.
Proof.
move => /infinite_setP p1 p2.
ex_middle nsc.
move: (sub_smaller (@sub_setC B A)) => h1.
move: (csum2_pr6_inf2 p1 p2 (conj h1 nsc)).
by rewrite setU2Cr2; move: (sub_smaller (@subsetU2r A B)) => l1 /(cleNgt l1).
Qed.

Lemma infinite_union2 x y z:
    infinite_c z -> cardinal x <c z -> cardinal y <c z ->
    nonempty (z -s (x \cup y)).
Proof.
move => h1 h2 h3.
move: (csum2_pr6_inf2 h1 h2 h3) => h.
apply /nonemptyP => h4; move: (sub_smaller (empty_setC h4)).
by rewrite (card_card (proj1 h1)) => /(cltNge h).
Qed.

Lemma csum_lt_inf a b c: infinite_c c -> a <c c -> b <c c ->
    (a +c b) <c c.
Proof.
move => hc; wlog : a b / a <=c b.
  move => aux l1 l2. 
  case: (cleT_ee (proj31_1 l1) (proj31_1 l2)) => l3.
     by apply: aux.
  by rewrite csumC; apply: aux.
move => ab ac bc.
case: (finite_dichot (proj32 ab)).
   move /NatP => bb; apply: finite_lt_infinite => //; apply /NatP.
   apply: NS_sum => //; apply: (NS_le_int ab bb).
by move => ib; rewrite csumC (csum_inf ab ib).
Qed.

Lemma cdiff_inf a b: infinite_c a -> b <c a -> a -c b = a.
Proof.
move => ica lba.
move: (cdiff_pr (proj1 lba)) => h1; symmetry in h1.
have cd: cardinalp (a -c b) by rewrite /cdiff;fprops.
by rewrite - (csum_inf2 cd ica lba h1).
Qed.

Lemma cdiff_Mle_gen a b c:
  cardinalp a -> cardinalp b -> cardinalp c ->
  c <=c (a +c b) -> (c -c b) <=c a.
Proof.
move => ca cb cc h.
case:(cleT_el cc cb); first by move /cdiff_wrong=> ->; apply:czero_least.
move => bc.
case: (finite_dichot cc) => fc; last first.
  rewrite (cdiff_inf fc bc).
  move: (cdiff_pr (proj1 bc)) => h1.
  case: (cleT_el cc ca) => // ac.
  case: (cleNgt h (csum_lt_inf fc ac bc)).
move: (le_finite_finite fc (proj1 bc)) => fb.
have h2: (c -c b) <=c c.
  by move: (sub_smaller (@sub_setC c b)); rewrite (card_card cc).
case: (finite_dichot ca) => fa.
  by apply: cdiff_Mle => //; apply /NatP.
by apply: (cleT h2);rewrite - (csum_inf (finite_le_infinite fb fa) fa).
Qed.

Lemma cdiff_pr1_gen a b: cardinalp a -> cardinalp b ->
  (finite_c b \/ b <c a) -> (a +c b) -c b = a.
Proof.
move => ca cb h.
case: (finite_dichot ca).   
   move /NatP => fa; apply: cdiff_pr1 => //; case: h; first by move /NatP.
   move => [pa _ ];apply: (NS_le_int pa fa).
move => ica.
have ba : b <c a by case: h => //; move => hh; apply: finite_lt_infinite.
by rewrite (csum_inf (proj1 ba) ica) cdiff_inf.
Qed.

Lemma cdiff_pr2_gen a b: infinite_c b -> a <=c b -> (a +c b) -c b = \0c.
Proof.
move => icb ab.
by rewrite csumC (csum_inf) // cdiff_nn.
Qed.

Lemma csum_Meqlt_gen a a' b: natp b -> a <c a' -> b +c a <c b +c a'.
Proof.
move => bN l1.
case: (finite_dichot (proj32_1 l1)) => h.
  by apply: csum_Meqlt => //; apply /NatP.
have l2: b <=c  a' by apply:finite_le_infinite => //; apply /NatP.
rewrite (csumC _ a') (csum_inf l2 h).
case: (finite_dichot (proj31_1 l1)) => h2.
  by apply:finite_lt_infinite => //;apply/NatP; apply: NS_sum => //; apply/NatP.
have l3: b <=c a by apply:finite_le_infinite => //; apply /NatP.
by rewrite csumC  (csum_inf l3 h2).
Qed.

Lemma cprod_inf3 E F: nonempty E ->
   (cardinal E) <=c (cardinal F) -> infinite_set F ->
   (F \times E) \Eq  F.
Proof.
move=>  [a aE] le1 infF.
have pa: (cardinal E <> \0c).
 move=> aux; move: (cardinal_nonemptyset aux) => fE; empty_tac1 a.
move: infF => /infinite_setP infF.
move: (cprod_inf le1 infF pa) => hh.
apply /card_eqP.
by rewrite - cprod2_pr1 -hh;apply: cprod2_pr2;rewrite double_cardinal.
Qed.

Lemma Exercise6_5a  E F: 
  (cardinal (functions E F)) <=c (cardinal (sub_functions E F)).
Proof.
apply: sub_smaller. 
move=> t /fun_set_P [pa pb pc];apply /sfun_set_P;split => //;rewrite pb; fprops.
Qed.

Lemma Exercise6_5b E F:
  (cardinal (sub_functions E F)) <=c (cardinal (powerset (product E F))).
Proof.
have injf: 
 injection (Lf graph  (sub_functions E F) (powerset (product E F))). 
  apply: lf_injective.
    by move=> t ta; apply /setP_P;apply: graph_of_function_sub.  
  move=> u v /sfun_set_P [y yp yq] /sfun_set_P [z zp zq] h.
  apply:  function_exten1 => //; ue.
move: (incr_fun_morph injf); aw.
Qed.


Lemma Exercise6_5c E: infinite_set E ->
  (cardinal (permutations E)) <=c  (cardinal (powerset E)).
Proof.
move=> isE.
set C:= (functions E E).
have pb: sub (Zo C bijection) C by apply: Zo_S.
apply: (cleT (sub_smaller pb)).
apply: (cleT (Exercise6_5a E E)).
case: (emptyset_dichot E) => neE.
  have ->: (sub_functions E E = singleton (identity E)).
  apply: set1_pr.
    apply /sfun_set_P; aw;split;fprops.
    move => z /sfun_set_P [sa sb sc]; apply /function_exten;aw; fprops.
      apply: extensionality => //; rewrite neE;fprops.
    by move => t te; move: (sb _ te); rewrite neE => /in_set0.
  rewrite cardinal_set1; apply: cge1; fprops => bad.
  move: (cardinal_nonemptyset bad) => bad1; empty_tac1 E; aw; apply /setP_Ti.
have cle2: (cardinal E) <=c (cardinal E) by fprops.
move: (Exercise6_5b E E) (cprod_inf3 neE cle2 isE) => ca cb.
have -> : (cardinal (powerset E) = cardinal (powerset (product E E))) => //.
symmetry;rewrite !card_setP; apply: cpow_pr; fprops. 
Qed.


(** ** EIII-6-4 Countable sets *)
Definition countable_set E:= equipotent_to_subset E Nat.
Definition countable_infinite E := countable_set E /\ infinite_set E.

Lemma countableP E:
  countable_set E  <-> (cardinal E) <=c (cardinal Nat).
Proof. 
apply (iff_trans (eq_subset_cardP E Nat)).
apply:cardinal_le_aux2P; fprops.
Qed.

Lemma countable_finite_or_N E: countable_set E ->
  finite_c (cardinal E) \/ cardinal E = cardinal Nat.
Proof. 
move /countableP => leB.
case: (finite_dichot (CS_cardinal E)); [ by left | move=> [_ nf]; right].
exact: (cleA leB (infinite_greater_countable1 nf)).
Qed.

Lemma countable_finite_or_N_b E: countable_set E ->
  finite_set E \/ equipotent E Nat.
Proof.  
by case /(countable_finite_or_N) => h; [ left | right; apply /card_eqP].
Qed.

Lemma countable_infiniteP E: countable_infinite E <-> equipotent E Nat.
Proof. 
split.
  move => [];case /(countable_finite_or_N_b) =>//.
  move=> h1 h2; case: (infinite_dichot2 h1 h2).
move /card_eqP => eq; split; first by  apply /countableP; rewrite eq; fprops.
apply /infinite_setP; rewrite eq;apply /infinite_setP; apply infinite_Nat.
Qed.

Theorem countable_sub E F: sub E F -> countable_set F  ->
  countable_set E.
Proof. 
move=> /sub_smaller h1 /countableP h; apply /countableP; exact:cleT h1 h.
Qed.

Lemma countable_Nat : countable_set Nat.
Proof. apply/countableP; apply:cleR; fprops. Qed.

Lemma countable_sub_Nat x : sub x Nat -> countable_set x.
Proof. move => h; apply: (countable_sub h countable_Nat).  Qed.

Lemma countable_fun_image z f:
  countable_set z -> countable_set (fun_image z f).
Proof.
move /countableP => h; apply /countableP.
exact:(cleT (fun_image_smaller z f) h).
Qed.

Theorem countable_product f:
  finite_set (domain f) -> 
  (allf f countable_set) -> 
  countable_set (productb f).
Proof. 
move=> fsd alc; apply /countableP.
set (g:= Lg (domain f) (fun i => cardinal (Vg f i))).
have ->:(cardinal (productb f) = cprod g) by rewrite cprod_pr =>//.
apply: (@cleT  (cprod (cst_graph (domain f) (cardinal Nat)))).
  apply: cprod_increasing.  
    rewrite /g; bw. 
   by rewrite /g; bw; move=> x xdf; bw; move: (alc _ xdf) => /countableP.
rewrite cpow_pr2 -(cpow_prb (cardinal Nat) (domain f)).
apply:cpow_inf1; last by apply /NatP. 
apply /infinite_setP; exact infinite_Nat.
Qed.


Theorem countable_union f:
  countable_set (domain f) -> 
  (allf f countable_set) ->
  countable_set (unionb f).
Proof. 
move=> cs alc; apply /countableP.  
set d:= domain f.
apply: (@cleT (csumb d (fun i => cardinal (Vg f i)))). 
  apply: csum_pr1 =>//.
set (h:=  cst_graph d (cardinal Nat)).
apply: (@cleT (csum h)).
  rewrite/h;apply: csum_increasing; fprops; bw. 
  move=> x xd; bw; move: (alc _ xd) => /countableP //. 
rewrite csum_of_same - (cprod2_pr2b (cardinal Nat) d).
apply: cprod_inf1 =>//. 
  move: cs => /countableP //. 
apply /infinite_setP; apply: infinite_Nat.
Qed.

Lemma countable_setU2 a b: 
   countable_set a -> countable_set b -> countable_set (a \cup b).
Proof.
move => /countableP ca /countableP cb; apply/countableP.
have icb: infinite_c (cardinal Nat) by apply/infinite_setP /infinite_Nat. 
move: (csum2_pr6 a b); rewrite - csum2_pr2a - csum2_pr2b => h.
by move: (csum_Mlele ca cb); rewrite csum_inf1 //; apply: cleT.
Qed.

Lemma countable_setX2 a b: 
   countable_set a -> countable_set b -> countable_set (a \times b).
Proof.
move => /countableP ca /countableP cb; apply/countableP.
have icb: infinite_c (cardinal Nat) by apply/infinite_setP /infinite_Nat.
rewrite - (cprod2_pr1 a b)- cprod2_pr2a - cprod2_pr2b. 
by apply: cprod_inf4.
Qed.

Theorem infinite_partition E: infinite_set E ->
  exists f, [/\ partition_w_fam f E, (domain f) \Eq E &
    (forall i, inc i (domain f) -> (countable_infinite (Vg f i)))]. 
Proof. 
move=> iE; move: (infinite_greater_countable1 iE) => h1.
have iE': infinite_c (cardinal E) by apply /infinite_setP.
move: (cprod_inf h1 iE' card_Nat_nz).
rewrite cprod2_pr1; move /card_eqP=> [f [bf sf tf]].
move: (bf) => [injf sjf].
pose G a := (indexedr a (cardinal Nat)).
set (g:= Lg (cardinal E) (fun a => image_by_fun f (G a))).
have ff: function f by fct_tac.
have ppa: forall i, inc i (cardinal E) -> sub (G i) (source f).
  by move => i ie;rewrite sf; apply: setX_Sl; apply/sub1setP.
exists g; rewrite /g;split => //; bw; [ | fprops | ].
  split => //; fprops.
    apply: mutually_disjoint_prop;bw; move=> i j y inE jnE; bw.
    move: (ppa _ inE)(ppa _ jnE) => pa pb.
    move /(Vf_image_P ff pa) => [u u1 u2] /(Vf_image_P ff pb)  [v v1 v2].
    rewrite u2 in  v2; move: (proj2 injf _ _ (pa _ u1) (pb _ v1) v2).
    by move: u1 v1 => /setX_P [_ /set1_P <-] _ /setX_P [_ /set1_P <-] _ ->.
   set_extens t.
     move=> /setUb_P1 [y ycE]; move/(Vf_image_P ff (ppa _ ycE)).
     by move => [u u1 ->]; rewrite -tf; Wtac; apply: (ppa _ ycE).
  move => tE.
  have tt: inc t (target f) by rewrite tf.
  move: (proj2 sjf _ tt);move=> [x xsf jG].
  move: xsf; rewrite sf; move /setX_P => [pax px qx].
  apply /setUb_P1;ex_tac; apply /(Vf_image_P ff (ppa _ px)).
  exists x => //; apply /setX_P; split;fprops.
move => i inE; bw.
have sd: (sub (G i) (source f)) by apply: ppa.
apply /countable_infiniteP /card_eqP.
by rewrite (cardinal_image sd injf) cardinal_indexedr double_cardinal.
Qed.

Lemma countable_inv_image f: surjection f ->
  (forall y, inc y (target f) -> 
    countable_set (inv_image_by_fun f (singleton y))) ->
  infinite_set (target f) ->
  (source f) \Eq (target f).
Proof. 
move => sf alc it.
have ff: function f by fct_tac.
set (pa := Lg (target f) (fun  z=> (inv_image_by_fun f (singleton z)))).
have upa: (unionb pa = (source f)). 
  set_extens t.
    by move /setUb_P1 => [y ytf] /(iim_fun_set1_P _ ff) [].
  move => tsf; apply /setUb_P1; exists (Vf f t)=> //; first by  Wtac.
  apply (iim_fun_set1_P _ ff);split => //.
have md: (mutually_disjoint pa). 
  rewrite /pa;apply: mutually_disjoint_prop;bw.
  move=> i j y itf jtf;bw; move /(iim_fun_set1_P _ ff) => [_ ->].
  by move /(iim_fun_set1_P _ ff) => [_ <-].
move: (csum_pr pa). 
rewrite /disjointU.
have: (equipotent (unionb (disjointU_fam pa)) (unionb  pa)).
  apply: equipotent_disjointU => //.
  - rewrite /pa /disjointU_fam; split => //;fprops.
  - rewrite/disjointU_fam /pa; bw. 
  - rewrite /disjointU_fam /pa; bw; move=> i idf; bw. 
  - eqsym; fprops.
  - fprops.
move /card_eqP => ->;  rewrite upa => aux.
apply/card_eqP.
have le2: ((cardinal (target f)) <=c (cardinal (source f))).
  apply:  surjective_cle; exists f; split => //.
apply: (cleA _ le2).
set (cf:= Lg (domain pa) (fun a  => cardinal (Vg a pa))).
set (g:= Lg (domain pa) (fun _:Set => cardinal Nat)).
have : (cardinal (source f)) <=c (csum g).
  rewrite aux /cf /g /pa;apply: csum_increasing; bw.
  move => x xa; bw; apply /countableP; apply: alc =>//.
rewrite csum_of_same.
have -> //:((cardinal Nat) *c (domain pa) = cardinal (target f)).
rewrite /pa; bw;rewrite - (cprod2_pr2b _ (target f)).
set (u:=  cardinal (target f)).
rewrite cprodC; apply: cprod_inf.
    apply: infinite_greater_countable1 =>//.
  split; first (by apply: CS_cardinal); done.
apply: card_Nat_nz.
Qed.

Theorem infinite_finite_subsets E: infinite_set E ->
  (Zo (powerset E) finite_set) \Eq E.
Proof. 
move=> inE.
have icE: infinite_c (cardinal E) by split; [apply: CS_cardinal | exact].
set bF:=Zo _ _ .
pose T n :=  Zo (powerset E) (fun z  => cardinal z = n).
have le1: (forall n, inc n Nat -> (cardinal (T n)) <=c (cardinal E)).
  move=> n nN; rewrite /T. 
  have cn: (cardinal n = n) by rewrite card_card; fprops.
  set (K:= injections n E).
  have ta: lf_axiom  (fun z => range (graph z)) K (T n).
    move=> z /Zo_P [] /fun_set_P [fz sz tz] inz; apply: Zo_i.
       apply /setP_P;rewrite -tz => //; apply: f_range_graph => //.
    by rewrite (cardinal_range inz) sz cn.
  set (f:= Lf (fun z => range (graph z)) K (T n)).
  have ff: function f by apply: lf_function.
  set (q:= permutations n).
  set (c := cardinal q). 
  have  fc: (finite_c c). 
    have fsn: finite_set n by rewrite /finite_set cn ; fprops.
    rewrite /c /q (number_of_permutations fsn). 
    apply : Nat_hi; apply: NS_factorial; ue.
  have cii: (forall x, inc x (target f) -> 
    cardinal (inv_image_by_fun f (singleton x)) = c). 
     move=> x ;rewrite/f; aw; move => /Zo_P [] /setP_P xE cx.
    have [x0 [bx0 sx0 tx0]]: (equipotent n x) by rewrite - cx; fprops.
    set x1 := triple n E (graph x0).
    have sx1: (source x1 = n) by rewrite /x1; aw.
    have tx1: (target x1 = E) by rewrite /x1; aw.
    have fx1: function x1.
      have fx0 := (proj1 (proj1 bx0)).
      have ha := (function_fgraph fx0).
      have hb:sub (range (graph x0)) E.
        apply: (sub_trans (f_range_graph fx0)); ue.
      by apply: (function_pr ha hb); aw.
    have ix1:  injection x1.
      apply:injective_pr_bis => //.
      rewrite /x1; aw;move=> u v w p1 p2.
      apply: (injective_pr (proj1 bx0) p1 p2).
    set iif := inv_image_by_fun _ _.
    set (g:= Lf (fun z=>  (compose x1 z)) q iif).
    have ta2:  (lf_axiom (fun z=> (x1 \co z)) q iif). 
      rewrite/q/iif => z zq; apply /iim_fun_set1_P => //.
      move: zq => /Zo_P [] /fun_set_P [fz sz tz] bz.
      have cxz: (x1 \coP z) by split => //; ue.
      set (t:= x1 \co z).  
      have fr: (function t) by rewrite /t; fct_tac.
      have tk: (inc t K).  
        apply: Zo_i; first by apply /fun_set_P;split => //; rewrite /t; aw.
        apply: inj_compose1 =>//; [ move: bz => [iz _] => // | ue].
      split; first by aw.
      have gz: sgraph (graph z) by fprops.
      suff: (x = Vf f t) by move => ->; rewrite /f; aw.
      rewrite /f/t/compose; aw. 
      rewrite (compg_range _ gz) (surjective_pr3 (proj2 bz)) tz /x1 /Lf; aw.
      move: (surjective_pr0 (proj2 bx0)).
      rewrite /image_of_fun /image_by_fun sx0 tx0 //. 
    have fg: function g by  rewrite /g; apply: lf_function =>//.
    apply /card_eqP;eqsym; exists g; split => //; rewrite /g; aw.
    apply: lf_bijective => //.  
        move=> u v /Zo_P [] /fun_set_P [fu su tu] bu
          /Zo_P [] /fun_set_P [fv sv tv] bv sc.
        apply: function_exten; try fct_tac; try ue.
        move=> w wsu.
        have : (Vf (compose x1 u) w = Vf (compose x1 v) w) by ue.
        have c1: composable x1 v by  split => //; try fct_tac; try ue.
        have c2: composable x1 u by  split => //; try fct_tac; try ue.
        have wsv :inc w (source v) by rewrite sv - su.
        aw; apply: (proj2 ix1). 
          rewrite sx1 -tu; Wtac;fct_tac. 
          rewrite sx1 -tv; Wtac;fct_tac.
     move => y /(iim_fun_set1_P _ ff) [].
     rewrite /f lf_source => pa; aw => rgy.
     move: pa; rewrite /f; aw; move => /Zo_P [y1 iy].
     have ww: sub (range (graph y)) (range (graph x1)).
       rewrite -rgy /x1; aw.
       move: bx0 => [_ bx0]; rewrite (surjective_pr3 bx0) tx0; fprops.
     move: y1 => /fun_set_P [fy sy ty].
     rewrite -tx1 in ty; move: ww;rewrite -(exists_right_composable fy ix1 ty). 
     move => [h [cph ch]]; exists h => //. 
     have sh: (source h = n) by rewrite - sy - ch; aw. 
     have th: (target h = n) by move: cph => [_ _ hq]; rewrite -hq.
     have ih: injection h. 
       split;first by fct_tac. 
       move => v w vsh wsh sw.
       have: Vf y v = Vf y w by rewrite - ch; aw; rewrite sw.
       apply: (proj2 iy); [rewrite sy - sh // |rewrite sy - sh //].
    have bh: bijection h.
      apply: bijective_if_same_finite_c_inj; rewrite ? th ? sh //.
        by rewrite /finite_set cn; fprops.
    apply: Zo_i => //; apply /fun_set_P;split => //; fct_tac.
  move: (shepherd_principle ff cii).
  rewrite /f; aw; move => cK.
  case: (equal_or_not n \0c) => nz.
    suff: (cardinal (T \0c) = \1c).
      rewrite nz; move => ->; apply: finite_le_infinite; [ fprops | exact].
    suff: (T \0c = singleton emptyset) by move => ->; apply: cardinal_set1.
    apply: set1_pr.
      apply: Zo_i; [apply:setP_0i | apply:cardinal_set0].
    move => z /Zo_P  [_ aux]; rewrite (cardinal_nonemptyset aux) //.
  case: (finite_dichot (CS_cardinal (T n))) => finT.
    apply: finite_le_infinite =>//.
  change ( (cardinal (T n)) <=c (cardinal E)).
  set (K1:= cardinal (functions n E)). 
  have k1E: (K1 = cardinal E).
    by rewrite /K1 -/(E ^c n) - cpow_pra; apply:  cpow_inf.
  have: ( (cardinal K) <=c (cardinal E)).
     rewrite -k1E; apply: sub_smaller;apply: Zo_S.
  apply: cleT; rewrite cK.
  have i2: (c <=c (cardinal (T n))) by  apply: finite_le_infinite. 
  have i3: (c <> \0c). 
    rewrite /c/q.
    have fsn: finite_set n by rewrite /finite_set cn; fprops.
    rewrite (number_of_permutations fsn) cn; apply: (factorial_nz nN).
  rewrite (cprod_inf i2 finT i3); fprops.
set Fn := Lg Nat T.
have bfu: (bF = unionb Fn). 
  set_extens t.
    move /Zo_P => [pa pb]; apply /setUb_P1;  exists (cardinal t).
       rewrite -/(natp _); fprops.
    apply /Zo_i => //. 
  move /setUb_P1 => [y yb] /Zo_P [pa pb]; apply /Zo_P;split => //.
  by apply /NatP; rewrite pb.
have: (equipotent (unionb (disjointU_fam Fn)) (unionb Fn)).
  apply: equipotent_disjointU; rewrite/Fn/disjointU_fam. 
     split => //; fprops; bw=> i iN; bw; eqsym; fprops.
    apply: mutually_disjoint_prop; rewrite /Fn; bw. 
    move=> i j y iN jN; bw.
    by move => /indexed_P [_ _ <-]  /indexed_P [_ _ <-].
  apply: mutually_disjoint_prop; rewrite /Fn; bw. 
  by move=> i j y iN jN; bw; move => /Zo_hi <- /Zo_hi <-.
move /card_eqP; rewrite -bfu => euu.
move: (csum_pr Fn); rewrite /disjointU euu  => cbf.
apply /card_eqP; apply: cleA; last first.
  apply /eq_subset_cardP1; apply /eq_subset_ex_injP.  
  exists (Lf (fun x => singleton x) E bF); split => //; aw.
  apply: lf_injective. 
    move=> x xe /=;  rewrite/bF; apply: Zo_i; last by apply: set1_finite.
    by apply /setP_P; apply /sub1setP.
  move=> u v _ _;apply: set1_inj. 
have: (csumb (domain Fn)(fun a => cardinal (Vg Fn a)))
    <=c (csumb  (domain Fn) (fun a  => cardinal E)).
  apply: csum_increasing; fprops; bw; rewrite /Fn;bw; move=> x xd; bw. 
  apply: le1 =>//.
rewrite {2} /csumb csum_of_same - cbf.
have -> //: ( (cardinal E) *c  (domain Fn) = cardinal E).
rewrite /Fn; bw.
rewrite - (cprod2_pr2b _ Nat).
apply: cprod_inf => //.
  apply: infinite_greater_countable1 =>//. 
apply: card_Nat_nz.
Qed.

Lemma infinite_finite_sequence E: infinite_set E ->
 (Zo (sub_functions Nat E) (fun z=> finite_set (source z))) \Eq E.
Proof.
move=> iE.
have iE': infinite_c (cardinal E) by  apply /infinite_setP.
set q:= Zo _ _.
apply /card_eqP; apply: cleA.
  move: (infinite_finite_subsets infinite_Nat).
  set Fn:=Zo _ _; move=> fse.
  have qu:q = unionb (Lg Fn (fun z=> (functions z E))).
    set_extens t.
      move => /Zo_P [] /sfun_set_P [ft sst tt] fst.
      apply /setUb_P1;exists (source t) => //. 
         apply: Zo_i =>//; by apply /setP_P.
      apply /fun_set_P;split => //.
    move => /setUb_P1 [y] /Zo_P [] /setP_P pa pb /fun_set_P [ft st tt].
    apply: Zo_i; [ apply /sfun_set_P;split => //; ue | rewrite st; fprops ].
  have ze: (forall z, inc z Fn -> 
    (cardinal (functions z E)) <=c (cardinal E)).
    move=> z zFn. rewrite -/(E ^c z) - cpow_pra - cpow_prb.
    apply: cpow_inf1 =>//; move: zFn => /Zo_P [_ pa]; fprops. 
  move: (csum_pr1 (Lg Fn (functions^~ E))). 
  rewrite -qu; bw; rewrite /csumb.
  set (g:= (Lg Fn (fun _:Set  => cardinal E))).
  set f0:= (Lg Fn _ ).
  have fg1: (fgraph f0) by rewrite /f0; fprops. 
  have fg2: (fgraph g) by rewrite /g; fprops.  
  have df0g: domain f0 = domain g by rewrite /f0  /g; bw.
  have ale: (forall x, inc x (domain f0) -> (Vg f0 x) <=c (Vg g x)).
    by rewrite /f0/g; bw; move=> x xdf; bw; apply: ze.
  move: (csum_increasing  df0g ale) => le1 le2.
  move: (infinite_greater_countable1 iE) => le3.
  move: (cprod_inf1 le3 iE'). 
  have <-: (cardinal Fn = cardinal Nat) by apply /card_eqP.
  have ->: ((cardinal E) *c (cardinal Fn) = (cardinal E) *c Fn).
    apply: cprod2_pr2; fprops; apply: double_cardinal.
  rewrite csum_of_same in le1.
   apply: cleT  (cleT le2 le1).
apply /eq_subset_cardP1; apply /eq_subset_ex_injP.
have aux: forall v, inc v E ->
   lf_axiom (fun _  => v) (singleton \0c) E by move=> vE t //=.  
exists (Lf (fun z  => (Lf ((fun _:Set => z))(singleton \0c) E)) E q).
split => //;aw; apply: lf_injective.
  move=> z  zE;  apply: Zo_i.
    apply /sfun_set_P;aw;split => //.
       apply: lf_function =>//.  
    apply: set1_sub; rewrite -/(natp _); fprops. 
   aw; apply: set1_finite.
move=> u v uE vE sv.
have zs: (inc  \0c (singleton \0c)) by fprops. 
move: (f_equal (Vf ^~ \0c) sv); aw; apply: aux =>//.
Qed.

(** Aleph zero *)

Definition aleph0 :=  omega0.

Lemma aleph0_pr1: aleph0 = cardinal Nat. 
Proof. by rewrite /aleph0 -/Nat cardinal_Nat. Qed.

Lemma aleph0_infinite: infinite_c aleph0. 
Proof. by move /infinite_setP: infinite_Nat; rewrite aleph0_pr1. Qed.

Lemma aleph0_pr2:  aleph0 +c aleph0 = aleph0.
Proof.  by apply: (csum_inf1 aleph0_infinite). Qed.

Lemma aleph0_pr3:  aleph0 *c aleph0 = aleph0.
Proof. by apply: (csquare_inf aleph0_infinite). Qed.

Lemma aleph0_countable E: cardinal E = aleph0 -> countable_set E.
Proof. move => h; apply /countableP; rewrite h aleph0_pr1; fprops. Qed.

Lemma aleph0_plus1: aleph0 +c \1c = aleph0.
Proof. 
move:aleph0_infinite => ai.
by rewrite -(csucc_pr4 (proj1 ai)); move/(infinite_cP): ai => [_ <-].
Qed.

(** ** EIII-6-5 Stationary sequences *)

Definition stationary_sequence f :=
  [/\ fgraph f, domain f = Nat &
  exists2 m, natp m & forall n, natp n -> m <=c n ->
    Vg f n = Vg f m].

Definition increasing_sequence f r:=
  [/\ fgraph f, domain f = Nat, sub (range f) (substrate r) &
  forall n m, natp n -> natp m -> n <=c m ->
    gle r (Vg f n) (Vg f m)].

Definition decreasing_sequence f r:=
  [/\ fgraph f, domain f = Nat, sub (range f) (substrate r) &
  forall n m, natp n -> natp m -> n <=c m ->
    gle r  (Vg f m) (Vg f n)].

Lemma increasing_seq_prop f r: order r ->
  function f -> source f = Nat -> sub (target f) (substrate r) ->
  (forall n, natp n -> gle r (Vf f n) (Vf f (csucc n))) ->
  increasing_sequence (graph f) r.
Proof. 
move => or ff sf tf ale.
split;fprops; aw.
   apply: (@sub_trans (target f)) => //;  apply: f_range_graph=>//.
move=> n m nN mN /cdiff_pr <-.
rewrite -/(Vf _ _ ) -/(Vf _ _).
move: (m -c n) (NS_diff n mN); apply: Nat_induction. 
   by aw; fprops;order_tac; apply: tf; Wtac;rewrite sf.
move=> p pN le1.
rewrite (csum_nS _ pN); move: (ale _ (NS_sum nN pN))=> le2; order_tac.
Qed.

Lemma decreasing_seq_prop f r: order r ->
  function f -> source f = Nat -> sub (target f) (substrate r) ->
  (forall n, natp n -> gle r (Vf f (csucc n))  (Vf f n)) ->
  decreasing_sequence (graph f) r.
Proof. 
move => or ff sf tf ale.
split;fprops; aw.
   apply: (@sub_trans (target f)) => //;  apply: f_range_graph=>//.
move=> n m nN mN nm;rewrite- (cdiff_pr nm).  
rewrite -/(Vf _ _ ) -/(Vf _ _).
move: (m -c n)  (NS_diff n mN);apply: Nat_induction. 
  by aw; fprops;order_tac; apply: tf; Wtac;rewrite sf.
move=> p pN le1.
rewrite (csum_nS _ pN).
move: (ale _ (NS_sum nN pN)) => le2; order_tac.
Qed.

Theorem  increasing_stationaryP r: order r ->
  ((forall X, sub X (substrate r) -> nonempty X ->
     exists a,  maximal (induced_order r X) a) <->
  (forall f, increasing_sequence f r -> stationary_sequence f)).
Proof. 
move=> or;split.
  move=> hyp f [fgf df rf incf]; rewrite /stationary_sequence;split => //. 
  have ner: (nonempty (range f)).
    exists (Vg f \0c); apply /range_gP => //. 
    exists \0c => //;rewrite df; apply:NS0.
  move: (hyp _ rf ner) => [a];rewrite /maximal; aw.
  move => [ha]; move: (ha) => /(range_gP fgf) [x pa eq] alv; rewrite df in pa.
  exists x => // n nN xn.
  move: (incf _ _ pa nN xn); rewrite -eq => h; apply: alv. 
  apply / iorder_gleP => //; apply /(range_gP fgf); exists n=> //; ue.
move=> h X Xsr neX.
pose T x := Zo X (fun y => glt r x y). 
case: (emptyset_dichot (productb (Lg X T))) => pe.
  have [x xX Tx]: (exists2 x, inc x X & T x = emptyset). 
    ex_middle eh.
    have p1: (fgraph (Lg X T)) by fprops.
    have p2: (forall i, inc i (domain (Lg X T)) -> nonempty (Vg (Lg X T) i)).
      bw;move=> i iX; bw;case: (emptyset_dichot (T i))=> //ie. 
      case: eh; ex_tac. 
    by move: (setXb_ne p2); rewrite pe; move /nonemptyP.
  exists x;split => //;aw=> t xt; move:(iorder_gle1 xt) => xt1.
  move: (iorder_gle3 xt) => [_ tX].
  ex_middle w; empty_tac1 t; apply: Zo_i =>//; split;fprops.
move:pe=> [y yp].  
have p1:  (forall x, inc x X -> glt r x (Vg y x)). 
  have aux: fgraph (Lg X T) by fprops.
  move: yp => /setXb_P; bw; move => [pa pb pc] x xX.
  by move: (pc _ xX); bw =>/Zo_P [].
have p2: (forall x, inc x X -> inc (Vg y x) X).
  have aux: fgraph (Lg X T) by fprops.
  move: yp => /setXb_P; bw; move => [pa pb pc] x xX.
  by move: (pc _ xX); bw =>/Zo_P [].
move:neX => [y0 y0X].
move: (integer_induction_stable y0X p2).
move:(induction_defined_pr (Vg y) y0); simpl.
set f:= induction_defined _ _.
move=> [sf [ff _] f0 fn] tfX.
have p3: (forall n, natp n -> glt r (Vf f n) (Vf f (csucc n))).
  move=> n nN; rewrite fn //; apply: p1; apply: tfX; Wtac; ue.
have p4: (forall n, natp n -> gle r (Vf f n) (Vf f (csucc n))).
  by move=> n nN; move: (p3 _ nN)=> [ok _].  
have s2: sub (target f) (substrate r) by apply: (sub_trans tfX).
move:(h _ (increasing_seq_prop or ff sf s2 p4)) => [_ _ [m mN sm]].
have sN: natp (csucc m) by fprops.
move: (sm _ sN (cleS mN)); move: (p3 _ mN); rewrite /Vf.
by move=> /proj2 h1 h2; move: h1; rewrite h2.
Qed.

Theorem  decreasing_stationaryP r: total_order r ->
  ((worder r) <->
  (forall f, decreasing_sequence f r -> stationary_sequence f)).
Proof. 
move=> tor; split  => hyp. 
  move=> f [fgf df rf incf]; rewrite /stationary_sequence;split => //.
  have ner: (nonempty (range f)). 
    exists (Vg f \0c); apply /range_gP => //. 
    exists \0c => //;rewrite df; apply: NS0.
  move: hyp=> [or wor]; move: (wor _ rf ner) => [y []]; aw.
  move /(range_gP fgf) => [x xdf xv] hb.
  rewrite df in xdf.
  exists x => // n nN xn.
  move: (incf _ _ xdf nN xn) => pa.
  have hh: inc (Vg f n) (range f) 
     by apply /(range_gP fgf); exists n => //; ue.
  move: (iorder_gle1 (hb _ hh)); rewrite xv => leq1; order_tac.
set (r':= opp_order r). 
have or: order r by move: tor => [or _].
move: (opp_osr or) => [or' sr'].
split =>// x xsr nex.
move: (total_order_lattice (total_order_sub tor xsr)) => la.
have srsr': (substrate r = substrate r') by rewrite sr'.
have aux: (forall f, increasing_sequence f r' -> stationary_sequence f).
  move=> f fi; apply: hyp; move : fi => [p1 p2 p3 p4]; split => //; try ue.
  move=> n m nN mN nm; move: (p4 _ _ nN mN nm); rewrite /r';bw.
  by move  /igraph_pP.
move: (iorder_osr or xsr) => [oi soi].
move: aux; move /(increasing_stationaryP or') => aux.
rewrite - srsr' in aux; move: (aux _ xsr nex) => [a].
rewrite /r'.
have ->: (induced_order (opp_order r) x = 
    opp_order (induced_order r x)) by apply: iorder_opposite.
move/(maximal_opp oi) => mi.
move: (lattice_directed la) => [_ ld].
exists a; apply: (left_directed_minimal ld mi).
Qed.


Definition decreasing_strict_sequence f r :=
 [/\ fgraph f, domain f = Nat, sub (range f) (substrate r)
   & forall n m, natp n -> natp m -> n <c m -> glt r (Vg f m) (Vg f n)].

Lemma total_order_worder_dichot r: total_order r ->
   (worder r \/ exists f, decreasing_strict_sequence f r).
Proof.
move => tor.
move:(decreasing_stationaryP tor) => H.
case: (p_or_not_p (worder r)) => ha; [by left | right].
ex_middle hb; case: ha; apply/H => f [qa qb qc qd]; split => //.
ex_middle hc; case hb.
pose Z n := Zo Nat (fun m => n <c m /\ glt r (Vg f m) (Vg f n)).
have sif: forall n, natp n -> nonempty (Z n).
   move => n nN; apply /nonemptyP => ze; case hc; exists n => // m mN lemn.
   move: (qd _ _ nN mN lemn) => ha; ex_middle he; empty_tac1 m.
   by apply Zo_i => //; split => //; split => // hf; case he; rewrite hf.
pose F n := rep (Z n).
have Fp: forall n, natp n -> 
  [/\ natp (F n),  n <c (F n) & glt r (Vg f (F n)) (Vg f n)].
  by move => n /sif /rep_i /Zo_P [hu [hv hw]].
move: (induction_defined_pr F \0c) =>[].
set g := (induction_defined F \0c). 
move => sg sjg g0 gs.
have gN: forall n, natp n -> natp (Vf g n).
  apply: Nat_induction; first by rewrite g0; apply:NS0.
  by move => n nN hr; rewrite (gs _ nN); exact: (proj31 (Fp _ hr)). 
pose h n := Vg f (Vf g n).
have hr: forall n, natp n -> glt r (h (csucc n))  (h n) .
   move => n nN; rewrite /h (gs _ nN); exact:(proj33 (Fp _ (gN _ nN))).
have fgh: fgraph (Lg Nat h) by fprops.
exists (Lg Nat h); split => //; bw.
  move => t /(range_gP fgh); bw; move => [z zN ->]; bw.
  move : (proj1 (hr _ zN)) => le1; order_tac.
move => n m nN mN nm.
rewrite - (cdiff_pr (proj1 nm)).
move:(NS_diff n mN) (cdiff_nz nm).
move:(m -c n); clear mN nm m; apply: Nat_induction => //.
  move => m mN Ha _.
  have sN:= NS_sum nN mN.
  have sN1:= (NS_succ sN).
  rewrite (csum_nS _ mN) /h; bw;rewrite (gs _ sN).
  case (equal_or_not m \0c).
     move => ->; rewrite (csum0r (CS_Nat nN)).
     exact: (proj33 (Fp _ (gN _ nN))).
    move /Ha; move:(proj33 (Fp _ (gN _ sN))); rewrite /h; bw. 
   move => pa pb; exact: (lt_lt_trans (proj1 tor) pa pb).
Qed.

Theorem  finite_increasing_stationary r: order r -> 
  finite_set (substrate r) ->
  (forall f, increasing_sequence f r -> stationary_sequence f).
Proof. 
move=> or fs; apply /(increasing_stationaryP or). 
move=> X Xsr neX; move: (iorder_osr or Xsr) => [pa pb].
apply: finite_set_maximal => //; rewrite pb => //.
apply: (sub_finite_set Xsr fs).
Qed.

Theorem noetherian_induction r F:  order r ->
  (forall X, sub X (substrate r) -> nonempty X ->
    exists a,  maximal (induced_order r X) a) ->
  sub F (substrate r) ->
  (forall a, inc a (substrate r) -> (forall x, glt r a x -> inc x F)
    -> inc  a F) 
  -> F = substrate r.
Proof.
move=> or p1 p2 p3.
set (c := (substrate r) -s F).
case: (emptyset_dichot c) => ce.
  apply: extensionality => //; move=> x xsr; ex_middle xsf.
  empty_tac1 x; apply /setC_P;split => //.
have cs: sub c (substrate r) by rewrite /c; apply: sub_setC.
move: (p1 _ cs ce) => [a]; rewrite / maximal; aw; move => [ac am].
have aux: (forall y, glt r a y -> inc y F). 
  move=> y xy;  ex_middle w.
  have yc: (inc y c) by apply /setC_P;split => //; order_tac.
  move: xy => [xy nxy]; case: nxy.
  have ham: gle (induced_order r c) a y by  apply /iorder_gleP.
  by rewrite  (am _ ham).  
by move: (p3 _ (cs _ ac) aux); move: ac => /setC_P [].
Qed.



Definition perm_int n := (permutations (Nint n)).

Lemma perm_intP f n: inc f (perm_int n) <->
   [/\ bijection f, source f = (Nint n) &
       target f = (Nint n)].
Proof.
split.
  by move/Zo_P => [/fun_set_P [ff sf tf] bf];split.
move => [pa pb pc]; apply:Zo_i => //; apply /fun_set_P; split=> //; fct_tac.
Qed.

Lemma perm_int_id n: inc (identity (Nint n)) (perm_int n).
Proof.
apply/perm_intP.
move:(identity_prop (Nint n)) => [_ pa pb].
split => //; apply: identity_fb.
Qed.

Lemma perm_int_inj n f: natp n -> inc f (perm_int n) ->
    (forall x y, x <c n -> y <c n -> Vf f x = Vf f y -> x = y).
Proof.
move => nN /perm_intP[ [[_ H] _] sf _].
rewrite sf in H.
by move => x y /(NintP nN) xs /(NintP nN) ys sv; apply:H.
Qed.

Lemma perm_int_surj n f: natp n -> inc f (perm_int n) ->
   forall y, y <c n -> exists2 x, x<c n & Vf f x = y.
Proof.
move => nN /perm_intP[ [_ [_ H]] sf tf].
rewrite sf tf in H.
by move => y /(NintP nN) /H [x /(NintP nN) sa sb]; exists x.
Qed.

Lemma perm_int_comp n f g: 
    inc f (perm_int n) -> inc g (perm_int n) ->
    inc (f \co g) (perm_int n).
Proof.
move /perm_intP => [sa sb sc]/perm_intP [sa' sb' sc'].
apply/perm_intP; split;aw;apply: compose_fb => //. 
by split; [fct_tac | fct_tac | rewrite sb sc'].
Qed. 

Lemma perm_int_inv n f: inc f (perm_int n) -> 
  inc (inverse_fun f) (perm_int n).
Proof.
move /perm_intP => [sa sb sc]; apply /perm_intP; split ; aw.
by apply: inverse_bij_fb.
Qed.

Lemma transposition_prop n i j 
   (f:=Lf (fun z => Yo (z = i) j (Yo (z = j) i z)) (Nint n) (Nint n)):
   inc n Nat -> inc i (Nint n) -> inc j (Nint n) ->
   [/\ inc f (perm_int n), Vf f i = j, Vf f j = i,
     forall k, inc k (Nint n) -> k <> i -> k <> j -> (Vf f k)  = k &
     forall k, inc k (Nint n) -> Vf f (Vf f k) = k].
Proof.
move => nN iE jE.
set F := fun z => Yo (z = i) j (Yo (z = j) i z).
have la: lf_axiom F (Nint n) (Nint n).
   rewrite /F => z; Ytac h1; [ ue | Ytac h2; [ ue | done]].
have pa: Vf f i = j by rewrite /f; aw; Ytac0.
have pb: Vf f j = i by rewrite /f; aw; Ytac h1 => //; Ytac0.
have pc: forall k, inc k (Nint n) -> k <> i -> k <> j -> Vf f k = k.
  by move => k kE ki kj; rewrite /f; aw; Ytac0; Ytac0.
have pd:forall k, inc k (Nint n) -> Vf f (Vf f k) = k.
  move => k kE; case (equal_or_not k i) => ki; first by rewrite ki pa pb.
  case (equal_or_not k j) => kj; first by rewrite kj pb pa.
  by rewrite !(pc _ kE ki kj).
have sf: source f =  (Nint n) by rewrite /f;aw.
have tf: target f =  (Nint n) by rewrite /f;aw.
have ff: function f by  apply: (lf_function la).
have bf: bijection f.
  split; split => //; first by rewrite sf => u v  /pd {2} <- /pd {2} <- ->.
  rewrite sf tf => y ye;exists (Vf f y); [ Wtac | by rewrite pd].
split => //; apply /perm_intP; split => //.
Qed.


Lemma permutation_exists1 n i: natp n -> i <c n ->
    exists2 f, inc f (perm_int n) & Vf f \0c = i.
Proof.
move => nN lin.
move /(NintP nN): (cle_ltT (czero_least (proj31_1 lin)) lin) => oE.
move /(NintP nN):lin => iI.
move:(transposition_prop nN iI oE) => [pa _ pb _ _].
move: pa pb;set f := Lf _ _ _ => pa pb; ex_tac.
Qed.
 
Lemma permutation_exists2 E n: natp n -> sub E (Nint n) ->
   exists2 f, inc f (perm_int n) & E = image_by_fun f (Nint (cardinal E)).
Proof.
move => nN;move:n nN E; apply: Nat_induction.
  move => E; rewrite Nint_co00 => /sub_set0  ->.
  rewrite cardinal_set0 Nint_co00. 
  by have h := (perm_int_id \0c); ex_tac; rewrite  fun_image_set0. 
move => n nN Hrec E seI.
case: (equal_or_not E  (Nint (csucc n))) => Ev.
  rewrite Ev (card_Nint (NS_succ nN)).
  set F := Nint (csucc n).
  have bi:= (identity_fb F).
  exists (identity  F); first apply: perm_int_id.
  have {3} <-: source (identity F) = F by apply: identity_s.
  by rewrite -/(image_of_fun _) (surjective_pr0 (proj2 bi)) identity_t.
move: (setC_ne (conj seI Ev)) => [k1 /setC_P [k1I k1E]].
move:(Nint_pr4 nN) =>[].
set In:= Nint n; set Im := (Nint (csucc n)) => sa sb.
have Inm: forall i, inc i In -> (inc i Im /\ i <> n).
  by move => i ein; rewrite - sa; split; [ fprops | dneg w; ue].
have nIm: inc n Im by exact : (Nint_si nN).
set k0:= Yo (inc n E) k1 n.
have kI: inc k0 Im by rewrite /k0; Ytac h.
have kE: ~(inc k0 E) by rewrite /k0; Ytac h.
set F := (E +s1 k0) -s1 n.
have eq0: F +s1 n = E +s1 k0.
  rewrite /k0;set_extens t; case/setU1_P.
  + by move/setC1_P => [].
  + move => ->; Ytac h; fprops. 
  + move => tE; case (equal_or_not t n) => etn; first by rewrite etn; fprops.
    apply /setU1_P; left; apply/setC1_P; split ; fprops.
  + rewrite /F/k0;move => ->;Ytac h; fprops; case (equal_or_not k1 n). 
      by move ->; fprops.
    move => h'; apply/setU1_P; left; apply/setC1_P; fprops.
have fIn: sub F In.
  move => t /setC1_P [ta tb].
  have: inc t Im by case /setU1_P:ta; [apply: seI | move => ->].
  rewrite - sa; case /setU1_P => //.
move: (Hrec _ fIn) => [f /Zo_P [/fun_set_P[ff sf tf] fb] fv].
rewrite -/In in sf tf.
pose g i := Yo (i = n) k0 (Yo (Vf f i = k0) n (Vf f i)).
have pa: lf_axiom g Im Im.
  hnf; rewrite/g  - {1} sa; move => t;case /setU1_P; last by move => ->; Ytac0.
  move => ti;Ytac h => //; Ytac h' => //; rewrite - sa; apply/setU1_r;Wtac.
have sg:surjection (Lf g Im Im).
  apply: lf_surjective => // y; rewrite - {1} sa /g; case /setU1_P.
    case: (equal_or_not y k0) => yk0; first by exists n => //; Ytac0.
    rewrite - tf => /(proj2 (proj2 fb)) [x xsf fx].
    rewrite sf in  xsf; move: (Inm _ xsf) => [xa xb]; ex_tac.
    by rewrite fx; Ytac0; Ytac0.
  move => ->;move: (kI) ; rewrite -/Im - {1} sa; case /setU1_P.
    rewrite - tf; move/(proj2 (proj2 fb)) => [x xsf fx].
    rewrite sf in xsf; move: (Inm _ xsf) =>[xa xb]; ex_tac.
    by rewrite /g; Ytac0; Ytac0.
  by move => kn;exists k0 => //;rewrite /g; Ytac0.
set G:=  (Lf g Im Im).
have bg:bijection (Lf g Im Im).
  have cc: cardinal (source G) = cardinal (target G) by rewrite/ G; aw.
  have fsg: finite_set (source G). 
    by rewrite /G /Im; aw; apply:finite_Nint.
   exact:(bijective_if_same_finite_c_surj cc fsg sg).
have gP: inc G (permutations Im).
  apply: Zo_i => //; apply/fun_set_P; rewrite /G;split;aw; fct_tac.
have nf: ~ (inc n F) by move/setC1_P => [].
have nk0: ~ inc k0 E by rewrite /k0; Ytac h.
move: (f_equal cardinal eq0).
rewrite (csucc_pr nf)(csucc_pr nk0).
move/(succ_injective1 (CS_cardinal F)(CS_cardinal E)) => <-.
move:(sub_smaller fIn); rewrite /In (card_Nint nN) => le1.
move:(Nint_M nN); rewrite -/In -/Im => sub2.
move: fv (Nint_M1 nN le1).
set Ik:= (Nint (cardinal F)); rewrite -/In => fv sub1.
move: (sub_trans sub1 sub2) => sub3.
have fG: function G by fct_tac.
have ssG: sub Ik (source G) by rewrite /G; aw. 
have ssF: sub Ik (source f) by rewrite sf.
ex_tac.
set_extens t.
  move => tE;case (equal_or_not t n) => etn.
    move:(setU1_1 E k0); rewrite - eq0 => /setU1_P; case.
      rewrite fv; move => /(Vf_image_P ff ssF) [u ua ub].
      apply/(Vf_image_P fG ssG); ex_tac;rewrite /G; aw; last by apply:sub3.
      rewrite /g;Ytac h1; first by move: (sub1 _ ua); rewrite h1.
      by rewrite - ub; Ytac0.
    by move => w; case kE; rewrite w - etn.
  move: (setU1_r k0 tE); rewrite - eq0 => /setU1_P;case => //.
  rewrite fv; move => /(Vf_image_P ff ssF) [u ua ub].
  apply/(Vf_image_P fG ssG); ex_tac;rewrite /G; aw; last by apply:sub3.
  rewrite /g;Ytac h1; first by move: (sub1 _ ua); rewrite h1.
  by rewrite - ub; Ytac h=> //; case kE; rewrite - h.
move /(Vf_image_P fG ssG) => [u uik ->]; move:(sub3 _ uik)=> uIm.
rewrite /G; aw;rewrite /g; Ytac un.
   by case sb; rewrite -un; exact:(sub1 _ uik).
have: inc (Vf f u) F by rewrite fv; apply/(Vf_image_P ff ssF); ex_tac.
move/setC1_P => [ha hb];Ytac h1; first by move: h1; rewrite /k0;Ytac h2.
by case/setU1_P: ha.
Qed.


End InfiniteSets.
Export  InfiniteSets.

