Require Import ZArith.

Add LoadPath "/Users/grimm/Formel/ssreflect-1.2/theories" .

Require Import ssreflect ssrbool eqtype ssrnat.

Section HilbertAxiom.

Variables A B C : Prop.

Lemma HilbertS :  (A -> B -> C ) -> (A -> B) -> A -> C.
Proof.
  move=> hAiBiC hAiB hA.
  move: hAiBiC.
  apply.
    by [].
  by apply: hAiB.
Qed.

Hypotheses (hAiBiC : A -> B -> C)  (hAiB : A -> B) (hA : A).

Lemma Hilbert_S2: C.
Proof.
 apply: hAiBiC; first by apply: hA. 
 exact: hAiB.
Qed.


Lemma Hilbert_S3: C.
Proof.
  by apply: hAiBiC; last exact: hAiB.
Qed.


Lemma Hilbert_S4: C.
Proof. exact: (hAiBiC _ (hAiB _)).
Qed.


Lemma Hilbert_S5: C.
Proof. exact: hAiBiC  (hAiB _).  Qed.

Lemma Hilbert_S6: C.
Proof. exact:  Hilbert_S5.  Qed.

Print Hilbert_S6.
Print Hilbert_S5.
Print Hilbert_S2.
Print HilbertS.

End HilbertAxiom.

Print Hilbert_S5.

Print bool. 

Section Symmetric_Conjunction_Disjunction.


Lemma andb_sym:  forall A B: bool, A && B -> B && A.
Proof.
  case.
    by case.
 by [].
Qed.


Lemma andb_sym2:  forall A B: bool, A && B -> B && A.
Proof. by case ; case. Qed.


Lemma andb_sym3:  forall A B: bool, A && B -> B && A.
Proof. by do 2! case. Qed.

Variables (C D : Prop) (hC : C) (hD : D).
Check (and C D).
Print and.
Print conj.

Lemma and_sym: forall A B: Prop, A /\ B -> B /\ A.
Proof.  by move => A1 B []. Qed.


Lemma or_sym: forall A B: Prop, A \/ B -> B \/ A.
Proof. 
  by move=> A B [ hA | hB ] ; [apply: or_intror | apply: or_introl ].
Qed.


Lemma or_sym2: forall A B: bool, A \/ B -> B \/ A.
Proof. by move => [] [] AorB ; apply/ orP; move/orP : AorB.
Qed.
Check @orP.

End Symmetric_Conjunction_Disjunction.

Section R_sym_trans.

Variables (D : Type) (R: D -> D -> Prop).
Hypothesis R_sym: forall x y, R x y -> R y x.
Hypothesis R_trans: forall x y z, R x y -> R y z -> R x z.

Lemma refl_if: forall x : D,  (exists y, R x y) -> R x x.
Proof. 
  move => x [y Rxy].
  by apply: R_trans _ (R_sym _ y _ ).
Qed.

End R_sym_trans.

Section Smullyan_drinker.
Variables (D : Type) (P: D -> Prop).
Hypothesis (d : D) (EM : forall A, A \/ ~ A).

Lemma drinker : exists x, P x -> forall y, P y.
Proof.
 case: (EM (exists y, ~ P y)) => [ [y notPy] | notnotPy ]; first by exists y. 
 exists d => _ y ; case (EM  (P y)) =>  // notPy.
 by case: notnotPy;  exists y.
Qed.


Lemma drinker1 : exists x, P x -> forall y, P y.
Proof.
  have [ [y notPy] | notnotPy ] := EM (exists y, ~ P y) ; first by exists y. 
 exists d => _ y ; case (EM  (P y)) =>  // notPy.
 by case: notnotPy;  exists y.
Qed.

End Smullyan_drinker.

Section Equality.

Variable f : nat -> nat.
Hypothesis f00 : f 0 = 0.

Lemma fkk : forall k, k = 0 -> f k = k.
Proof.
  move => k k0.
  by rewrite k0.
Qed.

Lemma fkk1 : forall k, k = 0 -> f k = k.
Proof. by move => k  ->. Qed.


Variables f10 : f 1 = f 0.

Lemma ff10 : f (f 1) = f 0.
Proof. by  rewrite f10 f00. Qed.

Variables (D : eqType) (x y : D).

Lemma eq_prop_bool : x = y -> x == y.
Proof. by  move/eqP. Qed.


Lemma eq_bool_prop : x == y -> x = y.
Proof. by  move/eqP. Qed.

End Equality.

Section Using_Definition.

Variable U : Type.
Definition set := U -> Prop.
Definition subset  (A  B: set) := forall x, A x ->  B x.
Definition transitive (T: Type) (R : T -> T -> Prop) :=
  forall x y z, R x y -> R y z -> R x z.


Lemma subset_trans: transitive set subset.
Proof.
rewrite /transitive /subset => x y z subxy subyz t tx.
by apply: subyz; apply: subxy.
Qed.

Lemma subset_trans2: transitive set subset.
Proof.
move => x y z subxy subyz t.
by move/subxy; move/subyz.
Qed.

End Using_Definition.

Lemma three : S (S (S O)) = 3 /\ 2 = 0.+1.+1.
Proof. by []. Qed.

Print plus.

Lemma concret_plus: plus 16 64 = 80.
Proof.   by []. Qed.

Print addn. 

Lemma concret_plus_bis: 16 + 64 = 80.
Proof. simpl. by []. Qed.

Print le.

Lemma concrete_le : le 1 3.
Proof. by apply: (Le.le_trans _ 2); apply: Le.le_n_Sn. Qed.


Lemma concrete_big_le : le 16 64.
Proof. by auto 47 with arith. Qed.
Print concrete_big_le.

Print leq.
Print subn. 
Print subn_rec.


Lemma concrete_big_leq :  16 <= 64.
Proof. by []. Qed.
Print concrete_big_leq.


Lemma concrete_big_leq2 :  0 <= 64.
Proof. by []. Qed.
Print concrete_big_leq2.

Lemma semi_concrete_leq : forall n m, n <= m ->  51 +n <= 51 + m.
Proof. by []. Qed.
Print semi_concrete_leq.

Lemma semi_concrete_leq2 : forall n m, n <= m -> n+ 51 <= m + 51.
Proof. move => n m H.   by rewrite addnC; rewrite [m + _ ] addnC. Qed.


Lemma concrete_arithmetic : (50< 100) && (3+4 < 3*4 <= 17-2).
Proof. by []. Qed.

Lemma plus_commute : forall n1 m1, n1 + m1= m1 + n1.
Proof.
  by elim => [| n IHn m] ; [elim | rewrite -[n.+1+m ] /(n+m).+1 IHn; elim: m].
Qed.


Definition  edivn_rec d := fix loop (m q : nat) { struct m} :=
  if m - d is m'.+1 then loop m' q.+1 else (q, m).

Definition edivn m d := if d >0 then edivn_rec d.-1 m 0 else (0, m).

CoInductive edivn_spec (m d : nat) : nat * nat -> Type :=
  EdivnSpec q r of m = q * d + r & (d>0) ==> (r < d) :
  edivn_spec m d (q,r).


(* alt syntax *)
CoInductive edivn_specA (m d : nat) : nat * nat -> Type :=
  EdivnSpecA: forall q r, m = q * d + r -> (d>0 -> r < d) ->
    edivn_specA m d (q,r).
Print edivn_specA. Print edivn_spec.


Lemma edivnP: forall m d, edivn_spec m d (edivn m d).
Proof.
rewrite /edivn => m [|d] //=; rewrite -{1}[m]/(0 * d.+1 + m).
elim: m {-2}m 0 (leqnn m) => [| n IHn] [| m] q //=; rewrite ltnS => le_mn.
rewrite subn_if_gt; case: ltnP => [// | le_dm].
rewrite -{1}(subnKC le_dm) -addSn  addnA -mulSnr; apply  :IHn.
apply: leq_trans le_mn; exact: leq_subr.
Qed.


Lemma edivn_eq : forall d q r, r < d -> edivn (q * d + r) d = (q, r).
Proof.
move=> d q r lt_rd; have d_gt0: 0 < d by exact: leq_trans lt_rd.
case: edivnP lt_rd => q' r' ; rewrite d_gt0 /=.
wlog: q q' r r' / q <= q' by case (ltnP q q'); last symmetry; eauto.
rewrite leq_eqVlt ; case: eqP  => [-> _|_] /=; first by move/addnI->.
rewrite -(leq_pmul2r d_gt0); move/leq_add => Hqr Eqr  _; move/Hqr {Hqr}.
by rewrite addnS ltnNge mulSn -addnA Eqr addnCA addnA leq_addr. 
Qed.

Check le_irrelevance.


(* OK *)